c---+----1----+----2----+----3----+----4----+----5----+----6----+----7--
c
c **********************************************************************
c * ****************************************************************** *
c * *                                                                * *
c * * MMM     MMM HHH  HHH DDDDD    RRRRRR    MMM     MMM NNN    NNN * *
c * *  MMM   MMM   HH  HH   DD  DD   RR   RR   MMM   MMM   NNN   NN  * *
c * *  MMM   MMM   HH  HH   DD   DD  RR    RR  MMM   MMM   NNN   NN  * *
c * *  MM M M MM   HH  HH   DD   DD  RR   RR   MM M M MM   NN N  NN  * *
c * *  MM M M MM   HHHHHH   DD   DD  RRRRR     MM M M MM   NN NN NN  * *
c * *  MM  M  MM   HH  HH   DD   DD  RR  RR    MM  M  MM   NN  N NN  * *
c * *  MM     MM   HH  HH   DD   DD  RR  RR    MM     MM   NN   NNN  * *
c * *  MM     MM   HH  HH   DD  DD   RR   RR   MM     MM   NN   NNN  * *
c * * MMMM   MMMM HHH  HHH DDDDD    RRR   RRR MMMM   MMMM NNN    NNN * *
c * *                                                                * *
c * ****************************************************************** *
c **********************************************************************
c
c
c                MHDRMN, VERSION 1.1, RELEASE OF 15DEC21
c
c                           David Alan Clarke
c                  Department of Astronomy and Physics
c                  Saint Mary's University, Halifax NS
c
c  PREAMBLE: From initial left and right states provided by the user,
c  this program computes the solution to the MHD Riemann problem, which
c  is unique when intermediate shocks are not considered.
c
c  This program is based in part on Torrilhon, 2002, research report
c  2002-06, Zurich: Seminar for Applied Mathematics, ETH (hereafter T02)
c  and Takahashu & Yamada, 2013, J. Pl. Ph., v. 79, p. 335 (hereafter
c  TY13), and in large part on Takahashu & Yamada, 2014, J. Pl. Ph., v.
c  80, p. 255 (hereafter TY14).
c
c  This program needs to be linked to library ~dclarke/libs/gfortran/
c  lib1dplot.a on compilation.  This is done automatically by the script
c  file mhdrmn11.s.
c
c  In addition to a brief summary of the converged solution that gets
c  printed to the screen, this program generates three output files:
c
c  mhdrmn.out   ascii file with the convergent values of the parameters,
c  psi_j, and the values of the primitive variables in each of the eight
c  states across the solution space (similar to the data reported to the
c  screen).
c
c  mhdrmn.dat   a ten-column ascii file suitable for use by widely-
c  available graphics packages such as GNUPLOT able to read and plot
c  column-separated datafiles.  Columns include the independent (self-
c  similar) position variable x/t, followed by nine variables d, p, e_T,
c  v_x, v_y, v_z, B_y, B_z, chi, where e_T is the total energy density
c  and chi = atan (B_y/B_z).
c
c  mhdrmn.ps    a post-script plot file with nine frames plotting each
c  of the variables listed above.
c
c
c  Notes from DAC, 5/12/21:
c
c  1.  Of the twelve Riemann problems from Ryu & Jones, two from Falle,
c      and one each from T02, TY13, and Jon, Jon's is the outrigger.  In
c      order for this program to converge on Jon's problem, I've had to
c      introduce the capability of jumbling up the initial guesses
c      ("phi0" using RAN2) when the secant root finder gets lost as it
c      does twice, first time because of a singular Jacobian, and the
c      second time because the maximum number of secant cycles was
c      exceeded (trying to widen an asymptoted fan).  On the third try,
c      convergence happens in 8 cycles.  This may yet point to a
c      weakness in the algorithm.
c
c  2.  In TY13, they claim some Riemann problems don't have a "regular"
c      solution.  In the literature, regular solutions generally don't
c      include "switch-off" waves, and in their example (inrmnty), the
c      only solution they find (and the one this routine finds) has a
c      switch-off shock and a switch-off rarefaction.  They are not, for
c      example, reporting a problem whose only solution includes an
c      intermediate shock.  I question, therefore, categorising "switch-
c      off" shocks as non-regular.  They are at the boundary between
c      slow and intermediate shocks.
c
c  3.  There remain vulnerabilities in this program where the Jacobian
c      can still be singular (e.g., point 1) and where fast and slow
c      wave parameters can skip off to infinity if the algorithm gets
c      into a local minimum where the only path to improvement is to
c      strengthen the wave.  I have set traps for when rarefactions get
c      too deep; it's not yet clear to me what the best way is for
c      shocks.
c
c  4.  Further traps in LUSOLVE, RAREFACT, and DFDPSI have been
c      introduced to detect when the secant solver is getting lost and a
c      parameters need to be reinitialised.  This has cured most of the
c      woes I had previously for R&J Fig4b.
c
c  5.  I have introduced coding to retain the "best solution found"
c      should the number of trials be exceeded.  For many problems, this
c      can give a solution whose error is perhaps one or two orders of
c      magnitude greater than the stated tolerance and may still be
c      "good enough".
c
c  6.  For the TY13 problem (inrmnty), I get two RDs, each by pi radians
c      (one where bperp=0!!), whereas TY13 do not.  Another "glitch" in
c      the program somewhere...
c
c=======================================================================
c
c    \\\\\\\\\\         B E G I N   P R O G R A M         //////////
c    //////////                M H D R M N                \\\\\\\\\\
c
c=======================================================================
c
       program mhdrmn
c
c    dac:mhdrmn <---------------------- evolutionary MHD Riemann problem
c                                                             june, 2018
c
c    written by: David Clarke
c    modified 1: November, 2021 by DAC; introduced "best solution" in
c                the event that "err" never converges to "merr".
c
c  PURPOSE: This program performs the iterations of the multi-variate
c  secant solver to solve the MHD Riemann problem including all switch-
c  on and switch-off waves, but excluding intermediate shocks (thereby
c  invoking the "evolutionary condition").  The final solution is then
c  reported and written to files suitable for visualisation for both
c  GNUPLOT and PSPLOT.
c
c  With no eigenvalue degeneracies, the solution to the Riemann problem
c  consists of eight states (including the original left and right
c  states) divided by seven families of waves, as depicted in the figure
c  below.  Primitive variables (rho, p, vx, vy, vz, By, Bz) are given in
c  each of the left and right states (states 1 and 8 in the figure),
c  with Bx also given and taken as the same for all states.
c
c                   t |
c                     4 C                  F = fast wave
c          S- 3       |       5 S+         R = rotational discontinuity
c              \      |      /             S = slow wave
c    R- 2.      \  4  |  5  /      .6 R+   C = contact discontinuity
c         `.  3  \    |    /  6  .'        +/- => right/left-moving wave
c  F- 1    `.     \   |   /    .'     7 F+
c     ` .  2  `.   \  |  /   .'  7  . '
c         ` .   `.  \ | /  .'   . '        state 1 = left  initial state
c         1   ` . `. \|/ .' . '   8        state 8 = right initial state
c    _ _ _ _ _ _ _`_`_V_'_'_ _ _ _ _ _ _ _
c                                        x
c
c  See subroutine FOFPSI for further details on the nature of the
c  Riemann solution.
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    NMLSTS
c    FOFPSI
c    COPYSOLN
c    SETPSIR
c    DFDPSI
c    LUSOLVE
c    REINIT
c    RIEMOUT
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       character*7   word
       integer       iter    , titer   , j       , jay     , soln
       real          cpubeg  , cpuend  , cputot
       real*8        err     , besterr
c
       real*8        psi     (mpar), fpsi    (mpar), dpsir   (mpar)
     1             , bestpsi (mpar)
c
       real*8        psir    (mpar,mpar), fpsir   (mpar,mpar)
     1             , jcbn    (mpar,mpar)
c
c      External statements
c
       external      nmlsts  , fofpsi  , copysoln, setpsir , dfdpsi
     1             , lusolve , reinit  , riemout
c
c-----------------------------------------------------------------------
c
c-----------------------------------------------------------------------
c-- opening procedures -------------------------------------------------
c-----------------------------------------------------------------------
c
c      Start up CPU time counter.
c
       call cpu_time ( cpubeg )
c
c      Set logical units, and initialise some variables.
c
       iotty = 6
       ioin  = 10
       ioout = 11
       ioplt = 99
       nwarn = 0
c
c      Open input file.
c
       open ( unit=ioin , file='inrmn', status='old' )
c
c      Opening banner.
c
       write ( iotty, 2010 )
c
c      Read namelists.
c
       call nmlsts
c
c      Initialise the number of trials taken; set the target for when a
c  new trial is needed.
c
       ntry    = 1
       soln    = 1
       besterr = huge
       titer   = 0
10     continue
c
c      Initialise secant convergence parameters "npar".  If "bx0" was
c  set by user to be "too close" to zero, set it to zero.
c
       npar = mpar
       if (abs(bx0) .le. zero) then
         npar = 2
         bx0  = 0.0d0
       endif
c
c      Initialise all other variables ("psi0" was read in NMLSTS or
c  reset by REINIT).  The flag "soln" is used by FOFPSI (RAREFACT) to
c  indicate whether the RF parameter causes density or pressure to go
c  negative, by DFDPSI in case all rows/columns of the Jacobian are
c  zero, and by LUSOLVE to indicate whether the Jacobian is singular.
c
       do 20 j=1,npar
         psi (j) = psi0(j)
         dpsi(j) = dpsi0
20     continue
       do 30 j=1,4
         vshk(j) = 0.0d0
30     continue
       soln = 1
c
c-----------------------------------------------------------------------
c----- Top of secant loop ----------------------------------------------
c-----------------------------------------------------------------------
c
       word = 'initial'
       iter = 0
40     continue
c
c      Evaluate convergence functions and calculate error.
c
         call fofpsi ( psi, fpsi, soln )
         if (soln .eq. 0) go to 10
         if (soln .lt. 0) go to 90
         err = 0.0d0
         do 50 j=1,npar
           err = err + fpsi(j)**2
50       continue
         err = sqrt ( err )
c
c      If "fpsi" is zero to within tolerance, report root values and
c  their error, and then exit secant loop.
c
         if (err .le. merr) then
           write ( iotty, 2020 ) iter, err, (psi(j),j=1,npar)
           go to 90
         endif
c
c      Keep track of the best solution so far.  In the event that "err"
c  never converges to "merr", this "best solution" will be better than
c  nothing!
c
         if (err .lt. besterr) then
           besterr = err
           do 51 j=1,npar
             bestpsi(j) = psi(j)
51         continue
           call copysoln ( 1 )
         endif
c
c      Report iteration progress.
c
         write ( iotty, 2030 ) iter, err, word, (psi(j),j=1,npar)
         word = 'current'
c
c      In preparation to compute corrections for the current guesses,
c  |dpsi>, compute the function Jacobian, "jcbn".  DFDPSI will return
c  "jcbn" without any rows or columns of zeros, and thus the dimension
c  of "jcbn", "npar", may have been reduced.
c
         call setpsir ( psi, dpsir, psir )
         do 60 j=1,npar
           call fofpsi ( psir(1,j), fpsir(1,j), soln )
           if (soln .eq. 0) go to 10
           if (soln .lt. 0) go to 90
60       continue
         call dfdpsi ( fpsi, dpsir, fpsir, jcbn, soln )
         if (soln .eq. 0) go to 10
         if (soln .lt. 0) go to 90
c
c      Solve the matrix equation jcbn |dpsi> = -|fpsi> for |dpsi>.  If
c  the flag "soln" is returned 0, the Jacobian provided no solution (is
c  singular), and LUSOLVE reinitialised the parameters "psi0".  In this
c  case return to target 10 for a new trial.  If "soln" is returned -1,
c  maximum number of trials have been executed, and we're giving up.
c
         call lusolve ( jcbn, fpsi, -1.0d0, dpsi, soln )
         if (soln .eq. 0) go to 10
         if (soln .lt. 0) go to 90
c
c      If DFDPSI has struck out any rows/columns of "jcbn", the vector
c  "dpsi" may no longer align with the vector "psi".  In this case,
c  restore "npar" to "mpar", and realign "dpsi" with "psi" by inserting
c  initial values for "dpsi" where needed.  Loop 70 is done backwards to
c  prevent the assignments of "dpsi" from clobbering itself.
c
         if (npar .lt. mpar) then
           jay  = npar
           npar = mpar
           if (bx0 .eq. 0.0) npar = 2
           do 70 j=npar,1,-1
             if (keeprow(j) .eq. 0) then
               dpsi(j) = dpsi0
             else
               dpsi(j) = dpsi(jay)
               jay     = jay - 1
             endif
70         continue
         endif
c
c      Increase iteration counter, "iter".  If "iter" exceeds "miter",
c  the secant root finder is probably lost.  Issue a warning message,
c  reinitialise parameters "psi0", and try again (if "mtry" has not been
c  exceeded).
c
         titer = titer + 1
         iter  = iter  + 1
         if (iter .gt. miter) then
           ntry = ntry + 1
           if (ntry .le. mtry) then
             write ( iotty, 2040 ) miter
             nwarn = nwarn + 1
             call reinit
             go to 10
           else
             soln =-1
             go to 90
           endif
         endif
c
c      Evaluate the next secant guess, and go back for next iteration.
c
         do 80 j=1,npar
           psi(j) = psi(j) + dpsi(j)
80       continue
         psi(5) = mod ( psi(5), 2.0d0*pi )
         go to 40
c
90     continue
c
c-----------------------------------------------------------------------
c----- Bottom of secant loop -------------------------------------------
c-----------------------------------------------------------------------
c
       if (soln .lt. 0) then
c
c      Execution arrives here only if the number of trials, "ntry",
c  exceeds "mtry", and thus no solution was found to within the desired
c  tolerance.  Issue a warning message and send the best solution found
c  (that with the lowest error) to RIEMOUT for output.
c
         write ( iotty, 2050 ) mtry, merr
         nwarn = nwarn + 1
         err   = besterr
         do 100 j=1,npar
           psi(j) = bestpsi(j)
100      continue
         call copysoln ( 2 )
         write ( iotty, 2060 ) titer, err, (psi(j),j=1,npar)
c
       endif
c
c      RIEMOUT reports the final solution and assembles the profiles of
c  the solution into arrays used to generate an ascii file suitable for
c  GNUPLOT and a publication-quality postscript file of nine profiles.
c
       call riemout ( psi, err )
c
c      End CPU time counter, report cpu time used, and close out.
c
       call cpu_time ( cpuend )
       cputot = cpuend - cpubeg
       write ( iotty, 2070 ) cputot, nwarn
       close ( ioin )
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-------------0----+----1----+----2----+----3----+----4----+----5----+--
c
2010   format(///,5x,'M   M H   H DDDD  RRRR  M   M N   N',/
     1           ,5x,'MM MM H   H  D  D R   R MM MM NN  N'
     3           ,5x,'MHDRMN1.1 - RELEASE OF 15DEC21',/
     2           ,5x,'M M M HHHHH  D  D RRRR  M M M N N N'
     3           ,8x,'Written by David Clarke',/
     4           ,5x,'M   M H   H  D  D R  R  M   M N  NN'
     3           ,8x,'Saint Mary''s University',/
     5           ,5x,'M   M H   H DDDD  R   R M   M N   N',///)
2020   format('MHDRMN  :',/
     1       ,'MHDRMN  : After ',i3,' iterations, err =',1pg12.5
     1       ,', final   guess for roots are:'/
     3       ,'MHDRMN  : ',5g13.5,/
     4       ,'MHDRMN  :')
2030   format('MHDRMN  : After iteration ',i4,', err =',1pg12.5,', ',a7
     1       ,' guess for roots are:'/
     3       ,'MHDRMN  : ',5g13.5)
2040   format('MHDRMN  : |---WARNING---> Maximum number of iterations, '
     1       ,i3,', exceeded. ',/
     2       ,'MHDRMN  : Resetting initial guess ("psi0") and '
     1       ,'restarting secant solver.')
2050   format('MHDRMN  : |---WARNING---> Maximum number of trials '
     1       ,i3,' exceeded.  Total error',/
     2       ,'MHDRMN  : reported is not within specified tolerance '
     2       ,'(merr):',1pg12.5,'.')
2060   format('MHDRMN  :',/
     1       ,'MHDRMN  : After ',i3,' iterations, err =',1pg12.5
     1       ,', *best*  guess for roots are:',/
     3       ,'MHDRMN  : ',5g13.5,/
     4       ,'MHDRMN  :')
2070   format('MHDRMN  :',/
     1       ,'MHDRMN  : Total cpu usage for this run is ',1pg12.5
     1       ,' seconds.',/
     1       ,'MHDRMN  : Execution terminated with ',i4
     1       ,' warnings issued.')
c
       stop
       end
c
c=======================================================================
c
c    \\\\\\\\\\           E N D   P R O G R A M           //////////
c    //////////                M H D R M N                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\        B E G I N   F U N C T I O N        //////////
c    //////////                A D D O N E                \\\\\\\\\\
c
c=======================================================================
c
       function addone ( i )
c
c    dac:mhdrmn.addone <------------------------------------ adds 1 to i
c    from PHYS3210/assign07/runkut.f                          june, 2018
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE:  This function adds one to the input integer, "i", so long
c  as the addition doesn't cause "i" to exceed "imax", the parameter in
c  common deck PAR.  If "i" does exceed "imax", an error message is
c  issued, and execution is aborted.
c
c  INPUT VARIABLES:
c    i           integer to add one to
c
c  OUTPUT VARIABLES:
c    addone      i+1, if i+1 .le. imax
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       integer       addone  , i
c
c-----------------------------------------------------------------------
c
       addone = i + 1
       if (addone .gt. imax) then
         write ( iotty, 2010 ) imax
         stop
       endif
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-------------0----+----1----+----2----+----3----+----4----+----5----+--
c
2010   format('ADDONE  : |---ERROR---> Attempt to add another plot '
     1       ,'point exceeds imax = ',i5,'.',/
     2       ,'ADDONE  :  Abort!')
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\          E N D   F U N C T I O N          //////////
c    //////////                A D D O N E                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\        B E G I N   F U N C T I O N        //////////
c    //////////               A R C S I N E               \\\\\\\\\\
c
c=======================================================================
c
       function arcsine ( dx, dy )
c
c    dac:zeus3d.arcsine <--------- returns an arcsine between -pi and pi
c                                                         february, 1992
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE:  Returns an arcsine between -pi and pi.  The branch point
c  lies along the negative x-axis.
c
c  INPUT VARIABLES:                         +y |
c    dx          run  (along +ve x-axis)       |--dx--|
c    dy          rise (along +ve y-axis)       |         _
c                                              |      /  |
c                                              |     /   |
c                                              |    /    |
c                                              |   /     dy
c                                              |  /      |
c                                              | / th    |
c                                              |/________|____+x
c  OUTPUT VARIABLES:
c    arcsine     value of "th" in diagram, such that:
c                0   < th < pi for angles counter-clockwise from +x-axis
c                -pi < th < 0  for angles         clockwise from +x-axis
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c
       real*8        dx      , dy      , arcsine
       real*8        r       , sgndx   , sgndy   , sine
c
c-----------------------------------------------------------------------
c
       r       = sqrt ( dx**2 + dy**2 )
       sgndx   = sign ( 1.0d0, dx )
       sgndy   = sign ( 0.5d0, dy )
       sine    = max ( -1.0d0, min ( 1.0d0, dy/(r+tiny) ) )
       arcsine = sgndx * asin(sine) + ( 1.0d0 - sgndx ) * sgndy * pi
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\          E N D   F U N C T I O N          //////////
c    //////////               A R C S I N E               \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////               C A R T A R 2               \\\\\\\\\\
c
c=======================================================================
c
       subroutine cartar2 ( a0, a1, a2, bperp, root )
c
c    dac:mhdrmn.cartar2 <----------- cardano-tartaglia formula for cubic
c    from dac:mhdjmp.cartar                                   june, 2015
c
c    written by: David Clarke
c    modified 1: June, 2018 by DAC; modified to return just the largest
c                real root
c
c  PURPOSE: This routine computes the three complex roots of a general
c  cubic using a modern version of the Cardano-Tartaglia formula for the
c  roots of a cubic (App. F of MHD text), and then returns the largest
c  real part in "root".
c
c  INPUT VARIABLES:
c    a0          real coefficient of constant  term
c    a1          real coefficient of linear    term
c    a2          real coefficient of quadratic term
c    bperp       magnitude of perpendicular component of B
c
c  OUTPUT VARIABLES:
c    root        the largest (real part) of the three roots
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c
       real*8        a0      , a1      , a2      , bperp   , root
     1             , disc    , del0    , del1    , q1
       complex*16    cdel0   , cdel1   , ca2     , omega   , three
     1             , qty     , zeta    , root1   , root2   , root3
c
c-----------------------------------------------------------------------
c
c      Compute the real constants.
c
c  One can show that "q1" (and thus the discriminant) is identically
c  zero if B_perp is zero (and thus why it is passed from the calling
c  routine).  Since "q1" is a difference, should the two quantities be
c  equal to within round off noise, "q1" could be negative, posing a
c  problem for the square root below.  Thus the check here.
c
       del0 = a2**2 - 3.0d0 * a1
       del1 = a2**3 - 4.5d0 * a2 * a1 + 13.5d0 * a0
       q1   = del1**2 - del0**3
       if (bperp .eq. 0.0d0) q1 = 0.0d0
       disc =-4.0d0 * q1 / 27.0d0
c
c      Set some complex constants.
c
       cdel0 = dcmplx ( del0 , 0.0d0 )
       cdel1 = dcmplx ( del1 , 0.0d0 )
       ca2   = dcmplx ( a2   , 0.0d0 )
       omega = dcmplx (-0.5d0, 0.5d0 * dsqrt(3.0d0) )
       three = dcmplx ( 3.0d0, 0.0d0 )
       qty   = sqrt ( dcmplx ( q1, 0.0d0 ) )
       if (del0 .eq. 0.0d0) qty = cdel1
       zeta  = ( cdel1 + qty )**(1.0d0/3.0d0)
c
c      Compute the three roots.
c
       root1 =-( ca2 +            zeta + cdel0            / zeta )
     1       / three
       root2 =-( ca2 + omega    * zeta + cdel0 / omega    / zeta )
     1       / three
       root3 =-( ca2 + omega**2 * zeta + cdel0 / omega**2 / zeta )
     1       / three
c
c      For the fast shock, return the maximum of the real part of the
c  three roots, which needs to be capped at 1.0 in case of round-off
c  error.
c
       root = max ( real(root1), real(root2), real(root3) )
       root = min ( 1.0d0, root )
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////               C A R T A R 2               \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              C O P Y S O L N              \\\\\\\\\\
c
c=======================================================================
c
       subroutine copysoln ( flag )
c
c    dac:mhdrmn.copysoln <-------------- copies to or from best solution
c                                                         november, 2021
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE:  Depending on "flag", this routine either saves the current
c  solution in "pvar" "vshk", "yrf", and "irf" to the best-solution-so-
c  far variables "bestpvar", "bestvshk", "bestyrf", and "bestirf" or, in
c  preparation for output by RIEMOUT, the reverse.  All variables are in
c  common block comvarr.
c
c  INPUT VARIABLES:
c    flag        =1 => copy "pvar", "vshk", "yrf", and "irf" to
c                      "bestpvar", "bestvshk", "bestyrf", and "bestirf";
c                otherwise, the reverse.
c
c  OUTPUT VARIABLES:
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       integer       flag    , i       , n       , j
c
c-----------------------------------------------------------------------
c
       if (flag .eq. 1) then
c
c      Error for current solution is the lowest so far.  In case it's
c  the best one we'll ever find, save it to the best-solution variables.
c
         do 20 i=1,8
           do 10 n=1,node
             bestpvar(n,i) = pvar(n,i)
10         continue
20       continue
         do 50 j=1,4
           bestirf (j) = irf (j)
           bestvshk(j) = vshk(j)
           do 40 n=1,node
             do 30 i=1,irf(j)
               bestyrf(i,n,j) = yrf(i,n,j)
30           continue
40         continue
50       continue
c
       else
c
c      If execution gets here, the maximum number of trials, "mtry", has
c  been reached and we're giving up.  To report the best solution found
c  even if its error is not below tolerance, copy the best-solution
c  variables to the solution variables in preparation for RIEMOUT.
c
         do 70 i=1,8
           do 60 n=1,node
             pvar(n,i) = bestpvar(n,i)
60         continue
70       continue
         do 100 j=1,4
           irf (j) = bestirf (j)
           vshk(j) = bestvshk(j)
           do 90 n=1,node
             do 80 i=1,irf(j)
               yrf(i,n,j) = bestyrf(i,n,j)
80           continue
90         continue
100      continue
c
       endif
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////              C O P Y S O L N              \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////               D E R I V S 7               \\\\\\\\\\
c
c=======================================================================
c
       subroutine derivs7 ( y, dpacc, yp )
c
c    dac:mhdrmn.derivs7 <--------- evaluate derivatives for R-K solution
c    from dac:fans10:derivs                                    may, 2018
c
c    written by: David Clarke
c    modified 1: February 2019, by DAC; aligned to version in FANS10.
c
c  PURPOSE:  For the 7 coupled 1st-order ODEs describing the rarefaction
c  fan, this routine returns the values of all function derivatives, and
c  thus the corresponding components in the right eigenket.
c
c  INPUT VARIABLES:
c    y           1-D array of the "node" dependent variables,
c    dpacc       accumulated difference between pressure and critical
c                pressure, pcr = bx0**2/gamma.  Used by SPEEDS to
c                calculate "deltax" should "alphax" get too close to 1.
c
c  OUTPUT VARIABLES:
c    yp          1-D array of derivatives needed by RK7.
c
c  EXTERNALS:
c    SPEEDS
c    MUNU
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       real*8        dpacc   , mu      , nu      , pm1
c
       real*8        y       (   *), yp      (   *)
c
       external      speeds  , munu
c
c-----------------------------------------------------------------------
c
       call speeds ( y(1), dpacc )
c
c      Compute "mu" and "nu", the "normalisation" factors for the fast
c  and slow eigenkets.
c
       call munu ( mu, nu )
c
c      The only difference between the left- and right- waves is a sign
c  in some terms, embodied by "pm1".
c
       pm1 = 1.0d0
       if (family .lt. 4 ) pm1 =-1.0d0
c
       if ( (family .eq. 1) .or. (family .eq. 7) ) then
c
c----- FAST RAREFACTION FANS -------------------------------------------
c
c      The fact there is no fast fan for bperp = 0 and alphax .ge. 1 is
c  handled in RAREFACT which sets the primitive variables constant
c  across the wave and its secant parameter "psi" to zero.  Here, we set
c  the fast eigenket for a fast regular fan (bperp > 0, any alphax), a
c  switch-off fan (saturated fast regular fan), or a fast Euler fan
c  (bperp = 0, alphax < 1) both unsaturated and saturated.
c
         yp(1) =-mu * y(1)
         yp(2) =-mu * gamma * y(2)
         yp(3) =-mu * pm1 * af
         yp(4) = nu * pm1 * sgnbx0 * as * coschi
         yp(5) = nu * pm1 * sgnbx0 * as * sinchi
         yp(6) =-nu * sqrt ( gamma * y(2) ) * coschi
         yp(7) =-nu * sqrt ( gamma * y(2) ) * sinchi
c
       else
c
c----- SLOW RAREFACTION FANS -------------------------------------------
c
c      The fact there is no slow fan for bx0 = 0 is handled in RAREFACT
c  which sets the primitive variables constant across the wave and skips
c  this routine altogether.  Here, we set the slow eigenket for a slow
c  regular fan (bperp > 0, any alphax), a switch-on fan (bperp = 0,
c  alphax < 1), or a slow Euler fan (bperp = 0, alphax .ge. 1).
c
         yp(1) =-nu * y(1)
         yp(2) =-nu * gamma * y(2)
         yp(3) =-nu * pm1 * as
         yp(4) =-mu * pm1 * sgnbx0 * af * coschi
         yp(5) =-mu * pm1 * sgnbx0 * af * sinchi
         yp(6) = mu * sqrt ( gamma * y(2) ) * coschi
         yp(7) = mu * sqrt ( gamma * y(2) ) * sinchi
c
       endif
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////               D E R I V S 7               \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                D F D P S I                \\\\\\\\\\
c
c=======================================================================
c
       subroutine dfdpsi ( fpsi, dpsir, fpsir, jcbn, soln )
c
c    dac:mhdrmn.dfdpsi <---------------- sets function Jacobian elements
c                                                             june, 2018
c
c    written by: David Clarke
c    modified 1: December 2021, by DAC; introduced "soln" sentinel used
c                in case "jcbn" has all zeros and "npar" is set to 0.
c
c  PURPOSE:  This routine returns approximate values of the npar X npar
c  elements of the Jacobian matrix for the functions in FOFPSI.  Here,
c  we use finite differences based on the function values at the current
c  best guess of the root ("fpsi"), and values of the functions at the
c  retarded parameters ("fpsir"), all sent to this routine via the
c  calling list.
c
c  The (i,j) element of the Jacobian is given by:
c
c          d f_i(psi)     f_i(psi(1:5)) - f_i(psir(1:5,j))
c          ----------  =  ----------------------------- ,       (1)
c           d psi(j)                dpsir(j)
c
c  where psir(1:5,j) are the same as psi(1:5) except the jth value has
c  been replaced with its "backward-stepped value" [psi(j)-dpsi(j)].
c  This is the multi-variate equivalent of the secant estimate of the
c  derivative for functions of one variable.
c
c  The trickiest part of this entire program is in managing the Jacobian
c  which can, for numerous reasons, be singular.  The most common reason
c  for singularity is the appearance of a row or column of (near) zeros.
c
c  A column (thinking of linear algebra matrices, and thus constant "j";
c  see note below on row-major vs. column-major) of zeros means none of
c  the functions depend upon psi(j) [all derivatives with respect to
c  psi(j) being zero].  Thus, at least for this round, this parameter is
c  not relevant in determining whether the functions have converged and
c  this column can be eliminated from the Jacobian.
c
c  A row (constant "i") of zeros is usually accompanied by the ith value
c  of "fpsi" being zero, indicating that particular function has
c  converged.  Thus, at least for this round, f_i(psi) need not be
c  considered in determining improvements for the parameters, and this
c  row can be eliminated from the Jacobian.
c
c  Since the Jacobian is a square matrix, for every row/column removed,
c  a corresponding column/row must be removed.  According to TY14, it
c  doesn't matter much which column/row is removed in squaring up the
c  matrix, although I find that it is best to remove the column/row with
c  the smallest elements in them.  It is often the case that when a row
c  of (near) zeros appears, a column of (near) zeros also appears (for
c  every converged function, there's often an irrelevant parameter), and
c  so this approach can remove both converged function(s) and irrelevant
c  parameter(s) at the same time.
c
c  Note that "npar" is reduced by the number of rows/columns removed,
c  with the original value of 5 retained in "mpar".  "npar" is then
c  restored to "mpar" in the main program at the beginning of each new
c  secant cycle.
c
c  FORTRAN77 note on column-major storage.
c
c  In linear algebra, we normally think of a matrix with element A(i,j)
c  living in the ith row and jth column.  As we scan from left to right,
c  top to bottom, the second index (j) runs its course before the first
c  index (i) is incremented.
c
c  A11 A12 ... A1n
c  A21 A22 ... A2n
c   :   :       :        (1)
c  An1 An2 ... Ann
c
c  In FORTRAN77, the first index (i) runs its course before the second
c  (j) is incremented and, as such, a matrix is effectively stored as
c  the transpose of how we normally think of it:
c
c  A11 A21 ... An1
c  A12 A22 ... An2
c   :   :       :        (2)
c  A1n A2n ... Ann
c
c  A column of data in FORTRAN77 (indexed by the first index, and thus
c  "column-major") is what we normally think of as a row in a matrix
c  from standard linear algebra ("row-major").
c
c  Now for the Jacobian, we want the lin-alg kth row to have the form:
c
c  d_1f_k  d_2f_k  d_3f_k  d_4f_k  d_5f_k,
c
c  in which case, for an array J(i,j), we need it stacked in a column:
c
c  d_1f_k
c  d_2f_k
c  d_3f_k
c  d_4f_k
c  d_5f_k
c
c  where k is the column number (thus corresponding to the first index;
c  i) and the integers 1:5 are the row numbers (thus corresponding to
c  the second index; j).
c
c  Thus, J(i,j) = d_jf_i; dpsi is indexed by j, fpsi is indexed by i.
c
c  INPUT VARIABLES:
c    fpsi        functional values at current guess of root
c    dpsir       differentials in "psi" set by SETPSIR
c    fpsir       functional values at retarded coordinates
c
c  OUTPUT VARIABLES:
c    jcbn        array of Jacobian elements, J(i,j) ~ df_i/dx_j.
c    soln        solution sentinel set if "jcbn" ends up with all zeros
c                0 => REINIT has reset initial parameters "psi0" to
c                     start a new trial
c               -1 => number of trials exceeded; abort!
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    ISMIN
c    REINIT
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       character*32  frmt
       integer       soln    , i       , eye     , imin    , j
     1             , jay     , jj      , jmin
       real*8        q1      , q2      , dfpsi   , sum
c
       real*8        fpsi    (   *), dpsir   (   *), sumi    (mpar)
     1             , sumj    (mpar)
c
       real*8        fpsir   (mpar,   *), jcbn    (mpar,   *)
c
c      External statements.
c
       integer       ismin
       external      ismin   , reinit
c
c-----------------------------------------------------------------------
c
c      Compute Jacobian elements.  Those dominated by machine round-off
c  error are set to 0.  This is a much more ticklish task than one might
c  at first think.  One needs to check that the absolute and fractional
c  differences from zero are not dominated by round-off, without ever
c  triggering zero-divides.  Not sure if even the following is 100%
c  fool-proof.
c
c  Comment from DAC, 12/21: An example where double-precision values
c  can effectively become single precision and lead to qualitatively
c  different behaviours on different compilers is for small values of
c  "dpsi0" (e.g., 1.0d-6 or less).  Should dpsir(j) be set to "dpsi0"
c  (e.g., when parameters are reinitialised or when a parameter is found
c  to be redundant), "dfpsi" would be the difference of two very nearly
c  equal values and its precision would be reduced by roughly the order
c  of magnitude of "dpsi0".  This can greatly exaggerate the effects of
c  round-off error.
c
c  A case in point is the R&J Fig5b problem with gamma=1.001.  On both
c  stefani (gfortran) and crux (f90), two values of "fpsir(i,j)" were
c  the same to double precision, leading one to expect the corresponding
c  values of "dfpsi" should be the same too.  While this was the case on
c  stefani, it wasn't so on crux (round-off differences occurred in the
c  sixth decimal place) causing LUSOLVE to fail to recognise the
c  Jacobian as singular.  From this point the paths the two computers
c  took to follow the secant solver diverged qualitatively.  Now, in
c  this particular case, both machines ran into dead ends and had to
c  reset the initial guesses, "psi0", and restart the secant solver, at
c  which point their paths reconverged and both ultimately came to the
c  same conclusion.  The point here is that this is a case where round-
c  off error can lead to qualitatively different behaviour, which
c  possibly speaks for a larger value of "dpsi0" to retain as much
c  precision in "dfpsi" as possible.  See however countering arguments
c  in NMLSTS.
c
       do 20 j=1,npar
         do 10 i=1,npar
           dfpsi     = fpsi(i) - fpsir(i,j)
           q1        = max ( abs( fpsi(i) ), abs( fpsir(i,j) ) )
           q2        = 0.0d0
           if (q1       .gt. zero ) q2 = abs ( dfpsi ) / q1
           jcbn(i,j) = 0.0d0
           if (dpsir(j) .ne. 0.0d0) jcbn(i,j) = dfpsi / dpsir(j)
           if (q2       .le. zero ) jcbn(i,j) = 0.0d0
10       continue
         keeprow(j) = 1
20     continue
c
c      Verbose output for diagnostics, including Jacobian, f(psi), and
c  f(psir).  Matrices are presented in row-major format (as in linear
c  algebra), and not column-major format (as in FORTRAN arrays).
c
       if (vrbs .eq. 1) then
         write ( iotty, 2010 )
         write ( frmt , 2020 ) npar
         do 30 i=1,npar
           write ( iotty, frmt ) i, ( jcbn(i,j), j=1,npar )
30       continue
         write ( iotty, 2030 )
         write ( frmt , 2020 ) npar+1
         do 40 i=1,npar
           write ( iotty, frmt ) i, fpsi(i), ( fpsir(i,j), j=1,npar )
40       continue
       endif
c
c-----------------------------------------------------------------------
c                     Check for parameter independence
c-----------------------------------------------------------------------
c
c      Should any row of constant "j" (linear algebra columns) of the
c  Jacobian be zero, then no function depends upon psi(j) and this row
c  is eliminated.  The possibility of "near-zeros" is accounted for in
c  loop 20 above with the use of "zero".
c
       jay = 0
       do 80 j=1,npar
         sum = 0.0d0
         do 60 i=1,npar
           sum = sum + abs ( jcbn(i,j) )
60       continue
         if (sum .eq. 0.0d0) then
           keeprow(j) = 0
         else
           jay = jay + 1
           do 70 i=1,npar
             jcbn(i,jay) = jcbn(i,j)
70         continue
         endif
80     continue
c
c      At this point, "jay" is the number of rows retained.  If jay =
c  npar, no rows of zeros were found, and we move on.  If jay < npar,
c  then their difference is the number of zero rows deleted, and we must
c  delete an equal number of columns of constant "i" (linear algebra
c  rows).  We eliminate the column whose sum of absolute elements is a
c  minimum.  Thus, if there are any columns of zeros, this step will
c  eliminate those, and loop 150 below will be unnecessary.
c
       if (jay .lt. npar) then
c
c      ISMIN returns the index of the smallest element of "sumi".
c
90       continue
         do 110 i=1,npar
           sumi(i) = 0.0d0
           do 100 j=1,jay
             sumi(i) = sumi(i) + abs ( jcbn(i,j) )
100        continue
110      continue
         imin = ismin ( npar, sumi, 1 )
c
c      Eliminate column "imin" of "jcbn" and element "imin" of "fpsi" .
c
         eye  = 0
         do 130 i=1,npar
           if (i .ne. imin) then
             eye = eye + 1
             do 120 j=1,jay
               jcbn(eye,j) = jcbn(i,j)
120          continue
             fpsi(eye) = fpsi(i)
           endif
130      continue
c
c      Having eliminated a column, reduce "npar" by 1.  If "jay" is
c  still less than "npar", multiple rows have been eliminated and we
c  must cycle through the column-elimination loop often enough to
c  restore the Jacobian to a square matrix.
c
         npar = npar - 1
         if (jay .lt. npar) go to 90
c
       endif
c
c      Trap in case all rows are zero!
c
       if (npar .eq. 0) go to 210
c
c      At this point, "jcbn" is a square matrix with its dimension,
c  "npar", possibly reduced and no rows (of constant "j") of zeros.
c
c-----------------------------------------------------------------------
c                      Check for converged functions
c-----------------------------------------------------------------------
c
c      Even if the preceding did not find any irrelevant parameters,
c  there still could be functions that have effectively converged.  For
c  these, f_i(psi)=0 and the ith column of the Jacobian (linear algebra
c  row) will be (near) zero.  These and the corresponding functions
c  should be eliminated.
c
       eye = 0
       do 150 i=1,npar
         if (abs(fpsi(i)) .gt. zero) then
           eye = eye + 1
           do 140 j=1,npar
             jcbn(eye,j) = jcbn(i,j)
140        continue
           fpsi(eye) = fpsi(i)
         endif
150    continue
c
c      At this point, "eye" is the number of columns retained.  If eye =
c  npar, no columns of zeros were found, and we're done.  If eye < npar,
c  then their difference is the number of zero columns deleted, and we
c  must delete an equal number of rows of constant "j": (linear algebra
c  columns).  We eliminate the row whose sum of absolute elements is a
c  minimum.
c
       if (eye .lt. npar) then
c
160      continue
         do 180 j=1,npar
           sumj(j) = 0.0d0
           do 170 i=1,eye
             sumj(j) = sumj(j) + abs ( jcbn(i,j) )
170        continue
180      continue
         jmin = ismin ( npar, sumj, 1 )
c
c      Eliminate row "jmin" of "jcbn".  Here, we must keep track of
c  "keeprow", and thus the three flavours of "j":
c
c      j    index running from 1:npar; indexes "keeprow"
c      jj   skips "j"  corresponding to rows already eliminated
c      jay  skips "jj" corresponding to a newly-found row of zeros
c
c  Comment from DAC, 11/21: bug fix.  I'm pretty sure that do-loop 200
c  should go to "mpar", and not "npar".  Going just to "npar" means "jj"
c  may never get to "jmin", in which case the appropriate row would not
c  have its "keeprow" set to zero leaving one more row kept than the
c  counter "npar".  This can cause index "jay" to go to zero in loop 70
c  in MHDRMN.
c
         jay  = 0
         jj   = 0
c        do 200 j=1,npar
         do 200 j=1,mpar
           if (keeprow(j) .eq. 1) then
             jj = jj + 1
             if (jj .eq. jmin) then
               keeprow(j) = 0
             else
               jay = jay + 1
               do 190 i=1,eye
                 jcbn(i,jay) = jcbn(i,jj)
190            continue
             endif
           endif
200      continue
c
c      Having eliminated a row, reduce "npar" by 1.  If "eye" is still
c  less than "npar", multiple columns have been eliminated and we must
c  cycle through the row-elimination loop often enough to restore the
c  Jacobian to a square matrix.
c
         npar = npar - 1
         if (eye .lt. npar) go to 160
c
       endif
c
c      Trap in case all columns are zero!
c
       if (npar .eq. 0) go to 210
c
c      The Jacobian should now be a square npar x npar matrix with no
c  rows or columns of zeros, and with "npar" reduced accordingly.  In
c  principle, none of the retained functions are converged, and they
c  should all depend upon the parameters still in play.  This being the
c  case, LUSOLVE should be able to find a solution so that the remaining
c  parameters may be improved by our next guess.  In short, "jcbn"
c  should now be non-singular.
c
c  Fingers crossed, we exit the subroutine....
c
       return
c
c      Execution gets here only if "jcbn" is all zeros, a condition that
c  can occur only if the secant solver is lost.  Issue warning message,
c  reset initial parameters "psi0", and try again (soln=0) or, if this
c  would exceed the maximum number of trials "mtry", set soln=-1 as a
c  signal to abort.
c
210    continue
       ntry = ntry + 1
       if (ntry .le. mtry) then
         write ( iotty, 2040 )
         call reinit
         soln = 0
       else
         write ( iotty, 2050 )
         soln =-1
       endif
       nwarn = nwarn + 1
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-------------0----+----1----+----2----+----3----+----4----+----5----+--
c
2010   format('DFDPSI  : Jacobian matrix elements, J(i,j) = d_jf_i.',/
     1       ,'   j=',5x,'1',12x,'2',12x,'3',12x,'4',12x,'5')
2020   format('(''i='',i1,1x,1p',i1,'g13.5)')
2030   format('DFDPSI  : f_i(psi) (column 0) and f_i(psir_j) (columns '
     1       ,'1-5)',/
     1       ,'   j=',5x,'0',12x,'1',12x,'2',12x,'3',12x,'4',12x,'5')
2040   format('DFDPSI  : |---WARNING---> Jacobian is all zeros.  '
     1       ,'Resetting initial guess',/
     2       ,'DFDPSI  : ("psi0") and restarting secant solver.')
2050   format('DFDPSI  : |---WARNING---> Jacobian is all zeros.')
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                D F D P S I                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                F O F P S I                \\\\\\\\\\
c
c=======================================================================
c
       subroutine fofpsi ( psi, fpsi, soln )
c
c    dac:mhdrmn.fofpsi <------------------------- sets values for f(psi)
c                                                             june, 2018
c
c    written by: David Clarke
c    modified 1: December 2021, by DAC; introduced "soln" sentinel that
c                is passed through to RAREFACT.
c
c  PURPOSE:  Starting separately from the left and right states, this
c  routine crosses all waves [shocks, rarefactions, and rotational
c  discontinuities (RD)] of strength set by parameters "psi", and finds
c  the values of the primitive variables, (rho, p, vx, vy, vz, by, bz)
c  in each intermediate state.  It then sets the functions fpsi(1:5) to
c  the differences of the quantities p, vx, vy, vz, and bperp across the
c  CD which the secant root finder in the calling routine works to make
c  zero.
c
c  In the case where bx0 = 0, the RDs, slow waves, and CD are degenerate
c  forming a tangential discontinuity (TD) and only the fast parameters,
c  psi(1:2), are improved.  In this case, functions fpsi(1:2) are set to
c  differences of the quantities vx and p + bperp**2/2 across the TD
c  which the secant root finder works to make zero.
c
c  In crossing rarefactions, a 6th order Runge-Kutta multiple first-
c  order ODE solver is used to integrate the right eigenket.  In
c  crossing shocks, the MHD Rankine-Hugoniot jump conditions are used.
c  Crossing an RD is just a matter of rotating "bperp" and "vperp" using
c  simple formulae.
c
c  The solution admits all switch-on and switch-off waves, and excludes
c  intermediate shocks (thereby invoking the "evolutionary condition").
c  As such, the solution to the MHD Riemann problem should be unique.
c
c  INPUT VARIABLES:
c    psi         ket of parameters to be improved; current guess for
c                wave strengths ("psi" or "psir" in calling routine).
c
c  OUTPUT VARIABLES:
c    fpsi        ket of differences in continuous variables across the
c                CD (TD) for given "psi".
c    soln        solution sentinel set by RAREFACT if density/pressure
c                go to zero before fan integration complete
c                0 => RAREFACT has reset initial parameters "psi0" to
c                     start a new trial
c               -1 => number of trials exceeded; abort!
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    SETFAM
c    SHOCK
c    RAREFACT
c    ROTDISC
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       integer       soln
       real*8        bperp4  , bperp5
c
       real*8        psi     (   *), fpsi     (   *)
c
       external      setfam  , shock   , rarefact, rotdisc
c
c-----------------------------------------------------------------------
c
c----- Left state to left side of CD -----------------------------------
c
c   1. State 1 -> state 2
c
       call setfam ( 1, psi(5) )
c
       if (psi(1) .ge. 0.0d0) then
c
c     a) Family 1 wave is a fast shock.  Use Rankine-Hugoniot jump
c        conditions to find downwind state 2.
c
         call shock ( psi(1) )
c
       else
c
c     b) Family 1 wave is a fast rarefaction fan.  Integrate fast
c        eigenket through RF to find downwind state 2.
c
         call rarefact ( psi(1), soln )
         if (soln .le. 0) return
c
       endif
c
c   2. State 2 -> state 3
c
c      Family 2 wave is a rotational discontinuity.  Rotate "bperp" and
c      compute new vperp to find downwind state 3.
c
       call setfam  ( 2, psi(5) )
       call rotdisc (    psi(5) )
c
c   3. State 3 -> state 4
c
       call setfam ( 3, psi(5) )
c
       if (psi(3) .ge. 0.0d0) then
c
c     a) Family 3 wave is a slow shock.  Use Rankine-Hugoniot jump
c        conditions to find downwind state 4.
c
         call shock ( psi(3) )
c
       else
c
c     b) Family 3 wave is a slow rarefaction fan.  Integrate slow
c        eigenket through RF to find downwind state 4.
c
         call rarefact ( psi(3), soln )
         if (soln .le. 0) return
c
       endif
c
c----- Right state to right side of CD ---------------------------------
c
c   4. State 8 -> state 7
c
       call setfam ( 7, psi(5) )
c
       if (psi(2) .ge. 0.0d0) then
c
c     a) Family 7 wave is a fast shock.  Use Rankine-Hugoniot jump
c        conditions to find downwind state 7.
c
         call shock ( psi(2) )
c
       else
c
c     b) Family 7 wave is a fast rarefaction fan.  Integrate fast
c        eigenket through RF to find downwind state 7.
c
         call rarefact ( psi(2), soln )
         if (soln .le. 0) return
c
       endif
c
c   5. State 7 -> state 6
c
c      Family 6 wave is a rotational discontinuity.  Rotate "bperp" and
c      compute new vperp to find downwind state 6.
c
       call setfam  ( 6, psi(5) )
       call rotdisc (    psi(5) )
c
c   6. State 6 -> state 5
c
       call setfam ( 5, psi(5) )
c
       if (psi(4) .ge. 0.0d0) then
c
c     a) Family 5 wave is a slow shock.  Use Rankine-Hugoniot jump
c        conditions to find downwind state 5.
c
         call shock ( psi(4) )
c
       else
c
c     b) Family 5 wave is a slow rarefaction fan.  Integrate slow
c        eigenket from 0 -> xi to find downwind state 5.
c
         call rarefact ( psi(4), soln )
         if (soln .le. 0) return
c
       endif
c
c-----------------------------------------------------------------------
c
c      Now evaluate the functions.
c
       if (bx0 .eq. 0.0d0) then
c
c      For bx0 = 0, the wave between states 4 and 5 is a TD across which
c  the x-component of v and the total pressure should be continuous.
c  Thus, fpsi(1:2) should both be zero.
c
         fpsi(1) = pvar(3,5) - pvar(3,4)
         fpsi(2) = pvar(2,5) - pvar(2,4)
     1           + 0.5d0 * ( pvar(6,5)**2 + pvar(7,5)**2
     1                     - pvar(6,4)**2 - pvar(7,4)**2 )
c
       else
c
c      For bx0 .ne. 0, the wave between states 4 and 5 is a CD across
c  which the thermal pressure, each component of v, and "bperp" should
c  be continuous.  Thus fpsi(1:5) should all be zero.
c
c      If bperp = 0 while vperp points in different directions in the L
c  and R states, there is nothing to twist the velocity vectors together
c  by the CD.  In this case, the CD is really a tangential discontinuity
c  (TD) with a residual shear velocity across it, and the fact that v_y
c  and v_z don't match across the TD should not be counted as "error".
c  To account for this possibility (e.g., for the HD Riemann problem),
c  functions 3 and 4 are initialised to zero, then evaluated only if
c  there is a non-zero "bperp".
c
         bperp4  = sqrt ( pvar(6,4)**2 + pvar(7,4)**2 )
         bperp5  = sqrt ( pvar(6,5)**2 + pvar(7,5)**2 )
         fpsi(1) = pvar(2,5) - pvar(2,4)
         fpsi(2) = pvar(3,5) - pvar(3,4)
         fpsi(3) = 0.0d0
         fpsi(4) = 0.0d0
         if ( (bperpL .gt. zero) .or. (bperpR .gt. zero) ) then
           fpsi(3) = pvar(4,5) - pvar(4,4)
           fpsi(4) = pvar(5,5) - pvar(5,4)
         endif
         fpsi(5) = bperp5 - bperp4
c
       endif
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                F O F P S I                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\        B E G I N   F U N C T I O N        //////////
c    //////////                 I S M I N                 \\\\\\\\\\
c
c=======================================================================
c
       function ismin ( nelt, arr, inc )
c
c    dac:scilib.ismin <------------------ returns index of minimum value
c                                                        september, 1992
c
c    written by: David Clarke
c    modified 1: July, 1999 by Jean-Pierre de Villers; introduced an
c                if-endif structure inside the do-loop, since this is
c                much faster on non-vectorised platforms.  (Note from
c                DAC, 7/05: He's right; 3-4 times faster!)
c
c  PURPOSE:  Returns the index of the minimum value in a real vector.
c  Note that "arr" is a 1-D vector but could be a multi-dimensional
c  array in the calling routine.
c
c  INPUT VARIABLES:
c    nelt        number of elements in vector "arr" to scan.
c    arr         input array
c    inc         scan every inc'th element in "arr".
c
c  OUTPUT VARIABLES:
c    ismin       index of minimum value in "arr".
c
c  EXTERNALS:  [NONE]
c
c-----------------------------------------------------------------------
c
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
       integer       ismin   , nelt    , inc     , i
       real*8        arrmin
c
       real*8        arr     (   *)
c
c-----------------------------------------------------------------------
c
       arrmin = huge
       ismin  = 1
       do 10 i=1,nelt,inc
         if (arr(i) .lt. arrmin) then
           arrmin = arr(i)
           ismin  = i
         endif
10     continue
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\          E N D   F U N C T I O N          //////////
c    //////////                 I S M I N                 \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////               L U S O L V E               \\\\\\\\\\
c
c=======================================================================
c
       subroutine lusolve ( a, b, fact, x, soln )
c
c    dac:mhdrmn.lusolve <----------------------------- L-U decomposition
c                                                         november, 2001
c
c    written by: David Clarke
c    modified 1: June, 2018 by DAC; replaced Devries' flawed row
c                swapping scheme; see header of linalgo.lusolve (old
c                version) for discussion.
c    modified 2: June, 2018 by DAC; adapted for MHDRMN.
c    modified 3: March, 2019 by DAC; use a singular matrix as a sign
c                that the solution is "lost", and that a new set of
c                initial parameters "psi0" needs to be tried.
c    modified 4: November, 2021 by DAC; completed modification 3 by
c                checking if matrix is singular during LU decomposition.
c
c  PURPOSE:  This routine solves the linear set of equations
c
c                          A |x> = |b>                              (1)
c
c  by the method of lower-upper triangular decomposition of the matrix A
c  (from Crout).  Any matrix A can be written as a product of a lower
c  and upper triangular matrix (L and U): A = LU.  In this case,
c
c                        L U |x> = |b>.
c
c  Let:
c                          U |x> = |z>.                             (2)
c
c  Then:
c                          L |z> = |b>.                             (3)
c
c  Equation (3) is easy to solve by "forward substitution".  Once |z> is
c  known, it is used in equation (2), which is then solved for |x> by
c  "backward substitution".
c
c  The trick is then to find L and U, given A, and the portion of this
c  routine that does that is Paul L. DeVries' version of Crout's method,
c  found in his text "A First Course in Computational Physics" (@1994,
c  John Wiley).
c
c  Before LU decomposition can be started, however, any zeros on the
c  diagonal have to be moved off of it by swapping rows.  This algorithm
c  is of my design, and was implemented when I discovered only in 6/18
c  that Devries' method is not general.
c
c  INPUT VARIABLES:
c    a          the npar x npar array to be decomposed
c    b          the npar-ket representing the RHS of the equation
c    fact       factor to multiply b by; useful for negating input ket
c
c  OUTPUT VARIABLES:
c    x          the n-ket of unknowns
c    soln       1 => solution was found
c               0 => no solution was found ["a" singular]; tells calling
c                    routine to restart secant solver
c              -1 => no solution was found, and maximum number of trials
c                    ("mtry") has been reached; give up.
c
c  EXTERNALS:
c    REINIT
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       integer       soln    , i       , j       , k       , eye
       real*8        fact    , prod    , sum
c
       integer       order   (mpar)
       real*8        x       (   *), b       (   *)
c
       real*8        a       (mpar,   *), p       (mpar,mpar)
     1             , pt      (mpar,mpar)
c
c      External statements.
c
       external      reinit
c
c-----------------------------------------------------------------------
c
c      In preparation for LU decomposition, determine if there are any
c  zeros along the diagonal.
c
       prod = 1.0d0
       do 10 i=1,npar
         prod = prod * a(i,i)
10     continue
c
       if (prod .ne. 0.0d0) then
c
c      No element along A as given is zero, and there is no need to swap
c  rows.  Set "order" to the current order of the rows, and move
c  directly to the LU decomposition.
c
         do 20 i=1,npar
           order(i) = i
20       continue
c
       else
c
c      At least one element along the diagonal is zero, and we must find
c  a permutation of the rows that moves all zeros off of it.  See
c  comments in PHYS3210/assignments/assign05/linalg.f for an explanation
c  of the algorithm.
c
c      Define a permutation array, "p", whose (row i, column j) entry
c  tells us if row j (yes, j) of matrix "a" *could* be moved to row i.
c
         do 40 j=1,npar
           do 30 i=1,npar
             p(i,j) = j
             if (a(j,i) .eq. 0) p(i,j) = 0
30         continue
40       continue
c
c      Use "p" to determine which permutation of the rows will move all
c  zeros away from the diagonal.
c
         i = 1
50       continue
c
c      If row counter "i" exceeds "npar", we've found a permutation that
c  works.  Move on to LU decomposition after permuting the rows of "a".
c
         if (i .gt. npar) go to 100
c
c      If we arrive here, we've still not found a complete permutation
c  to rid the diagonal of zeros.  Start by loading a temporary
c  permutation array, "pt".
c
         do 60 j=1,npar
           pt(i,j) = p(i,j)
60       continue
c
c      Set values of "pt" already used in "order" to zero (so they won't
c  be used again).
c
         do 70 eye=1,i-1
           pt(i,order(eye)) = 0
70       continue
80       continue
c
c      Set order for row "i" to first non-zero value in row "i" of "pt".
c
         do 90 j=1,npar
           if (pt(i,j) .ne. 0) then
             order(i) = pt(i,j)
             i = i + 1
             go to 50
           endif
90       continue
c
c      If execution gets here, there were no non-zero values in row "i"
c  of "pt".  The row pointer "i" is diminished by 1 and we check if
c  there are any rows left.
c
         i = i - 1
         if (i .ge. 1) then
c
c      There are still rows to check.  Go back up one row, eliminate the
c  current value of "order" from that row of "pt", and resume search.
c
           pt(i,order(i)) = 0
           go to 80
c
         else
c
c      If execution gets here, no permutation moves all zeros off the
c  diagonal, and there is no solution.  Redirect execution to the bottom
c  of the routine and return.
c
           go to 300
c
         endif
c
100      continue
c
c      Now that a suitable permutation of the rows of "a" has been
c  found, rearrange the matrix accordingly.
c
         do 120 i=1,npar
           do 110 j=1,npar
             pt(i,j) = a(order(i),j)
110        continue
120      continue
         do 140 i=1,npar
           do 130 j=1,npar
             a(i,j) = pt(i,j)
130        continue
140      continue
c
       endif
c
c      Start the LU decomposition.  The original matrix "a" will be
c  overwritten by the elements of L and U as they are determined since
c  those elements of "a" are never needed again.  This saves having to
c  declare two additional arrays L and U.  Note further that "a" can
c  accommodate the non-zero unknown elements (those not equal to 1) of
c  both L and U simultaneously.
c
c  The first row and column are specially treated, as is L(npar,npar).
c
       do 200 k=1,npar-1
c
c      Do a column of L.  The first column of L is already in place
c  since L is overwriting "a".
c
         if (k .ge. 2) then
           do 160 i=k,npar
             sum = a(i,k)
             do 150 j=1,k-1
               sum = sum - a(i,j) * a(j,k)
150          continue
             a(i,k) = sum
160        continue
         endif
c
c      Even though all zeros have been moved off the diagonal, "a" could
c  still be singular if one or more of the rows or columns is linearly
c  dependent upon the others.  In an LU decomposition, this manifests as
c  a diagonal element [a(k,k)] becoming zero, in which case redirect
c  execution to the bottom of the routine and return.
c
         if (abs(a(k,k)) .lt. zero) go to 300
c
c      Now do a row of U.  The first row of U is computed differently
c  from the rest.
c
         if (k .eq. 1) then
           do 170 j=2,npar
             a(1,j) = a(1,j) / a(1,1)
170        continue
         else
           do 190 j=k+1,npar
             sum = a(k,j)
             do 180 i=1,k-1
               sum = sum - a(k,i) * a(i,j)
180          continue
             a(k,j) = sum / a(k,k)
190        continue
         endif
c
200    continue
c
c      Finally, compute the bottom right element of L.
c
       sum = a(npar,npar)
       do 210 j=1,npar-1
         sum = sum - a(npar,j) * a(j,npar)
210    continue
       a(npar,npar) = sum
c
c      If last diagonal element is zero, redirect execution to the
c  bottom of the routine and return.
c
       if (abs(sum) .lt. zero) go to 300
c
c      LU decomposition is now complete, and we can start the solution
c  phase.  Since the equations may have been interchanged, interchange
c  the elements of "b" the same way, putting the result into "x".
c
       do 220 i=1,npar
         x(i) = fact * b(order(i))
220    continue
c
c      Perform forward substitution.  This puts the values of "z" [see
c  equation (2) in the subroutine header] into the array "x".  Note the
c  division by the diagonal element of "a" that we were worried about in
c  the row swapping segment above.
c
       x(1) = x(1) / a(1,1)
       do 240 i=2,npar
         sum = x(i)
         do 230 k=1,i-1
           sum = sum - a(i,k) * x(k)
230      continue
         x(i) = sum / a(i,i)
240    continue
c
c      Perform backward substitution.  This puts the values of "x" [see
c  equation (3) in the subroutine header] into the array "x".
c
       do 260 i=1,npar-1
         sum = x(npar-i)
         do 250 k=npar-i+1,npar
           sum = sum - a(npar-i,k) * x(k)
250      continue
         x(npar-i) = sum
260    continue
c
       return
c
c      Execution gets here only if the matrix "a" is singular, and no
c  solution exists.  In this case, issue a warning message and reset the
c  initial parameters "psi0" with REINIT.  Solution sentinel, "soln", is
c  set to 0 to tell calling routine to launch a new trial.  If number of
c  resets, "ntry", exceeds "mtry", set "soln" to -1 to tell calling
c  routine to give up.
c
300    continue
       ntry = ntry + 1
       if (ntry .le. mtry) then
         write ( iotty, 2010 )
         call reinit
         soln = 0
       else
         write ( iotty, 2020 )
         soln =-1
       endif
       nwarn = nwarn + 1
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-------------0----+----1----+----2----+----3----+----4----+----5----+--
c
2010   format('LUSOLVE : |---WARNING---> Input array has no inverse.  '
     1       ,'Resetting initial guess',/
     2       ,'LUSOLVE : ("psi0") and restarting secant solver.')
2020   format('LUSOLVE : |---WARNING---> Input array has no inverse.')
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////               L U S O L V E               \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                  M U N U                  \\\\\\\\\\
c
c=======================================================================
c
       subroutine munu ( mu, nu )
c
c    dac:mhdrmn.munu <------------------- computes mu and nu for derivs7
c    from dac:fans.munu                                    january, 2019
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE: This routine is called by DERIVS7 and computes the factors
c  "mu" and "nu" for the fast and slow RF.  By strategic use of binomial
c  expansions, "mu" and "nu" are returned with at least nine significant
c  figures, regardless of how small they may be.  The input variables
c  "alphap" and "deltax" are set by the latest call to SPEEDS, which
c  should be immediately before this routine is called.
c
c  INPUT VARIABLES (from common block comvarr):
c    alphap      apsq / cssq
c    deltax      axsq / cssq - 1
c
c  OUTPUT VARIABLES:
c    mu, nu      factors used in DERIVS7
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       real*8        mu      , nu      , nusq    , musq    , dxpap
     1             , eps     , qty
c
c-----------------------------------------------------------------------
c
       if (alphap .eq. 0.0d0) then
c
c      Take care of the alphap = 0 (bperp = 0) case first.
c
c  Programming notes:
c
c  1.  There is no fast RF for alphap = 0 and deltax .ge. 0, and thus
c      execution never gets here in that stead.  Thus, the "else" case
c      is only for slow waves.
c
c  2.  Algebraically, the limiting case for deltax = alphap = 0 (triple
c      umbilic; TU) is musq = nusq = 1/2.  Further, "alphax" ("deltax")
c      necessarily increases across an RF.  Thus, when starting from the
c      TU, musq = 1/2 for the very first step only seeding "bperp" (and
c      "alphap") with a small but not infinitesimal value.  Then, on the
c      second step, since "alphap" is not strictly zero, calculation of
c      "musq" is done by the second "else" portion below which does its
c      best to keep musq ~ 0 but, not being exactly zero, allows "bperp"
c      to grow like a switch-on fan.  As calculated by this program
c      (by setting musq = 1/2 at the TU), "bperp" grows to ~ 1.0d-16.
c
c      Now, as dp0 -> 0, maximum "bperp" for switch-on fans asymptotes
c      to 0 (program FANS10 can follow switch-on fans for "dp0" as small
c      as ~2.5d-95 resulting in a maximum "bperp" of ~2.0d-38).  Thus, a
c      slow fan originating from the TU should be an Euler fan, not a
c      switch-on fan, and the observation that "bperp" grows from the
c      TU to ~1.0d-16 is a result of round-off error.
c
c      A practical way to get around this problem is to initialise
c      musq = 0 at the TU (deltax = 0) rather than 0.5 (as done below),
c      so that no seed value is assigned to "bperp".  Thus, "alphap"
c      stays 0 and "musq" continues to be set to 0 by this upper portion
c      of the routine maintaining bperp = 0 exactly across the RF.
c
         if (deltax .lt. 0.0d0) then
           musq = 1.0d0
           nusq = 0.0d0
         else
           musq = 0.0d0
           nusq = 1.0d0
         endif
c
       else
c
c      For alphap > 0, "musq" and "nusq" can be computed from the full
c  expression most of the time.  The exception is when either expression
c  becomes a difference between 1 and a number very close to 1 resulting
c  in values dominated by round-off errors.
c
c  The quantity monitored is eps = alphap / (deltax + alphap)**2.  If
c  eps < emin, "musq" (for deltax + alphap > 0) or "nusq" (for deltax
c  + alphap < 0) is evaluated by a three-term binomial expansion.  See
c  notes "Limiting behaviour for mu**2 and nu**2", January, 2019.
c
         dxpap = deltax + alphap
         musq  = 0.5d0
     1         * ( 1.0d0 - dxpap / sqrt ( dxpap**2 + 4.0d0 * alphap ) )
         nusq  = 1.0d0 - musq
c
         if (dxpap .ne. 0.0d0) then
           eps = alphap / dxpap**2
           if (eps .lt. emin) then
             qty = eps - 3.0d0 * eps**2 + 10.0d0 * eps**3
             if (dxpap .gt. 0.0d0) musq = qty
             if (dxpap .lt. 0.0d0) nusq = qty
           endif
         endif
c
       endif
c
       mu = sqrt ( musq )
       nu = sqrt ( nusq )
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                  M U N U                  \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                N M L S T S                \\\\\\\\\\
c
c=======================================================================
c
       subroutine nmlsts
c
c    dac:mhdrmn.nmlsts <---------------- reads namelists from input deck
c                                                              may, 2018
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE: Reads input parameters from the single namelist RPCON.
c
c  INPUT VARIABLES:
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    ARCSINE
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       integer       i
       real*8        rhoL    , pL      , vxL     , vyL     , vzL
     1             , byL     , bzL     , bpL     , rhoR    , pR
     1             , vxR     , vyR     , vzR     , byR     , bzR
     2             , bpR
c
       real*8        arcsine
       external      arcsine
       character*32  namelist (   2,1000)
       external      nlsdac01, nlsdac02, nlsdac03, nlsdac04, nlsdac05
     1             , nlsdac06, nlsdac07, nlsdac08, nlsdac09, nlsdac10
     2             , nlsdac11, nlsdac12, nlsdac13, nlsdac14, nlsdac15
     3             , nlsdac16, nlsdac17, nlsdac18, nlsdac19, nlsdac20
     4             , nlsdac21, nlsdac22, nlsdac23, nlsdac24
c
c-----------------------------------------------------------------------
c
c----------------------  RIEMANN PROBLEM CONTROL  ----------------------
c
c  gamma           ratio of specific heats (default 5/3)
c  rhoL, pL        density and pressure in left state
c  vxL             normal velocity component in left state
c  vyL, vzL        transverse velocity components in left state
c  byL, bzL        transverse magnetic field components in left state
c  rhoR, pR        density and pressure in right state
c  vxR             normal velocity component in right state
c  vyR, vzR        transverse velocity components in right state
c  byR, bzR        transverse magnetic field components in right state
c  bx0             normal magnetic field component (same for all eight
c                  states)
c
c  Default is the HD Sod shocktube: (rhoL=pL=3, rhoR=pR=1, rest 0)
c
c  psi0    (   5)  initial guesses for the five independent secant
c                  convergence parameters in order: left FW, right FW,
c                  left SW, right SW, left RD (default: 4*0.1, half the
c                  difference between phases of "bperp" in R, L states)
c
c  There is no right RD parameter since it depends upon the left RD
c  parameter.  The secant convergence process obviates the need for a
c  CD parameter.  The FW parameters are first since for bx0=0, these are
c  the only parameters the secant convergence routine must work on, and
c  it is easier to limit the list of "psi" by setting npar to 2.
c
c  dpsi0           initial perturbation applied to secant parameters.
c
c  Note from DAC, 12/21: After a great deal of experimentation with all
c  test problems and varying gamma from 5/3 to 1.000001, I determined
c  that, for the vast majority of cases, solutions are found most easily
c  for a small---very small---"dpsi0", whence the default of 1.0d-10.
c  Some cases, like R&J Fig5b, will not converge for gamma=1.001 unless
c  "dpsi0" is 1.0d-9 or smaller, and for gamma=1.00001, 1.0d-10 or
c  smaller.  For "gamma" still closer to 1, "dpsi0" needs to be smaller
c  still.  This sensitivity to gamma approaching 1 was not observed for
c  any other problem.
c
c  I can only imagine that for some ticklish problems, a very small
c  "dpsi0" means that any parameter converging before the others can be
c  "held onto" better if it is perturbed by as small a perturbation as
c  possible (with "dpsi0") while the other parameters continue to
c  converge.  This is just a hypothesis at this writing.  I note that at
c  least one other ticklish problem (R&J Fig4b with gamma=1.3) "fails"
c  to converge with dpsi0=1.0d-10 (that is, best error is 2.7e-7 where
c  the tolerance is 1.0e-8, so it's still probably OK), but converges
c  for dpsi0=1.0d-6.  Thus, I have changed "dpsi0" from being a global
c  parameter in *cd riemvar to a namelist parameter here, in case the
c  user wishes to alter its default value.
c
c  For an argument for higher values of "dpsi0", see note in SETFAM.
c
c  tag             character*2 problem tag for output filenames
c  aspect          aspect ratio for plots (default, 0.9)
c  phdr            character*80 plot header in NCAR mark-up language
c                  blank => no title (default)
c  iftr            = 1 => include footer on plot
c                  = 0 => not (default)
c  vrbs            = 1 => verbose output (debugging)
c                  = 0 => not (default)
c
       call nlsdac01 ( namelist, 1000 )
       namelist (1,   1) = 'gamma                        r s'
       namelist (1,   2) = 'rhoL                         r s'
       namelist (1,   3) = 'pL                           r s'
       namelist (1,   4) = 'vxL                          r s'
       namelist (1,   5) = 'vyL                          r s'
       namelist (1,   6) = 'vzL                          r s'
       namelist (1,   7) = 'byL                          r s'
       namelist (1,   8) = 'bzL                          r s'
       namelist (1,   9) = 'rhoR                         r s'
       namelist (1,  10) = 'pR                           r s'
       namelist (1,  11) = 'vxR                          r s'
       namelist (1,  12) = 'vyR                          r s'
       namelist (1,  13) = 'vzR                          r s'
       namelist (1,  14) = 'byR                          r s'
       namelist (1,  15) = 'bzR                          r s'
       namelist (1,  16) = 'bx0                          r s'
       namelist (1,  17) = 'psi0                    mpar r v'
       namelist (1,  18) = 'dpsi0                        r s'
       namelist (1,  19) = 'tag                          c s'
       namelist (1,  20) = 'aspect                       r s'
       namelist (1,  21) = 'phdr                         c s'
       namelist (1,  22) = 'iftr                         i s'
       namelist (1,  23) = 'vrbs                         i s'
c
c      Default values (HD Sod shock tube).
c
       gamma  = 5.0d0 / 3.0d0
       rhoL   = 3.0d0
       pL     = 3.0d0
       vxL    = 0.0d0
       vyL    = 0.0d0
       vzL    = 0.0d0
       byL    = 0.0d0
       bzL    = 0.0d0
       rhoR   = 1.0d0
       pR     = 1.0d0
       vxR    = 0.0d0
       vyR    = 0.0d0
       vzR    = 0.0d0
       byR    = 0.0d0
       bzR    = 0.0d0
       bx0    = 0.0d0
       dpsi0  = 1.0d-10
       tag    = '00'
       aspect = 0.9d0
       phdr   = ''
       iftr   = 0
       vrbs   = 0
       do 10 i=1,mpar
         psi0(i) = huge
10     continue
c
c      Read namelist
c
       call nlsdac02 ( ioin    , 'rpcon   ', namelist, 1000,   23 )
       call nlsdac05 ( gamma   , 'gamma   ', namelist, 1000 )
       call nlsdac05 ( rhoL    , 'rhoL    ', namelist, 1000 )
       call nlsdac05 ( pL      , 'pL      ', namelist, 1000 )
       call nlsdac05 ( vxL     , 'vxL     ', namelist, 1000 )
       call nlsdac05 ( vyL     , 'vyL     ', namelist, 1000 )
       call nlsdac05 ( vzL     , 'vzL     ', namelist, 1000 )
       call nlsdac05 ( byL     , 'byL     ', namelist, 1000 )
       call nlsdac05 ( bzL     , 'bzL     ', namelist, 1000 )
       call nlsdac05 ( rhoR    , 'rhoR    ', namelist, 1000 )
       call nlsdac05 ( pR      , 'pR      ', namelist, 1000 )
       call nlsdac05 ( vxR     , 'vxR     ', namelist, 1000 )
       call nlsdac05 ( vyR     , 'vyR     ', namelist, 1000 )
       call nlsdac05 ( vzR     , 'vzR     ', namelist, 1000 )
       call nlsdac05 ( byR     , 'byR     ', namelist, 1000 )
       call nlsdac05 ( bzR     , 'bzR     ', namelist, 1000 )
       call nlsdac05 ( bx0     , 'bx0     ', namelist, 1000 )
       call nlsdac09 ( psi0    , 'psi0    ', namelist, 1000 )
       call nlsdac05 ( dpsi0   , 'dpsi0   ', namelist, 1000 )
       call nlsdac03 ( tag     , 'tag     ', namelist, 1000 )
       call nlsdac05 ( aspect  , 'aspect  ', namelist, 1000 )
       call nlsdac03 ( phdr    , 'phdr    ', namelist, 1000 )
       call nlsdac04 ( iftr    , 'iftr    ', namelist, 1000 )
       call nlsdac04 ( vrbs    , 'vrbs    ', namelist, 1000 )
c
c      Load left and right states into solution array.
c
       pvar(1,1) = rhoL
       pvar(2,1) = pL
       pvar(3,1) = vxL
       pvar(4,1) = vyL
       pvar(5,1) = vzL
       pvar(6,1) = byL
       pvar(7,1) = bzL
       pvar(1,8) = rhoR
       pvar(2,8) = pR
       pvar(3,8) = vxR
       pvar(4,8) = vyR
       pvar(5,8) = vzR
       pvar(6,8) = byR
       pvar(7,8) = bzR
c
c      Set constants involving "gamma".
c
       gamm2 = gamma - 2.0d0
       gamm1 = gamma - 1.0d0
       gamp1 = gamma + 1.0d0
c
c      Set left and right orientation angles for "bperp".  If "bperp" is
c  zero on one side but not the other, angular data from the side with
c  zero "bperp" are taken from the side with non-zero "bperp".  If both
c  sides have zero "bperp", "chi" is arbitrarily set to 0 on both sides.
c
       bperpL = sqrt ( byL**2 + bzL**2 )
       if (bperpL .gt. zero) then
         coschiL = byL / bperpL
         sinchiL = bzL / bperpL
         chiL    = arcsine ( byL, bzL )
       endif
       bperpR = sqrt ( byR**2 + bzR**2 )
       if (bperpR .gt. zero) then
         coschiR = byR / bperpR
         sinchiR = bzR / bperpR
         chiR    = arcsine ( byR, bzR )
       endif
       if (bperpL .le. zero) then
         if (bperpR .le. zero) then
           coschiR = 1.0d0
           sinchiR = 0.0d0
           chiR    = 0.0d0
         endif
         coschiL = coschiR
         sinchiL = sinchiR
         chiL    = chiR
       else
         if (bperpR .le. zero) then
           coschiR = coschiL
           sinchiR = sinchiL
           chiR    = chiL
         endif
       endif
c
c      Any unset parameters are set to default values.
c
       do 20 i=1,mpar-1
         if (psi0(i) .eq. huge) psi0(i) = 0.1d0
20     continue
       if (psi0(mpar) .eq. huge) psi0(mpar) = 0.5d0 * ( chiR - chiL )
       psi0(mpar) = mod ( psi0(mpar), 2.0d0*pi )
c
c      Set all other global variables.
c
       sgnbx0 = sign ( 1.0d0, bx0 )
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                N M L S T S                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                P L O T I T                \\\\\\\\\\
c
c=======================================================================
c
       subroutine plotit ( x, den, prs, et, vx, vy, vz, by, bz, chi
     1                   , ntot, filename )
c
c    dac:mhdrmn.plotit <------------------------- plots solution profile
c    from ASTR5700/programs/hdriemann/hdrsx.plotit            june, 2018
c
c    written by: David Clarke
c    modified 1: June, 2018 by DAC; adapted to MHDRMN
c
c  PURPOSE:  This subroutine plots nine variables for the MHD Riemann
c  solution to a postscript file mhdrmn##.ps using calls to the library
c  ~dclarke/libs/gfortran/lib1dplot.a.
c
c  INPUT VARIABLES:
c    x           abscissa against which all variables are plotted
c    den         density
c    prs         thermal pressure
c    et          total energy density
c    vx          x-component of velocity
c    vy          y-component of velocity
c    vz          z-component of velocity
c    by          y-component of magnetic field
c    bz          z-component of magnetic field
c    chi         atan(by/bz); orientation angle of "bperp"
c    ntot        total number of points in each plot
c    phdr        title for graph (blank => no title)
c    filename    name of file written to disc for plots
c
c  OUTPUT VARIABLES:
c
c  EXTERNALS:
c    GRFXOPN   |
c    HDRFTR    |
c    GRFX1D    |---from library lib1dplot.a
c    GRFXCLS   |
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       character*(*) filename
       character*64  xlab    , ylab
       integer       ntot
       real          aspect4
c
       real          x       (   *), den     (   *), prs     (   *)
     1             , et      (   *), vx      (   *), vy      (   *)
     1             , vz      (   *), by      (   *), bz      (   *)
     1             , chi     (   *)
c
       external      grfxopn , hdrftr  , grfx1d  , grfxcls
c
c-----------------------------------------------------------------------
c
c      Open graphics file.
c
       call grfxopn ( filename )
c
c      Write header and footer.
c
       call hdrftr ( phdr, iftr, filename )
c
       xlab    = '|F10|x/t'
       aspect4 = real ( aspect )
c
c      Plot density.
c
       ylab = '|F5|r'
       call grfx1d ( x, den, 0, 1, ntot, 1, 1, 0.0, 0.0, 0.0, 0.0
     1             , 2, aspect4, 1, 2, 3, 1, 3, 1, xlab, ylab )
c
c      Plot pressure.
c
       ylab = '|F10|p'
       call grfx1d ( x, prs, 0, 1, ntot, 1, 1, 0.0, 0.0, 0.0, 0.0
     1             , 2, aspect4, 1, 2, 3, 2, 3, 1, xlab, ylab )
c
c      Plot total energy.
c
       ylab = '|F10|e|BF0|T'
       call grfx1d ( x, et, 0, 1, ntot, 1, 1, 0.0, 0.0, 0.0, 0.0
     1             , 2, aspect4, 1, 2, 3, 3, 3, 1, xlab, ylab )
c
c      Plot x-velocity.
c
       ylab = '|F10|v|B|x'
       call grfx1d ( x, vx, 0, 1, ntot, 1, 1, 0.0, 0.0, 0.0, 0.0
     1             , 2, aspect4, 1, 2, 3, 1, 3, 2, xlab, ylab )
c
c      Plot y-velocity.
c
       ylab = '|F10|v|B|y'
       call grfx1d ( x, vy, 0, 1, ntot, 1, 1, 0.0, 0.0, 0.0, 0.0
     1             , 2, aspect4, 1, 2, 3, 2, 3, 2, xlab, ylab )
c
c      Plot z-velocity.
c
       ylab = '|F10|v|B|z'
       call grfx1d ( x, vz, 0, 1, ntot, 1, 1, 0.0, 0.0, 0.0, 0.0
     1             , 2, aspect4, 1, 2, 3, 3, 3, 2, xlab, ylab )
c
c      Plot y-magnetic field.
c
       ylab = '|F10|B|B|y'
       call grfx1d ( x, by, 0, 1, ntot, 1, 1, 0.0, 0.0, 0.0, 0.0
     1             , 2, aspect4, 1, 2, 3, 1, 3, 3, xlab, ylab )
c
c      Plot z-magnetic field.
c
       ylab = '|F10|B|B|z'
       call grfx1d ( x, bz, 0, 1, ntot, 1, 1, 0.0, 0.0, 0.0, 0.0
     1             , 2, aspect4, 1, 2, 3, 2, 3, 3, xlab, ylab )
c
c      Plot orientation angle of Bperp.
c
       ylab = '|F5|c'
       call grfx1d ( x, chi, 0, 1, ntot, 1, 1, 0.0, 0.0, 0.0, 0.0
     1             , 2, aspect4, 1, 2, 3, 3, 3, 3, xlab, ylab )
c
c      Close graphics file.
c
       call grfxcls
       write ( iotty, 2010 ) filename
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-------------0----+----1----+----2----+----3----+----4----+----5----+--
c
2010   format('PLOTIT  : PSPLOT output file ',a,' written to disc.')
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                P L O T I T                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\        B E G I N   F U N C T I O N        //////////
c    //////////                  R A N 2                  \\\\\\\\\\
c
c=======================================================================
c
       function ran2 ( idum )
c
c    whp:zeus3d.ran2 <----------------- "perfect" uniform random deviate
c    from numerical recipes, ed. 2                            june, 2008
c
c    written by: W. Press et al.
c    modified 1: March, 2019 by DAC; incorporated into mhdrmn
c
c  PURPOSE:  This routine returns a "perfect" uniform random deviate
c  between 0.0 and 1.0 (up to 2 x 10**18 calls).  See Numerical Recipes,
c  ed. 2, pg. 272.  It must be initialised with a negative seed and then
c  called with a positive number from then on.
c
c  INPUT VARIABLES:
c    idum        .lt. 0 => initialise sequence depending on absolute
c                          value
c                .ge. 0 => get next random number from sequence
c
c  OUTPUT VARIABLES:
c    ran2        deviate with uniform distribution between 0.0 and 1.0
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c
       integer       im1     , im2     , ia1     , ia2     , iq1
     1             , iq2     , ir1     , ir2     , ntab    , ndiv
       real*8        am      , eps     , rnmx
       parameter   ( im1=2147483563, im2=2147483399, ia1=40014
     1             , ia2=40692, iq1=53668, iq2=52774, ir1=12211
     2             , ir2=3791, ntab=32, ndiv=1+(im1-1)/ntab )
       parameter   ( am=1.0d0/im1, eps=1.2d-13, rnmx=1.0d0-eps )
c
       integer       idum    , idum2   , j       , k       , iy
       real*8        ran2
c
       integer       iv      (ntab)
c
       save          iv      , iy      , idum2
c
       data          idum2   , iv      , iy
     1             / 123456789, ntab*0, 0 /
c
c-----------------------------------------------------------------------
c
c      If sequence is to be initialised, start here ...
c
       if (idum .le. 0) then
         idum  = max (-idum, 1 )
         idum2 = idum
         do 10 j=ntab+8,1,-1
           k    = idum / iq1
           idum = ia1 * ( idum - k * iq1 ) - k * ir1
           if (idum .lt. 0   ) idum  = idum + im1
           if (j    .le. ntab) iv(j) = idum
10       continue
         iy = iv(1)
       endif
c
c      Otherwise, start here ...
c
       k     = idum  / iq1
       idum  = ia1 * ( idum  - k * iq1 ) - k * ir1
       if (idum  .lt. 0) idum  = idum  + im1
       k     = idum2 / iq2
       idum2 = ia2 * ( idum2 - k * iq2 ) - k * ir2
       if (idum2 .lt. 0) idum2 = idum2 + im2
       j     = 1 + iy / ndiv
       iy    = iv(j) - idum2
       iv(j) = idum
       if (iy .lt. 1) iy = iy + im1 - 1
       ran2  = min ( am*iy, rnmx )
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\          E N D   F U N C T I O N          //////////
c    //////////                  R A N 2                  \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              R A R E F A C T              \\\\\\\\\\
c
c=======================================================================
c
       subroutine rarefact ( psi, soln )
c
c    dac:mhdrmn.rarefact <------------- computes rarefaction fan profile
c                                                              may, 2018
c
c    written by: David Clarke
c    modified 1: February 2019, by DAC; aligned with the main program of
c                FANS10
c    modified 2: December 2021, by DAC; introduced "soln" sentinel to
c                tell MHDRMN to restart trial should density or pressure
c                go to zero before fan integration is complete.
c
c  PURPOSE:  This subroutine uses 6th-order Runge-Kutta to integrate the
c  right eigenket of the rarefaction fan (fast or slow) to obtain its
c  profile as a function of a generalised coordinate, "s".  Limits of
c  integration are (0, slim), "where "slim" is determined from "psi".
c
c  This generalised coordinate can be converted to an actual "distance
c  through the rarefaction", x (i.e., a self-similar coordinate suitable
c  for use as the abscissa on a plot showing the rarefaction profile) by
c  the following transformation (done in RIEMOUT):
c
c    x = vx(s) +/- af(s)  for right (+) or left (-) moving fast waves
c    x = vx(s) +/- as(s)  for right (+) or left (-) moving slow waves
c
c  INPUT VARIABLES:
c    psi         secant solver parameter used to set integration depth.
c
c  OUTPUT VARIABLES:
c    soln        solution sentinel indicating a problem integrating RF
c                = 0 => initial parameters "psi0" have been reset so
c                       that MHDRMN can restart secant loop
c                =-1 => number of resets of "psi0" exceeds "mtry"; best
c                       solution provided to RIEMOUT.
c
c  EXTERNALS:
c    RK7
c    REINIT
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       logical       nofrf   , nosrf   , fasym
       integer       soln    , m       , k       , fors    , n
     1             , i
       real*8        psi     , q1      , q2      , ds      , sleft
     1             , slim    , sright  , dshalf  , dptwid  , dpmid
     1             , dpacc   , dpstar  , err     , errn    , frds
     1             , frac
c
       real*8        yleft   (node), ytwid   (node), ymid    (node)
     1             , ystar   (node), yright  (node)
c
c      External statements.
c
       external      rk7     , reinit
c
c-----------------------------------------------------------------------
c
c      The index "m" runs from 1 to 4 (for families 1, 3, 5, 7) to
c  identify four possible rarefaction fans.  Index "k" is for "dpsi".
c  "fors" is a fast/slow toggle.
c
       m = ( family + 1 ) / 2
       k = 1
       if (m .eq. 4) k = 2
       if (m .eq. 2) k = 3
       if (m .eq. 3) k = 4
       fors = 1
       if ( (m .eq. 2) .or. (m .eq. 3) ) fors = 2
c
c      Set upwind state for rarefaction wave ("uwst" set in SETFAM).
c  The array "yrf" collects the rarefaction profiles for this run.
c  Should this iteration be the one to converge on the solution, "yrf"
c  is used to assemble the datafiles to be plotted.
c
       do 10 n=1,node
         yrf(1,n,m) = pvar(n,uwst)
         yleft  (n) = pvar(n,uwst)
10     continue
c
c      To avoid plotting noise, reset signed quantities (components of v
c  and B) to 0 should they be less than "merr".
c
       q1    = merr * sqrt ( gamma * yrf(1,2,m) )
       q2    = q1 / sqrt ( yrf(1,1,m) )
       if (abs(yrf(1,3,m)) .lt. q2) yrf(1,3,m) = 0.0d0
       if (abs(yrf(1,4,m)) .lt. q2) yrf(1,4,m) = 0.0d0
       if (abs(yrf(1,5,m)) .lt. q2) yrf(1,5,m) = 0.0d0
       if (abs(yrf(1,6,m)) .lt. q1) yrf(1,6,m) = 0.0d0
       if (abs(yrf(1,7,m)) .lt. q1) yrf(1,7,m) = 0.0d0
c
c      For bx0 = 0, a slow rarefaction fan (m = 2 or 3) does not exist.
c  For bperp = 0 and ax > cs, a fast RF does not exist.  In either case,
c  just set the DW primitive variables to the UW values, and return to
c  calling routine.
c
       nosrf = (fors .eq. 2) .and. (bx0    .eq. 0.0d0)
       nofrf = (fors .eq. 1) .and. (bperp  .eq. 0.0d0)
     1                       .and. (alphax .ge. 1.0d0)
c
       if ( nosrf .or. nofrf ) then
c
         do 20 n=1,node
           yrf  (2,n,m) = yrf  (1,n,m)
           pvar(n,dwst) = pvar(n,uwst)
20       continue
         irf(m) = 2
c
c      For no fast RF, reset "psi" to 0.  For no slow RF (bx0 = 0),
c  there's no "psi" to reset.
c
         if ( nofrf ) psi = 0.0d0
         return
c
       endif
c
c      Initialise parameters for first Runge-Kutta integration.
c
       i     = 1
       ds    = ds0
       sleft = 0.0d0
       slim  =-psi
       dptot = pvar(2,uwst) - bx0**2 / gamma
       if (abs(dptot) .le. zero) dptot = 0.0d0
c
c------start of adaptive loop-------------------------------------------
c
30     continue
       sright = min ( sleft + ds, slim )
       ds     = sright - sleft
       dshalf = 0.5d0 * ds
c
c      Runge-Kutta integrates "yleft" by a full step to "ytwid", then
c  again by two half steps to "ystar".  Quantities "dp*" are to manage
c  total accumulated difference between pressure and critical pressure,
c  and used by SPEEDS to evaluate "deltax" should the current pressure
c  be too close to the critical pressure (B_x**2/gamma).
c
       call rk7 ( yleft, ytwid, ds    , dptot, dptwid )
       call rk7 ( yleft, ymid , dshalf, dptot, dpmid  )
       dpacc = dptot + dpmid
       call rk7 ( ymid , ystar, dshalf, dpacc, dpstar )
c
c      Evaluate "yright" ("y" at the right side of the interval) using
c  Richardson extrapolation on "ytwid" and "ystar", then estimate the
c  error.
c
       err = 0.0d0
       do 40 n=1,node
         yright(n) = ( 16.0d0 * ystar(n) - ytwid(n) ) / 15.0d0
         errn      = abs ( ( yright(n) - ystar(n) )
     1                   / ( ( yright(n) + ystar(n) ) + tiny ) ) * 2.0d0
         err       = max ( err, errn )
40     continue
c
c      A symptom that a solution is going awry is if the density and/or
c  pressure go to zero before the other variables do.  In this stead, a
c  new trial is initialised (REINIT), "soln" is set to 0, and execution
c  is returned to calling routine.  Should number of reinitialisations
c  ("ntry") exceed "mtry", "soln" is set to -1 telling calling routine
c  to abort execution.
c
       if ( (yright(1) .lt. tiny) .or. (yright(2) .lt. tiny) ) then
         ntry = ntry + 1
         if (ntry .le. mtry) then
           write ( iotty, 2010 ) family
           call reinit
           soln = 0
         else
           write ( iotty, 2020 ) family
           soln =-1
         endif
         nwarn = nwarn + 1
         return
       endif
c
c      If error is too large, repeat iteration with a half step size.
c
       if (err .gt. merr) then
         ds = dshalf
         go to 30
       endif
c
c      For fast fans only, if either or both components of "bperp" have
c  changed sign from the previous step (which analytically, should not
c  happen), the solution has overshot the terminus of a switch-off fan.
c  In this case, use a bisection approach by repeating last iteration
c  with a half step size.  This allows b_y and b_z to approach zero
c  without ever changing sign, and converges within 100 steps or fewer.
c
       if (fors .eq. 1) then
         if ( ( yright(6) * yleft(6) .lt. 0.0d0 ) .or.
     1        ( yright(7) * yleft(7) .lt. 0.0d0 ) ) then
           ds = dshalf
           go to 30
         endif
       endif
c
c      If error is much smaller than needed, increase step size for next
c  time (but don't redo integration).
c
       if (err .lt. 0.1d0*merr) ds = 1.5d0 * ds
c
c      For slow and HD fans, the fan depth ("slim") can, in principle,
c  be infinite as the pressure and density asymptote to zero.  Should a
c  fan be integrated too far, floating point exceptions in SPEEDS can
c  occur when the quantity (alphax + alphap)**2 required for "disc" is
c  > 1.0d+256).  It is also a strong sign that the solver is lost, as no
c  solution should include states with zero pressure and density.
c
c  Thus, declare a fan to have asymptoted (by setting flag "fasym" to
c  true) once changes to quantities other than rho and p are zero to
c  within machine round-off error.  In this case, set slim = sright to
c  exit adaptive loop below, and set psi =-sright.
c
       if ( (fors .eq. 2) .or. (alphax+alphap .eq. 0.0d0) ) then
         fasym = .true.
         do 50 n=3,node
           q1    = abs ( yright(n) - yleft(n) )
           q2    = max ( abs ( yright(n) ), abs ( yleft(n) ) )
           frac  = 0.0d0
           if (q2 .gt. 0.0d0) frac = q1 / q2
           fasym = fasym .and. (frac .lt. zero)
50       continue
         if ( fasym ) then
           slim = sright
           psi  =-sright
         endif
       endif
c
c      If execution arrives here, "yright" has been integrated to within
c  error tolerance and is stored in the array "yrf", which forms part of
c  the profiles of the final solution should this iteration be the last.
c  Signed quantities (components of v and B) are set to zero should they
c  be effectively less than "merr" to avoid plotting noise.  The left-
c  values ("sleft", "yleft") are then shifted to the right-values
c  ("sright", "yright") in preparation for the next iteration.
c
c  Other quantities set here:
c
c    dptot   accumulated deviation from critical pressure (B_x**2/gamma)
c            is updated using the same Richardson extrapolation step as
c            used above to calculate "yright".
c
c    frds    fractional change in generalised coordinate, s, used below
c            to detect whether a fast fan has been saturated.
c
       i = i + 1
       if (i .le. imax) then
c
         do 60 n=1,node
           yrf(i,n,m) = yright(n)
           yleft  (n) = yright(n)
60       continue
         q1    = merr * sqrt ( gamma * yrf(i,2,m) )
         q2    = q1 / sqrt ( yrf(i,1,m) )
         if (abs(yrf(i,3,m)) .lt. q2) yrf(i,3,m) = 0.0d0
         if (abs(yrf(i,4,m)) .lt. q2) yrf(i,4,m) = 0.0d0
         if (abs(yrf(i,5,m)) .lt. q2) yrf(i,5,m) = 0.0d0
         if (abs(yrf(i,6,m)) .lt. q1) yrf(i,6,m) = 0.0d0
         if (abs(yrf(i,7,m)) .lt. q1) yrf(i,7,m) = 0.0d0
c
         dptot = dptot
     1         + ( 16.0d0 * ( dpmid + dpstar ) - dptwid ) / 15.0d0
         frds  = abs ( ( sright - sleft ) / max ( sleft, sright ) )
         sleft = sright
c
c      For fast fans only, we test for "saturation".
c
c  If frds < merr, fan is declared "saturated", which occurs when:
c
c  1.  for alphax > 1 and bperp > 0, fast RF is a switch-off fan, and
c      bperp has fallen essentially to zero; or
c
c  2.  for alphax < 1 and bperp = 0, RF is a fast Euler fan that has
c      reached the triple umbilic (bperp = 0, alphax = 1) through which
c      it cannot pass.
c
c  In either case, set "slim" to "sright" to exit adaptive loop below,
c  and set "psi" to -sright.  Finally, since the downwind "bperp" should
c  now be 0, set it to 0 to avoid propagating round-off noise to the
c  Jacobian and LUSOLVE.
c
         if ( (fors .eq. 1) .and. (frds .lt. merr) ) then
           slim      = sright
           psi       =-sright
           yright(6) = 0.0d0
           yright(7) = 0.0d0
         endif
c
c      If execution gets here and desired limit of integration has not
c  yet been reached, do another R-K integration step.  Otherwise, exit.
c
         if (sleft .lt. slim) go to 30
c
       else
c
c      If execution gets here, maximum number of steps has been
c  exceeded.  Issue warning message and exit.
c
         write ( iotty, 2030 ) family, imax
         i     = imax
         nwarn = nwarn + 1
c
       endif
c
c------end of adaptive loop---------------------------------------------
c      Set the primitive variables on the downwind state and store the
c  number of steps used to determine the RF profiles.
c
       do 70 n=1,node
         pvar(n,dwst) = yright(n)
70     continue
       irf(m) = i
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-------------0----+----1----+----2----+----3----+----4----+----5----+--
c
2010   format('RAREFACT: |---WARNING---> Input value of "psi" forces '
     1       ,'d/p to 0 on ',i1,'-family fan.',/
     2       ,'RAREFACT: Resetting initial guess ("psi0") and '
     3       ,'restarting secant solver.')
2020   format('RAREFACT: |---WARNING---> Input value of "psi" forces '
     1       ,'d/p to 0 on ',i1,'-family fan.')
2030   format('RAREFACT: |---WARNING---> RK7 failed to finish '
     1       ,'integrating through the ',i1,'-family',/
     2       ,'RAREFACT: fan; may wish to recompile with imax > '
     2       ,i6,'.')
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////              R A R E F A C T              \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                R E I N I T                \\\\\\\\\\
c
c=======================================================================
c
       subroutine reinit
c
c    dac:mhdrmn.reinit <----------------------- reinitialises parameters
c                                                            march, 2019
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE:  In the event that the previous set of initial parameters,
c  "psi0", led to a Jacobian with a zero determinant or caused the
c  maximum number of secant iterations to be reached, solution may be
c  lost.  In either case, this routine reinitialises those parameters in
c  the hopes that the null solution was a result of ill-chosen values.
c
c  INPUT VARIABLES:
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    RAN2
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       integer       iseed   , n
c
c      External statements
c
       real*8        ran2
       external      ran2
c
c      Data statements
c
       data          iseed    / -1 /
       save          iseed
c
c-----------------------------------------------------------------------
c
       do 10 n=1,mpar
         psi0(n) = ran2(iseed) - 0.5d0
10     continue
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                R E I N I T                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////               R I E M O U T               \\\\\\\\\\
c
c=======================================================================
c
       subroutine riemout ( psi, err )
c
c    dac:mhdrmn.riemout <-------------------- output for riemann problem
c                                                             june, 2018
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE:  This routine takes care of all the output.  It first
c  reports the levels in each state to both the screen and an ascii
c  file.  It then assembles the data for the converged solution from all
c  states and outputs them in an ascii text file suitable for GNUPLOT,
c  as well as a publication quality plot of nine variables: rho, p, et,
c  vx, vy, vz, By, Bz, and chi, the orientation angle of "bperp".
c
c  INPUT VARIABLES:
c    psi         five parameters used by the secant algorithm for
c                convergence; a measure of the strength of each wave
c    err         convergence error
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    SPEEDS
c    ADDONE
c    ARCSINE
c    PLOTIT
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       character*1   ud
       character*13  filename
       integer       i       , j       , n       , itot
       real          vsq4    , bsq4
       real*8        err     , abx0    , vsq     , bsq     , et
     1             , chi     , velL    , velR    , delv    , smll
     1             , xleft   , xright  , deni    , q1      , q2
     1             , xpos    , cssq    , axsq    , apsq    , afsq
c
       character*3   tran    (   7)
       real          xvar    (imax), et4     (imax), chi4    (imax)
       real*8        psi     (   *), eks     (  14)
c
       real          yvar    (imax,node)
c
       integer       addone
       real*8        arcsine
       external      speeds  , addone  , arcsine , plotit
c
c-----------------------------------------------------------------------
c
c      Adjust all signed quantities (e.g., components of v and B) in
c  each state that are effectively smaller than "merr" to zero before
c  reporting them to the monitor.
c
       write ( iotty, 2010 )
       do 10 j=1,8
         q1 = sqrt ( merr * gamma * pvar(2,j) )
         q2 = q1 / sqrt ( pvar(1,j) )
         if (abs(pvar(3,j)) .lt. q2) pvar(3,j) = 0.0d0
         if (abs(pvar(4,j)) .lt. q2) pvar(4,j) = 0.0d0
         if (abs(pvar(5,j)) .lt. q2) pvar(5,j) = 0.0d0
         if (abs(pvar(6,j)) .lt. q1) pvar(6,j) = 0.0d0
         if (abs(pvar(7,j)) .lt. q1) pvar(7,j) = 0.0d0
         write ( iotty, 2020 ) j, (pvar(i,j),i=1,node)
10     continue
       write ( iotty, 2030 ) bx0
       abx0 = abs ( bx0 )
c
c      For plotting purposes, we'll plot the profiles vs. the self-
c  similar coordinate x/t which, for constant speeds are just the
c  velocities.  One can think of x/t as just x when t=1.
c
c      First, set the domain of the solution, xleft:xright, where SPEEDS
c  is called just to set "af".
c
c      On the left side, the 1-family wave gets as far as:
c
       call speeds ( pvar(1,1), -1.0d0 )
       if (psi(1) .le. 0.0d0) then
         velL = pvar(3,1) - af
       else
         velL = vshk(1)
       endif
c
c      On the right side, the 7-family wave gets as far as:
c
       call speeds ( pvar(1,8), -1.0d0 )
       if (psi(2) .le. 0.0d0) then
         velR = pvar(3,8) + af
       else
         velR = vshk(4)
       endif
c
c      Thus, take the left and right extrema of the domain to be:
c
       delv   = 0.25d0 * ( velR - velL )
       smll   = 1.0d-5 * delv
       xleft  = velL - delv
       xright = velR + delv
c
c      Fill the abscissa and ordinates.  Start at the left-most point.
c
       i       = 1
       xvar(i) = sngl ( xleft )
       do 20 j=1,node
         yvar(i,j) = sngl ( pvar(j,1) )
20     continue
c
c  1.  Across family 1, left-moving fast wave:
c
       if (psi(1) .lt. 0.0d0) then
c
c      Fill in the 1-family rarefaction.  Here, "af" is calculated
c  within the do-loop (and not by calling SPEEDS) since it is needed at
c  each point within the fan using data gathered for plotting ("yrf").
c  The form for "af" used ensures all radicals are of positive numbers.
c
         do 40 n=1,irf(1)
           i       = addone ( i )
           deni    = 1.0d0 / yrf(n,1,1)
           axsq    = bx0**2 * deni
           apsq    = ( yrf(n,6,1)**2 + yrf(n,7,1)**2 ) * deni
           cssq    = gamma * yrf(n,2,1) * deni
           q1      = axsq + apsq + cssq
           q2      = axsq + apsq - cssq
           afsq    = ( q1 + sqrt ( q2**2 + 4.0d0 * apsq * cssq ) )
     1             * 0.5d0
           af      = sqrt ( afsq )
           xvar(i) = sngl ( yrf(n,3,1) - af )
           do 30 j=1,node
             yvar(i,j) = sngl ( yrf(n,j,1) )
30         continue
           if (n .eq.      1) eks(1) = yrf(n,3,1) - af
           if (n .eq. irf(1)) eks(2) = yrf(n,3,1) - af
40       continue
c
       else
c
c      Fill in the 1-family shock
c
         i       = addone ( i )
         xvar(i) = sngl ( vshk(1) - smll )
         do 50 j=1,node
           yvar(i,j) = sngl ( pvar(j,1) )
50       continue
         eks (1) = vshk(1)
c
         i       = addone ( i )
         xvar(i) = sngl ( vshk(1) + smll )
         do 60 j=1,node
           yvar(i,j) = sngl ( pvar(j,2) )
60       continue
         eks (2) = vshk(1)
c
       endif
c
c  2.  Across family 2, left-moving rotational discontinuity:
c
       xpos    = max ( dble ( xvar(i) ) + 2.0d0 * smll
     1               , pvar(3,2) - abx0 / sqrt ( pvar(1,2) ) )
       i       = addone ( i )
       xvar(i) = sngl ( xpos - smll )
       do 70 j=1,node
         yvar(i,j) = sngl ( pvar(j,2) )
70     continue
       eks (3) = pvar(3,2) - abx0 / sqrt ( pvar(1,2) )
c
       i       = addone ( i )
       xvar(i) = sngl ( xpos + smll )
       do 80 j=1,node
         yvar(i,j) = sngl ( pvar(j,3) )
80     continue
       eks (4) = eks(3)
c
c  3.  Across family 3, left-moving slow wave:
c
       if (psi(3) .lt. 0.0d0) then
c
c      Fill in the 3-family rarefaction.  Here, "as" is calculated
c  within the do-loop (and not by calling SPEEDS) since it is needed at
c  each point within the fan using data gathered for plotting ("yrf").
c  The form for "as" used ensures all radicals are of positive numbers.
c
         do 100 n=1,irf(2)
           i       = addone ( i )
           deni    = 1.0d0 / yrf(n,1,2)
           axsq    = bx0**2 * deni
           apsq    = ( yrf(n,6,2)**2 + yrf(n,7,2)**2 ) * deni
           cssq    = gamma * yrf(n,2,2) * deni
           q1      = axsq + apsq + cssq
           q2      = axsq + apsq - cssq
           afsq    = ( q1 + sqrt ( q2**2 + 4.0d0 * apsq * cssq ) )
     1             * 0.5d0
           as      = sqrt ( cssq * axsq / afsq )
           xvar(i) = sngl ( yrf(n,3,2) - as )
           do 90 j=1,node
             yvar(i,j) = sngl ( yrf(n,j,2) )
90         continue
           if (n .eq.      1) eks(5) = yrf(n,3,2) - as
           if (n .eq. irf(2)) eks(6) = yrf(n,3,2) - as
100      continue
c
       else
c
c      Fill in the 3-family shock
c
         xpos    = max ( dble ( xvar(i) ) + 2.0d0 * smll, vshk(2) )
         i       = addone ( i )
         xvar(i) = sngl ( xpos - smll )
         do 110 j=1,node
           yvar(i,j) = sngl ( pvar(j,3) )
110      continue
         eks (5) = vshk(2)
c
         i       = addone ( i )
         xvar(i) = sngl ( xpos + smll )
         do 120 j=1,node
           yvar(i,j) = sngl ( pvar(j,4) )
120      continue
         eks (6) = vshk(2)
c
       endif
c
c  4.  Across family 4, contact discontinuity:
c
       xpos    = max ( dble ( xvar(i) ) + 2.0d0 * smll, pvar(3,4) )
       i       = addone ( i )
       xvar(i) = sngl ( xpos - smll )
       do 130 j=1,node
         yvar(i,j) = sngl ( pvar(j,4) )
130    continue
       eks (7) = pvar(3,4)
c
       i       = addone ( i )
       xvar(i) = sngl ( xpos + smll )
       do 140 j=1,node
         yvar(i,j) = sngl ( pvar(j,5) )
140    continue
       eks (8) = pvar(3,4)
c
c  5.  Across family 5, right-moving slow wave:
c
       if (psi(4) .lt. 0.0d0) then
c
c      Fill in the 5-family rarefaction.  Note that for right-moving
c  waves, "yrf" was stored backwards by RAREFACT.  That is, data were
c  collected from the upwind to downwind state and thus, for right-
c  moving waves, from right to left.  For plotting purposes, the data
c  are required from left to right, and thus the inversion of the do-
c  loop limits in loop 160 (and 220 for family 7 wave).
c
         do 160 n=irf(3),1,-1
           i       = addone ( i )
           deni    = 1.0d0 / yrf(n,1,3)
           axsq    = bx0**2 * deni
           apsq    = ( yrf(n,6,3)**2 + yrf(n,7,3)**2 ) * deni
           cssq    = gamma * yrf(n,2,3) * deni
           q1      = axsq + apsq + cssq
           q2      = axsq + apsq - cssq
           afsq    = ( q1 + sqrt ( q2**2 + 4.0d0 * apsq * cssq ) )
     1             * 0.5d0
           as      = sqrt ( cssq * axsq / afsq )
           xvar(i) = sngl ( yrf(n,3,3) + as )
           do 150 j=1,node
             yvar(i,j) = sngl ( yrf(n,j,3) )
150        continue
           if (n .eq.      1) eks(10) = yrf(n,3,3) + as
           if (n .eq. irf(3)) eks( 9) = yrf(n,3,3) + as
160      continue
c
       else
c
c      Fill in the 5-family shock
c
         xpos    = max ( dble ( xvar(i) ) + 2.0d0 * smll, vshk(3) )
         i       = addone ( i )
         xvar(i) = sngl ( xpos - smll )
         do 170 j=1,node
           yvar(i,j) = sngl ( pvar(j,5) )
170      continue
         eks( 9) = vshk(3)
c
         i       = addone ( i )
         xvar(i) = sngl ( xpos + smll )
         do 180 j=1,node
           yvar(i,j) = sngl ( pvar(j,6) )
180      continue
         eks(10) = vshk(3)
c
       endif
c
c  6.  Across family 6, right-moving rotational discontinuity:
c
       xpos    = max ( dble ( xvar(i) ) + 2.0d0 * smll
     1               , pvar(3,6) + abx0 / sqrt ( pvar(1,6) ) )
       i       = addone ( i )
       xvar(i) = sngl ( xpos - smll )
       do 190 j=1,node
         yvar(i,j) = sngl ( pvar(j,6) )
190    continue
       eks(11) = pvar(3,6) + abx0 / sqrt ( pvar(1,6) )
c
       i    = addone ( i )
       xvar(i) = sngl ( xpos + smll )
       do 200 j=1,node
         yvar(i,j) = sngl ( pvar(j,7) )
200    continue
       eks(12) = eks(11)
c
c  7.  Across family 7, right-moving fast wave:
c
       if (psi(2) .lt. 0.0d0) then
c
c      Fill in the 7-family rarefaction.
c
         do 220 n=irf(4),1,-1
           i       = addone ( i )
           deni    = 1.0d0 / yrf(n,1,4)
           axsq    = bx0**2 * deni
           apsq    = ( yrf(n,6,4)**2 + yrf(n,7,4)**2 ) * deni
           cssq    = gamma * yrf(n,2,4) * deni
           q1      = axsq + apsq + cssq
           q2      = axsq + apsq - cssq
           afsq    = ( q1 + sqrt ( q2**2 + 4.0d0 * apsq * cssq ) )
     1             * 0.5d0
           af      = sqrt ( afsq )
           xvar(i) = sngl ( yrf(n,3,4) + af )
           do 210 j=1,node
             yvar(i,j) = sngl ( yrf(n,j,4) )
210        continue
           if (n .eq.      1) eks(14) = yrf(n,3,4) + af
           if (n .eq. irf(4)) eks(13) = yrf(n,3,4) + af
220      continue
c
       else
c
c      Fill in the 7-family shock.
c
         xpos    = max ( dble ( xvar(i) ) + 2.0d0 * smll, vshk(4) )
         i       = addone ( i )
         xvar(i) = sngl ( xpos - smll )
         do 230 j=1,node
           yvar(i,j) = sngl ( pvar(j,7) )
230      continue
         eks(13) = vshk(4)
c
         i       = addone ( i )
         xvar(i) = sngl ( xpos + smll )
         do 240 j=1,node
           yvar(i,j) = sngl ( pvar(j,8) )
240      continue
         eks(14) = vshk(4)
c
       endif
c
c      Finish off at the right-most point.
c
       i       = addone ( i )
       xvar(i) = sngl ( xright )
       do 250 j=1,node
         yvar(i,j) = sngl ( pvar(j,8) )
250    continue
c
c      Solution profiles are now complete.  Now compute the secondary
c  profiles: total energy, "etot"; and orientation of "bperp", "chi".
c
       itot = i
       do 260 i=1,itot
         vsq4    = yvar(i,3)**2    + yvar(i,4)**2 + yvar(i,5)**2
         bsq4    = sngl ( bx0**2 ) + yvar(i,6)**2 + yvar(i,7)**2
         et4 (i) = yvar(i,2) / sngl ( gamm1 )
     1           + 0.5 * ( yvar(i,1) * vsq4 + bsq4 )
         chi4(i) = sngl ( arcsine ( dble ( yvar(i,6) )
     1                            , dble ( yvar(i,7) ) ) )
260    continue
c
c      Write datafile "mhdrmn.##.dat".
c
       filename      = "mhdrmn.00.dat"
       filename(8:9) = tag
       open ( unit=ioout, file=filename, status='unknown' )
       do 270 i=1,itot
         write ( ioout, 2040 ) xvar(i), yvar(i,1), yvar(i,2), et4(i)
     1                       , (yvar(i,j), j=3,7), chi4(i)
270    continue
       close ( ioout )
       write ( iotty, 2050 ) filename
c
c      Create PSPLOT file.
c
       filename(11:13) = 'ps '
       call plotit ( xvar, yvar(1,1), yvar(1,2), et4, yvar(1,3)
     1             , yvar(1,4), yvar(1,5), yvar(1,6), yvar(1,7), chi4
     1             , itot, filename )
c
c      Finally, create the "output" data file, which lists the positions
c  and variable values at the foot and head of each transition, along
c  with final values for the secant parameters, "psi".
c
       filename(11:13) = 'out'
       open ( unit=ioout, file=filename, status='unknown' )
       write ( ioout, 2060 )
c
c      Set transition labels.
c
       if (psi(1) .lt. 0.0d0) tran(1) = 'FR-'
       if (psi(1) .gt. 0.0d0) tran(1) = 'FS-'
       if (psi(2) .lt. 0.0d0) tran(7) = 'FR+'
       if (psi(2) .gt. 0.0d0) tran(7) = 'FS+'
       tran(2) = '   '
       tran(3) = '   '
       tran(5) = '   '
       tran(6) = '   '
       if (bx0 .eq. 0.0d0) then
         tran(4) = 'TD '
       else
         if (psi(1) .lt. 0.0d0) tran(1) = 'FR-'
         if (psi(1) .gt. 0.0d0) tran(1) = 'FS-'
         if (psi(5) .ne. 0.0d0) tran(2) = 'RD-'
         if (psi(3) .lt. 0.0d0) tran(3) = 'SR-'
         if (psi(3) .gt. 0.0d0) tran(3) = 'SS-'
         tran(4) = 'CD '
         if (psi(4) .lt. 0.0d0) tran(5) = 'SR+'
         if (psi(4) .gt. 0.0d0) tran(5) = 'SS+'
         if (psi(5) .ne. 0.0d0) tran(6) = 'RD+'
         if (psi(2) .lt. 0.0d0) tran(7) = 'FR+'
         if (psi(2) .gt. 0.0d0) tran(7) = 'FS+'
       endif
c
c      Write two lines for each transition, one on the UW side, the
c  other on the DW side.
c
       do 280 j=1,7
         ud  = 'U'
         if (j .gt. 4) ud = 'D'
         if (j .eq. 4) ud = ' '
         vsq = pvar(3,j  )**2 + pvar(4,j  )**2 + pvar(5,j  )**2
         bsq = bx0**2         + pvar(6,j  )**2 + pvar(7,j  )**2
         et  = pvar(2,j  ) / gamm1 + 0.5d0 * ( pvar(1,j  ) * vsq + bsq )
         chi = arcsine ( pvar(6,j  ), pvar(7,j  ) )
         write ( ioout, 2070 ) j  , ud, tran(j), eks(2*j-1), pvar(1,j  )
     1                       , pvar(2,j  ), et, (pvar(i,j  ),i=3,7), chi
         ud  = 'D'
         if (j .gt. 4) ud = 'U'
         if (j .eq. 4) ud = ' '
         vsq = pvar(3,j+1)**2 + pvar(4,j+1)**2 + pvar(5,j+1)**2
         bsq = bx0**2         + pvar(6,j+1)**2 + pvar(7,j+1)**2
         et  = pvar(2,j+1) / gamm1 + 0.5d0 * ( pvar(1,j+1) * vsq + bsq )
         chi = arcsine ( pvar(6,j+1), pvar(7,j+1) )
         write ( ioout, 2070 ) j+1, ud, '   '  , eks(2*j  ), pvar(1,j+1)
     1                       , pvar(2,j+1), et, (pvar(i,j+1),i=3,7), chi
         write ( ioout, 2080 )
280    continue
c
c      In the output file report, include "bx0", the final values for
c  "psi", and the residual convergence error.
c
       write ( ioout, 2030 ) bx0
       write ( ioout, 2090 ) (psi(i),i=1,npar), err
       close ( ioout )
       write ( iotty, 2100 ) filename
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-------------0----+----1----+----2----+----3----+----4----+----5----+--
c
2010   format('RIEMOUT : Levels in each of the eight states are as '
     1       ,'follows:',//
     1       ,'State |',6x,'d',9x,'p',8x,'vx',8x,'vy',8x,'vz',8x,'by'
     1       ,8x,'bz',/
     1       ,'------|-------------------------------------------------'
     1       ,'----------------------')
2020   format(2x,i1,'   | ',1p7g10.3)
2030   format(/,'bx0 = ',1pg13.5,' across all states.')
2040   format(1p10g13.5)
2050   format(/
     1       ,'RIEMOUT : Ascii  output file ',a13,' written to disc.')
2060   format(' State       x/t      |',7x,'d',12x,'p',11x,'et',11x,'vx'
     1       ,11x,'vy',11x,'vz',11x,'by',11x,'bz',11x,'chi',/
     1       ,'----------------------|---------------------------------'
     1       ,'--------------------------------------------------------'
     2       ,'-----------------------------')
2070   format(i1,1x,a1,1x,a3,1x,1pg13.5' | ',9g13.5)
2080   format(22x,'|')
2090   format(/,'Convergence parameters, psi = ',1p5g15.7,', error = '
     1       ,g10.3)
2100   format('RIEMOUT : Ascii  output file ',a13,' written to disc.')
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////               R I E M O U T               \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                   R K 7                   \\\\\\\\\\
c
c=======================================================================
c
       subroutine rk7 ( yl, yr, h, dpl, delp )
c
c    dac:mhdrmn.rk7 <-------------- performs the Runge-kutta integration
c                                                              may, 2018
c
c    written by: David Clarke
c    modified 1: February 2019, by DAC; added tracking of accumulated
c                deviation from critical pressure
c
c  PURPOSE:  This routine performs a single step of the 6th order Runge-
c  Kutta algorithm for the seven coupled first order ODEs describing the
c  rarefaction fan.  For this application, the independent variable is
c  not needed.
c
c  INPUT VARIABLES:
c    yl         dependent variables at the left side of the interval
c    h          step size
c    dpl        accumulated pressure difference from the critical
c               pressure, "pcr", at left side of interval
c
c  OUTPUT VARIABLES:
c    yr         dependent variables at the right side of the interval
c    delp       amount by which pressure changed across interval which
c               calling routine uses to update accumulated difference
c               between pressure and "pcr".  Used by SPEEDS to calculate
c               "deltax" should "alphax" gets too close to 1.
c
c  LOCAL VARIABLES:
c    fl     |
c    fhtwid |__ temporary 1-D arrays dimensioned to "node" for storing
c    fh     |   the "node" derivatives.
c    f1twid |
c
c  EXTERNALS:
c    DERIVS7
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       integer       n
       real*8        h       , dpl     , delp    , halfh   , dpr
 
       real*8        yl      (node), yr      (node), fl      (node)
     1             , fhtwid  (node), fh      (node), f1twid  (node)
c
       external      derivs7
c
c-----------------------------------------------------------------------
c
       halfh = 0.5d0 * h
c
c      Evaluate derivatives and calculate anew "yr" at each quarter
c  step.  "dpr" is the accumulated pressure difference from the critical
c  pressure, "pcr", up to and including the right side of the quarter
c  step, and is used by the next call to DERIVS7-->SPEEDS to calculate
c  "deltax" should "alphax" gets too close to 1.
c
       call derivs7 ( yl, dpl, fl )
       do 10 n=1,node
         yr(n) = yl(n) + halfh * fl(n)
10     continue
       dpr = dpl + halfh * fl(2)
c
       call derivs7 ( yr, dpr, fhtwid )
       do 20 n=1,node
         yr(n) = yl(n) + halfh * fhtwid(n)
20     continue
       dpr = dpl + halfh * fhtwid(2)
c
       call derivs7 ( yr, dpr, fh )
       do 30 n=1,node
         yr(n) = yl(n) + h * fh(n)
30     continue
       dpr = dpl + h * fh(2)
c
       call derivs7 ( yr, dpr, f1twid )
       do 40 n=1,node
         yr(n) = yl(n) + h * ( fl(n) + 2.0d0 * ( fhtwid(n) + fh(n) )
     1                       + f1twid(n) ) / 6.0d0
40     continue
       delp = h * ( fl(2) + 2.0d0 * ( fhtwid(2) + fh(2) ) + f1twid(2) )
     1      / 6.0d0
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                   R K 7                   \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////               R O T D I S C               \\\\\\\\\\
c
c=======================================================================
c
       subroutine rotdisc ( delchi )
c
c    dac:mhdrmn.rotdisc <---------------------- rotational discontinuity
c                                                              may, 2018
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE: This routine sets the transverse magnetic field and velocity
c  after a rotational discontinuity (RD). No other primitive variables
c  are changed.
c
c  INPUT VARIABLES:
c    delchi      orientation angle of "bperp" downwind of both RDs is
c                chiL + delchi.
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       real*8        delchi  , chi     , srdi    , pm1
c
c-----------------------------------------------------------------------
c
c      Family 2 RD (RD-) rotates "bperp" by angle "delchi", and thus the
c  orientation angle of "bperp" in all states between and including
c  state 3 (downwind of RD-) and state 6 (downwind of RD+) is chiL +
c  delchi.  Thus, RD+ rotates "bperp" by an angle chiL + delchi - chiR
c  (though this angle is never needed in the computation).
c
c  There is no RD if bx0 = 0, in which case vperp and bperp are left
c  alone.
c
       if (bx0 .ne. 0.0d0) then
c
         chi = delchi + chiL
c
c      Set orientation of "bperp" to "chi".  Note that an RD does not
c  change the magnitude of "bperp", so we may use the upwind magnitude
c  of "bperp" for this purpose.
c
          pvar(6,dwst) = bperp * cos ( chi )
          pvar(7,dwst) = bperp * sin ( chi )
c
c      Update "vperp" as a consequence of the change to "bperp".  "pm1"
c  distinguishes between left- and right-moving RDs.  Since downstream
c  vperp depends upon flow speed **relative to RD**, vx = -ax for the
c  right-moving wave, +ax for the left-moving wave.  Thus, for bx0 > 0
c  (and sgnbx0 = +1), pm1 = +1 for family 2 wave, -1 for family 6 wave
c  (reversed for bx0 < 0).
c
         srdi         = 1.0d0 / sqrt ( pvar(1,uwst) )
         pm1          = sgnbx0
         if (family .eq. 6) pm1 =-sgnbx0
         pvar(4,dwst) = pvar(4,uwst)
     1                + pm1 * srdi * ( pvar(6,dwst) - pvar(6,uwst) )
         pvar(5,dwst) = pvar(5,uwst)
     1                + pm1 * srdi * ( pvar(7,dwst) - pvar(7,uwst) )
c
       else
c
c      If "bx0" is zero, all variables are continuous.
c
         pvar(4,dwst) = pvar(4,uwst)
         pvar(5,dwst) = pvar(5,uwst)
         pvar(6,dwst) = pvar(6,uwst)
         pvar(7,dwst) = pvar(7,uwst)
c
       endif
c
c      All other primitive variables remain unchanged after an RD.
c
       pvar(1,dwst) = pvar(1,uwst)
       pvar(2,dwst) = pvar(2,uwst)
       pvar(3,dwst) = pvar(3,uwst)
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////               R O T D I S C               \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                S E T F A M                \\\\\\\\\\
c
c=======================================================================
c
       subroutine setfam ( fam, delchi )
c
c    dac:mhdrmn.setfam <--------------- sets family-dependent parameters
c                                                              may, 2018
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE: This routine sets the family and upwind/downwind state
c  flags, as well as speeds, "bperp", and the orientation of "bperp" in
c  the upwind-state.  All variables set are in common blocks comvari and
c  comvarr.
c
c  INPUT VARIABLES:
c    fam         wave family number, as indicated in the diagram in the
c                header of MHDRMN.
c    delchi      rotation angle by RD-; current guess for psi(5)
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    SPEEDS
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       integer       fam
       real*8        delchi
c
       external      speeds
c
c-----------------------------------------------------------------------
c
       family = fam
       uwst   = family
       dwst   = uwst + 1
       if (family .gt. 4) then
         dwst = family
         uwst = dwst + 1
       endif
c
c      Set the characteristic speeds of the upwind state.
c
       call speeds ( pvar(1,uwst), -1.0d0 )
c
c      Set "bperp" in the upwind state.
c
c  Note from DAC, 12/21: In an earlier version I had commented out the
c  test on "bperp" and made the statement: "Commenting out for now the
c  check (of 'bperp') on 'nearly zero', as there is no such check on
c  'alphap' in SPEEDS."  Why I was worried about there being no
c  commensurate check on "alphap" in SPEEDS I can't recall.  What I do
c  know now is that with the test uncommented, all test problems work
c  fine.  However, with the test commented out, the most ticklish of
c  problems, Fig 4b from R&J, fails in the AZEuS version (but not in
c  MHDRMN) because in the former, pvar(6,1) is -6.661...e-16 and not
c  dead 0 as it is in MHDRMN11.  This gives a (barely) non-zero value
c  for "bperp" below and this has deleterious consequences in RAREFACT
c  where the logical variable "nofrf" is .true. if bperp is dead 0,
c  false otherwise.  Throughout this program, I have tried to maintain
c  the spirit that 0 means "zero", something "close" to 0 in double
c  precision.  It's so easy to write lines of code comparing quantities
c  to certain values without necessarily thinking "How close to 0 does
c  zero really mean?", and I'm still rooting out lines of code that
c  don't comply with this spirit and that turn numerical noise into
c  qualitative consequences.  This is one such case.
c
       bperp = sqrt ( pvar(6,uwst)**2 + pvar(7,uwst)**2 )
       if (bperp .le. zero) bperp = 0.0d0
c
c      Set the unit vector for "bperp" in the upwind state.  "delchi"
c  [psi(5)] is the current best guess for the rotation of "bperp" by the
c  family 2 RD (RD-).
c
       if (family .le. 2) then
         coschi = coschiL
         sinchi = sinchiL
       endif
       if ( (family .eq. 3) .or. (family .eq. 5) ) then
         coschi = cos ( chiL + delchi )
         sinchi = sin ( chiL + delchi )
       endif
       if (family .ge. 6) then
         coschi = coschiR
         sinchi = sinchiR
       endif
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                S E T F A M                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////               S E T P S I R               \\\\\\\\\\
c
c=======================================================================
c
      subroutine setpsir ( psi, dpsir, psir )
c
c    dac:mhdrmn.setpsir <---------------------- sets retarded parameters
c                                                             june, 2018
c
c    written by: David Clarke
c    modified 1: March 2018, by DAC; introduced "dpsir".
c
c  PURPOSE:  This routine sets "psir", the retarded values of "psi", at
c  which f(psi) is evaluated by FOFPSI so that DFDPSI may calculate the
c  npar-gradient of the function "fpsi" needed to compute the Jacobian
c  in DFDPSI.
c
c  INPUT VARIABLES:
c    psi         current guess of parameters (roots)
c    dpsi        differential in parameters from previous guess (via
c                common block rootvarr)
c
c  OUTPUT VARIABLES:
c    dpsir       differentials in "psi" used to set "psir"
c    psir        retarded parameters
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       integer       i       , j
c
       real*8        psi     (   *), dpsir   (   *)
c
       real*8        psir    (mpar,   *)
c
c-----------------------------------------------------------------------
c
       do 20 j=1,npar
         dpsir(j) = dpsi(j)
         do 10 i=1,npar
           psir(i,j) = psi(i)
           if (i .eq. j) psir(i,j) = psi(i) - dpsir(j)
10       continue
20     continue
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////               S E T P S I R               \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                 S H O C K                \\\\\\\\\\
c
c=======================================================================
c
       subroutine shock ( psi )
c
c    dac:mhdrmn.shock <---------- primitive variable jumps across shocks
c                                                              may, 2018
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE: This routine computes the jumps in the primitive variables
c  across a fast or slow shock of strength indicated by "psi" according
c  to the MHD Rankine-Hugoniot jump conditions.
c
c  For the fast shock, a cubic whose coefficients depend upon the square
c  of the upwind Mach number must be solved (using the Cartano-Tartaglia
c  formula; CARTAR2) to find the jump in normal speed *relative to the
c  shock* "phix".  The correct root is the one with the largest real
c  part.
c
c  For the slow shock, a quadratic whose coefficients depend upon the
c  jump in perpendicular magnetic field strength, "bee", must be solved
c  to find "phix", with the smallest root being the physical one (the
c  largest is >1, violating the entropy condition.
c
c  INPUT VARIABLES:
c    psi         secant solver parameter used to set "msq" or "bee".
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    CARTAR2
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       logical       noshock
       integer       fors    , ishk    , k
       real*8        psi     , msq     , bee     , pm1     , alphaxi
     1             , maxsqi  , mapsqi  , hh      , a2      , a1
     1             , a0      , phix    , bm1     , bp1     , eta
     2             , zeta    , vuw     , eps     , xi      , q1
     1             , q2      , beep    , maxsq   , mapsq   , phix1
     1             , phix2
c
c      External statements.
c
       external      cartar2
c
c-----------------------------------------------------------------------
c
c      Identify shock type, fast or slow, and initialise parameters.
c
c  "pm1"   direction of the upwind speed *relative to the shock*
c  "ishk"  index for the shock speed variable "vshk"
c  "k"     index for "psi" and "dpsi".
c
       fors = 1
       if ( (family .eq. 3) .or. (family .eq. 5) ) fors = 2
       pm1  = 1.0d0
       if (family .gt. 4 ) pm1 =-1.0d0
       ishk = ( family + 1 ) / 2
       k    = 1
       if (family .eq. 7) k = 2
       if (family .eq. 3) k = 3
       if (family .eq. 5) k = 4
       beep    = 0.0d0
       alphaxi = huge
       if (alphax .ne. 0.0d0) alphaxi = 1.0d0 / alphax
c
       if (fors .eq. 1) then
c
c----- Fast shock ------------------------------------------------------
c
c      Set inverse square of Alfven Mach numbers, "maxsqi" (relative to
c  "bx0") and "mapsqi" (relative to "bperp").  "msq" is the square of
c  the upwind ordinary Mach number as determined by the secant solver
c  parameter "psi".
c
         msq    = ( af / cs + psi )**2
         maxsqi = alphax / msq
         mapsqi = alphap / msq
c
c      Set "phix" and "bee".
c
         if (psi .eq. 0.0d0) then
           phix = 1.0d0
           bee  = 1.0d0
         else
c
c      Set coefficients of the cubic giving jump in normal velocity,
c  "phix" (equation 5.112, MHD text), and use the Cardano-Tartaglia
c  formula to compute the roots, the largest of which is returned as
c  "phix".  "hh" is the "hydrodynamical term" (defined after equation
c  5.112).
c
           hh = ( gamm1 + 2.0d0 / msq ) / gamp1
           a2 =-hh - 2.0d0 * maxsqi - gamma * mapsqi / gamp1
           a1 = maxsqi * ( 2.0d0 * hh + maxsqi + mapsqi )
     1        + mapsqi * gamm2 / gamp1
           a0 =-maxsqi * ( hh * maxsqi + mapsqi * gamm1 / gamp1 )
c
           call cartar2 ( a0, a1, a2, bperp, phix )
c
c      Compute "bee" (jump in "bperp") from a NaN-proof version of
c  equation 5.103 in the MHD text.  If phix = maxsqi, this is either a
c  "switch-on" or "Euler" shock, and in the former case "bperp" is set
c  by "beep" (ratio of downwind "bperp" to "bx0"; equation 5.117 in MHD
c  text) rather than "bee".  The distinction between switch-on and Euler
c  is made by a test on "msq" (equation 5.120 in the MHD text).
c
c  Note that the else part should not happen if bx0 = 0, in which case
c  the test on "bx0" should be moot.  The only possibility of a bx0 = 0
c  case slipping through to the else part is if "zero" is not small
c  enough.
c
           eps = abs ( phix - maxsqi ) / phix
           if (eps .gt. zero) then
             if (eps .gt. 0.001d0) then
               bee   = ( 1.0d0 - maxsqi ) / ( phix - maxsqi )
             else
               mapsq = huge
               if (alphap .ne. 0.0d0) mapsq = 1.0d0 / mapsqi
               xi    = 0.5d0 * ( 1.0d0 - phix ) / phix
               q1    = gamm1 * xi
               q2    = 2.0d0 * xi * mapsq * gamp1 * ( phix - hh )
               bee   = q1 + sqrt ( ( q1 - 1.0d0 )**2 + q2 )
             endif
           else
             maxsq = huge
             if (bx0 .ne. 0.0d0) maxsq = 1.0d0 / maxsqi
             bee   = huge
             q1    = ( gamp1 * alphax - 2.0d0 ) / gamm1
             if (msq .lt. q1) then
               beep = ( maxsq - 1.0d0 )
     1              * ( gamp1 - 2.0d0 * alphaxi - gamm1 * maxsq )
               beep = sqrt ( beep )
             endif
           endif
         endif
c
       else
c
c----- Slow shock ------------------------------------------------------
c
c      A slow shock does not exist if bx0 = 0 or if bperp = 0 and
c  alphax < 1 (weak "bx0").  In these cases, set DW primitive variables
c  to UW values, and return to calling routine.
c
         noshock = (bperp .eq. 0.0d0) .and. (alphax .lt. 1.0d0)
         if ( (abs(bx0) .le. zero) .or. noshock ) then
c
           pvar(1,dwst) = pvar(1,uwst)
           pvar(2,dwst) = pvar(2,uwst)
           pvar(3,dwst) = pvar(3,uwst)
           pvar(4,dwst) = pvar(4,uwst)
           pvar(5,dwst) = pvar(5,uwst)
           pvar(6,dwst) = pvar(6,uwst)
           pvar(7,dwst) = pvar(7,uwst)
           vshk  (ishk) = pvar(3,uwst)
c
c      For bperp = 0 and alphax < 1, we reset the parameter "psi" to
c  zero; for bx0 = 0, there is no slow parameter to reset.
c
           if ( noshock ) psi = 0.0d0
           return
c
         else
c
           if (bperp .eq. 0.0d0) then
c
c      For alphax > 1, the slow shock is part of the Euler branch, and
c  thus hydrodynamical.  In this case, "psi" is used to set "msq" which
c  must be between 1 and 1 + deltax (alphax), and thus the use of the
c  tanh function.  Setting "bee" to 1 allows the expression for "q1"
c  below (jump in vperp) to be used for slow Euler shocks as well.
c
             msq  = 1.0d0 + tanh ( psi ) * deltax
             phix = ( 2.0d0 / msq + gamm1 ) / gamp1
             bee  = 1.0d0
c
           else
c
c      A slow shock with a non-zero upwind "bperp" is indicated.  The
c  jump in "bperp", "bee", can be between 1 (no shock) and 0 (switch-off
c  shock), and thus psi = 1 - bee must be between 0 (no shock) and 1
c  (switch-off shock).  Thus, we limit psi to 1.
c
             if (psi .ge. 1.0d0) psi = 1.0d0
             bee = 1.0d0 - psi
             bm1 = bee - 1.0d0
             bp1 = bee + 1.0d0
c
c      Set quadratic coefficients giving the jump in "phix" in terms of
c  "bee" (equation 5.131, MHD text, multiplied through by "bm1" to make
c  it NaN-proof).
c
             a2  = 2.0d0 * bee
     1           - bm1 * ( gamp1 * alphax
     2                   + bee * ( gamma + bee ) * alphap )
             a1  =-2.0d0 * bp1
     1           + bm1 * gamma * ( 2.0d0 * alphax + bp1 * alphap )
             a0  = 2.0d0 - bm1 * gamm1 * ( alphax + alphap )
c
c      To satisfy the entropy condition, "phix" must be between 0 and 1.
c  "phix1" and "phix2" are the two roots and while I've not proven it
c  algebraically, it looks like the smaller of the two roots is the
c  physical one.
c
c  Comment from DAC, 11/21: bug fix in case "a2" gets very small, which
c  can happen if bee = 0 and if the solution is sufficiently lost that
c  "alphap" and "alphax", even with an order unity magnetic field, can
c  go to zero if cs**2 gets ridiculously large.
c
             if (a2 .lt. zero) then
               phix =-a0 / a1
             else
               q1    = sqrt ( a1**2 - 4.0d0 * a2 * a0 )
               phix1 = 0.5d0 * (-a1 + q1 ) / a2
               phix2 = 0.5d0 * (-a1 - q1 ) / a2
               phix  = min ( phix1, phix2 )
             endif
             phix  = max ( 0.0d0, min ( phix, 1.0d0 ) )
c
c      Once "phix" is known, compute "msq" from a variant of equation
c  5.103 in the MHD text.  Note that phix*bee = 1 only when bee = 1
c  (taken care of by the if statement) or when bx0 = 0, under which
c  circumstances this portion of the routine is never called.
c
             if (bee .eq. 1.0d0) then
               msq = ( as / cs )**2
             else
               msq = alphax * ( bm1 ) / ( phix * bee - 1.0d0 )
             endif
c
           endif
c
         endif
c
c-----------------------------------------------------------------------
c
       endif
c
c      With "phix", "bee" (or "beep"), and "msq" now known for all types
c  of shocks, set the primitive variables in the downwind state.
c
c  eta    density jump (equation 5.106)
c  zeta   pressure jump (equation 5.104)
c  vuw    upwind fluid velocity relative to shock (thus, <0 for right-
c         moving shock, >0 for left-moving shock, as set by "pm1")
c  q1     quantity to control change in vperp (equation 5.107)
c
       eta  = 1.0d0 / phix
       vuw  = pm1 * sqrt ( msq ) * cs
       pvar(1,dwst) = pvar(1,uwst) * eta
       pvar(3,dwst) = pvar(3,uwst) - ( 1.0d0 - phix ) * vuw
c
c      The else-condition is for switch-on/Euler fast shocks.  "Regular"
c  fast and slow shocks, as well as slow Euler shocks are taken care of
c  by the if-condition.
c
       if (bee .lt. huge) then
         zeta         = 1.0d0 + gamma * msq * ( 1.0d0 - phix )
     1                - 0.5d0 * gamma * alphap * ( bee**2 - 1.0d0 )
         q1           = bx0 * ( bee - 1.0d0 ) / ( vuw * pvar(1,uwst) )
         pvar(2,dwst) = pvar(2,uwst) * zeta
         pvar(4,dwst) = pvar(4,uwst) + q1 * pvar(6,uwst)
         pvar(5,dwst) = pvar(5,uwst) + q1 * pvar(7,uwst)
         pvar(6,dwst) = pvar(6,uwst) * bee
         pvar(7,dwst) = pvar(7,uwst) * bee
       else
         zeta         = 1.0d0 + gamma * msq * ( 1.0d0 - phix )
     1                - 0.5d0 * gamma * alphax * beep**2
         q1           = bx0**2 * beep / ( vuw * pvar(1,uwst) )
         pvar(2,dwst) = pvar(2,uwst) * zeta
         pvar(4,dwst) = pvar(4,uwst) + q1 * coschi
         pvar(5,dwst) = pvar(5,uwst) + q1 * sinchi
         pvar(6,dwst) = bx0 * beep * coschi
         pvar(7,dwst) = bx0 * beep * sinchi
       endif
c
c      Finally, set and store the shock speed relative to the "lab
c  frame".
c
       vshk(ishk) = pvar(3,uwst) - vuw
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-------------0----+----1----+----2----+----3----+----4----+----5----+--
c
2010   format('SHOCK   : No slow shock for bperp = 0 and alphax < 1; '
     1       ,'dpsi(k) = ',1pg13.5)
2020   format('SHOCK   : For slow shock, psi > 1 is reset to 1; '
     1       ,'dpsi(k) = ',1pg13.5)
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                 S H O C K                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                S P E E D S                \\\\\\\\\\
c
c=======================================================================
c
       subroutine speeds ( var, dpacc )
c
c    dac:mhdrmn.speeds <-------------- computes speed-related parameters
c                                                              may, 2018
c
c    written by: David Clarke
c    modified 1: February 2019, by DAC; aligned with version in FANS10
c
c  PURPOSE: This routine computes the speed-related quantities "cs",
c  "af", "as", "alphap", "alphax", and "deltax", some or all of which
c  are needed by the calling routine.  Two considerations are used to
c  ensure values are returned with full machine accuracy:
c
c  1.  To avoid the difference in the slow alpha, "alphas", it is
c      computed as alphax/alphaf.  This requires there to be reliable
c      values for both "alphax" and "alphaf", which there are.
c
c  2.  Since deltax = alphax - 1, it is dominated by round-off error
c      when alphax -> 1.  Thus, if alphax ~ 1, an alternative expression
c      for "deltax" is used based on a binomial expansion on dpacc/pcr,
c      where "dpacc" is the accumulated difference between the current
c      pressure and the critical pressure, pcr = bx0**2/gamma.  See
c      notes "deltax without machine round-off error", January 2019.
c
c  INPUT VARIABLES:
c    var         array dimensioned "node" with seven primitive variable
c                values (rho, p, vx, vy, vz, by, bz); bx0 is in common.
c    dpacc       accumulated difference between pressure and "pcr"
c                =-1.0 => not used.
c
c  OUTPUT VARIABLES: (all in common block comvarr)
c    alphap      apsq / cssq
c    alphax      axsq / cssq
c    deltax      alphax - 1
c    cs          sound speed
c    af          fast speed
c    as          slow speed
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c------------------------ IMPLICIT STATEMENTS --------------------------
c-----------------------------------------------------------------------
c
       implicit      none
c
c-----------------------------------------------------------------------
c----------------------------- PARAMETERS ------------------------------
c-----------------------------------------------------------------------
c
c  mpar    maximum number of free parameters for secant convergence (5)
c  miter   maximum number of iterations for secant convergence
c  mtry    maximum number of trials; each trial started with randomly
c          selected initial parameters ("psi0")
c  node    number of coupled 1st order ODEs solved by Runge-Kutta (7)
c  imax    maximum number of points in Runge-Kutta solutions; also used
c          as array dimension for plot variables.
c  merr    tolerance for R-K solutions and secant convergence.
c  emin    minimum value of "eps" before binomial expansions are used.
c  zero    what is taken as close enough to zero when examining
c          eigenkets for degeneracies/singularities
c  ds0     initial step size
c  tiny    a tiny number
c  huge    a huge number
c  pi      3.14159...
c
c  This routine must be run separately for each set of left and right
c  values.
c
       integer       mpar    , miter   , mtry    , node    , imax
       real*8        merr    , emin    , zero    , ds0     , tiny
     1             , huge    , pi
c
       parameter     ( mpar=5, miter=50, mtry=10, node=7, imax=10000 )
       parameter     ( merr=1.0d-8, emin=1.0d-4, zero=1.0d-13
     1               , ds0=0.001d0, tiny=1.0d-99, huge=1.0d99
     1               , pi=3.141592653589793116d0 )
c
c-----------------------------------------------------------------------
c--------------------------- ROOT VARIABLES ----------------------------
c-----------------------------------------------------------------------
c
       character*2   tag
       character*80  phdr
c
       integer       iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
c
       real*8        gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
c
       integer       irf     (   4), bestirf (   4), keeprow (mpar)
       real*8        psi0    (mpar), dpsi    (mpar), vshk    (   4)
     1             , bestvshk(   4)
c
       real*8        pvar    (node,   8), bestpvar(node,   8)
c
       real*8        yrf     (imax,node,   4), bestyrf (imax,node,   4)
c
       common / rootvarc /
     1               phdr    , tag
c
       common / rootvari /
     1               iotty   , ioin    , ioout   , ioplt   , nwarn
     1             , family  , uwst    , dwst    , npar    , ntry
     1             , iftr    , vrbs
     1             , irf     , bestirf , keeprow
c
       common / rootvarr /
     1               gamma   , gamm2   , gamm1   , gamp1   , dpsi0
     1             , bx0     , sgnbx0  , bperp   , alphap  , alphax
     1             , deltax  , cs      , af      , as      , coschi
     1             , sinchi  , bperpR  , coschiR , sinchiR , chiR
     1             , bperpL  , coschiL , sinchiL , chiL    , dptot
     1             , aspect
     3             , psi0    , dpsi    , vshk    , bestvshk
     3             , pvar    , bestpvar
     3             , yrf     , bestyrf
c
c
       real*8        dpacc   , eps     , disc    , alphaf  , alphas
c
       real*8        var     (   *)
c
c-----------------------------------------------------------------------
c
c      Compute "alphap" and "alphax", then calculate deltax = alphax - 1
c  directly, or by a three-term binomial expansion of eps = dpacc/pcr
c  should eps < emin.
c
       alphap = ( var(6)**2 + var(7)**2 ) / ( gamma * var(2) )
       alphax = bx0**2 / ( gamma * var(2) )
       if (abs(alphax-1.0d0) .lt. zero) alphax = 1.0d0
       deltax = alphax - 1.0d0
       if ( (bx0 .ne. 0.0d0) .and. (dpacc .ne. -1.0d0) ) then
         eps = dpacc * gamma / bx0**2
         if (abs(eps) .lt. emin) then
           deltax =-eps + eps**2 - eps**3
           alphax = 1.0d0 + deltax
         endif
       endif
c
c      Compute the sound, fast, and slow speeds, "cs", "af", and "as".
c
       disc   = sqrt ( ( deltax + alphap )**2 + 4.0d0 * alphap )
       alphaf = 1.0d0 + 0.5d0 * ( deltax + alphap + disc )
       alphas = alphax / alphaf
       cs     = sqrt ( gamma * var(2) / var(1) )
       af     = sqrt ( alphaf ) * cs
       as     = sqrt ( alphas ) * cs
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                S P E E D S                \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 0 1             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac01 ( nmlst, nmax )
c
c    dac:namelist.nlsdac01 <----------------------- reset namelist array
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine initialises all the elements of the array
c  "nmlst" to 32 character blanks words.
c
c  EXTERNALS:  NONE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*32  blank32
       integer       nmax    , i
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank32 /'                                '/
c
c-----------------------------------------------------------------------
c
       do 10 i=1,nmax
         nmlst(1,i) = blank32
         nmlst(2,i) = blank32
10     continue
c
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 0 1             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 0 2             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac02 ( nlunit, cname, nmlst, nmax, ival )
c
c    dac:namelist.nlsdac02 <----------------- reads input character data
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1: February, 1991 by David Clarke; Fixed various bugs
c                  discovered in the previous 20 months of use.
c                  Character strings of indefinite length can now be
c                  read.
c      Modified 2: January, 1993 by David Clarke; Generalised array
c                  reader so that ar(1,1) is understood as ar(1:1,1:1).
c
c  PURPOSE:  This subroutine reads the input decks originally designed
c  for CFT "namelist" (under CTSS), and stores the character
c  representations of the input data in "nmlst(n,2)".  Meanwhile,
c  "nmlst(n,1)" contains a list of the corresponding variable names,
c  which were set by the calling program.  Other subroutines (called by
c  the calling program) convert the character strings to the correct
c  attribute.
c
c  EXTERNALS:  NLSDAC15 - 23
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*1   char1   , char2   , char3
       character*8   cname   , word8
       character*32  blank32 , word32  , temp32
       character*72  line    , oldline
       integer       nlunit  , nmax    , ival    , nlist   , jbeg
     1             , jend    , jran    , jbegold , jendold , i
     2             , i1      , i2      , ilist   , ilisto  , match
     3             , iqte
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data blank32 /'                                '/
c
c      External statements
c
       external      nlsdac15, nlsdac16, nlsdac17, nlsdac18, nlsdac19
     1             , nlsdac20, nlsdac21, nlsdac22, nlsdac23
c
c-----------------------------------------------------------------------
c
c      Initialise parameters.  Note that all the second elements of the
c  array "nmlst" (nmlst(2,*)) and all but the first "ival" of the first
c  elements (nmlst(1,*)) are reset to 32 character blanks words.  This
c  is to allow for the possibility that the READ statement appears in a
c  loop, in which case, most of "nmlst" must be reset.
c
       do 10 i=1,nmax
         nmlst(2,i) = blank32
10     continue
       if (ival .ge. nmax) return
       do 20 i=ival+1,nmax
         nmlst(1,i) = blank32
20     continue
       nlist = ival
       iqte  = 0
30     continue
c
c      Read line from input deck.
c
       read (nlunit, 1010, end=150, err=160) line
c
c      Error message if column 1 contains anything but a 'c', 'C', or a
c  blank.
c
       char2 = line(1:1)
       if ( (char2 .ne. 'c') .and. (char2 .ne. 'C') .and.
     1      (char2 .ne. ' ') ) call nlsdac15 ( line, 1, 1 )
c
c      Error message if column 2 contains anything but a '$' or a blank
c  (unless column 1 contains a 'c' or a 'C').
c
       char3 = line(2:2)
       if ( (char2 .ne. 'c') .and. (char2 .ne. 'C') .and.
     1      (char3 .ne. '$') .and. (char3 .ne. ' ') )
     2   call nlsdac15 ( line, 2, 2 )
c
c      If column 1 contains a 'c' or a 'C', the line is "commented out",
c  and is echoed on the terminal.
c
       if ( (char2 .eq. 'c') .or. (char2 .eq. 'C') ) then
         write (6, 2010) line
         go to 30
       endif
c
c      If column 2 contains a '$' sentinal, then a new namelist begins.
c  Determine if name of namelist matches "cname", and if it does, read
c  the contents of the namelist.  The next '$' sentinal will close the
c  namelist.
c
       if (char3 .eq. '$') then
         jbeg = 3
         call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
         if (char1 .eq. '$') go to 30
         call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte, word32 )
         word8 = word32(1:8)
         if (word8 .eq. cname) then
           jbeg = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           if (char1 .eq. '$') return
40         continue
           i1 = 0
           i2 = 0
c
c      Read the next specified variable.
c
           call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte
     1                   , word32 )
           do 50 i=1,nmax
             ilist  = i
             temp32 = blank32
             temp32(1:8) = nmlst(1,ilist)(1:8)
             if (temp32 .eq. blank32) go to 60
             if (temp32 .eq.  word32) then
               if (nmlst(2,ilist) .ne. blank32) then
                 temp32 = nmlst(1,ilist)
                 nlist  = nlist + 1
                 ilist  = nlist
                 nmlst(1,ilist)( 1: 8) = temp32( 1: 8)
                 nmlst(1,ilist)(25:32) = temp32(25:32)
               endif
               go to 70
             endif
50         continue
c
c      Error:  variable specified in input deck was not included in
c  "nmlst".
c
60         continue
           call nlsdac15 ( line, jbeg, 3 )
c
c      Specified variable found in "nmlst".  Read the next character.
c  If "char1" .ne. '(' or '=', issue error message.  If "char1" = '(', a
c  portion of an array will be set.  If "char1" = '=', the variable
c  value will be read next.
c
70         continue
           jbeg = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           match = 1
           if (char1 .eq. '(') match = 2
           if (char1 .eq. '=') match = 3
           go to (80, 90, 130) match
c
c      Error: unexpected character.
c
80         continue
           call nlsdac15 ( line, jbeg, 4 )
c
c      Specify portion of vector or array to be set.  The limits are
c  stored in columns 10 through 28  of "nmlst(1,ilist)", and are used by
c  other subroutines to initiate the variable.
c
90         continue
           jbeg = jend + 1
           call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte
     1                   , word32 )
           call nlsdac18 ( line, jbeg, word32 )
           call nlsdac19 ( word32, i1 )
           i2 = i1
           nmlst(1,ilist)(9:12) = word32(1:4)
           jbeg = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           match = 1
           if (char1 .eq. ')') match = 2
           if (char1 .eq. ':') match = 3
           if (char1 .eq. ',') match = 4
           go to (80, 100, 110, 111) match
100        continue
           jbeg = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           if (char1 .ne. '=') go to 80
           go to 130
110        continue
           jbeg = jend + 1
           call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte
     1                   , word32 )
           call nlsdac18 ( line, jbeg, word32 )
           nmlst(1,ilist)(13:16) = word32(1:4)
           jbeg = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           match = 1
           if (char1 .eq. ')') match = 2
           if (char1 .eq. ',') match = 3
           go to (80, 100, 120) match
111        continue
           nmlst(1,ilist)(13:16) = word32(1:4)
120        continue
           jbeg = jend + 1
           call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte
     1                   , word32 )
           call nlsdac18 ( line, jbeg, word32 )
           nmlst(1,ilist)(17:20) = word32(1:4)
           jbeg = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           match = 1
           if (char1 .eq. ':') match = 2
           if (char1 .eq. ')') match = 3
           go to (80, 121, 122) match
121        continue
           jbeg = jend + 1
           call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte
     1                   , word32 )
           call nlsdac18 ( line, jbeg, word32 )
           nmlst(1,ilist)(21:24) = word32(1:4)
           jbeg = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           match = 1
           if (char1 .eq. ')') match = 2
           go to (80, 123) match
122        continue
           nmlst(1,ilist)(21:24) = word32(1:4)
123        continue
           jbeg = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           if (char1 .ne. '=') go to 80
c
c      Read the character representation of the variable value.
c
130        continue
           jbeg    = jend + 1
           call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte
     1                   , word32 )
           jbegold = jbeg
           jendold = jend
           oldline = line
           jbeg    = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           if (char1 .eq. '*') then
             char1 = nmlst(1,ilist)(32:32)
             if (char1 .ne. 'v') call nlsdac15 ( line, jbeg, 8 )
             if (i1 .eq. 0) i1 = 1
             call nlsdac18 ( line, jbeg, word32 )
             call nlsdac19 ( word32, i2 )
             i2 = i1 + i2 - 1
             call nlsdac20 ( i2, word32 )
             nmlst(1,ilist)(13:16) = word32(1:4)
             call nlsdac20 ( i1, word32 )
             nmlst(1,ilist)( 9:12) = word32(1:4)
             jbeg = jend + 1
             call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte
     1                     , word32 )
           else
             jbeg = jbegold
             jend = jendold
             if (line .ne. oldline) then
               backspace (nlunit)
               line = oldline
             endif
           endif
           ilisto = ilist
           char1  = nmlst(1,ilist)(30:30)
           if (char1 .eq. 'l') call nlsdac22 ( line, jbeg, word32 )
           if (char1 .eq. 'i') call nlsdac18 ( line, jbeg, word32 )
           if (char1 .eq. 'r') call nlsdac23 ( line, jbeg, word32 )
           if (char1 .eq. 'c') then
c
c      Single or double quotes must surround each character string.  If
c  a second quote is not found, it will be assumed that the character
c  variable is being assigned a string longer than 30 characters.  In
c  this case, multiple entries are made in "nmlst(1,ilist)" for the
c  character variable, with successive pieces of the assigned value
c  stored in "nmlst(2,ilist)".  Continuation characters (&) are
c  appended to each piece of the input character value to tell the
c  subroutines which set the values of character variables that each
c  entry is to be concatenated, rather than overwriting the previous
c  entry.
c
             char1 = word32(1:1)
             if ( (char1 .ne. '"') .and. (char1 .ne. '''') )
     1         call nlsdac15 ( line, jbeg, 10 )
             temp32           = blank32
             temp32(1:jran-1) = word32(2:jran)
             word32           = temp32
             jran             = jran - 1
140          continue
             char1 = word32(jran:jran)
             if ( (char1 .ne. '"') .and. (char1 .ne. '''') ) then
               char1 = nmlst(1,ilist)(32:32)
               if ( (char1 .eq. 'v') .and. (i1 .eq. 0) ) then
                 nmlst(1,ilist)(9:9) = '1'
                 i1 = 1
               endif
               if (jran .le. 30) then
                 jbeg = - jend - 1
               endif
               if (jran .eq. 31) then
                 jran = 30
                 jbeg = - jend
               endif
               if (jran .eq. 32) then
                 jran = 30
                 jbeg = - jend + 1
               endif
               word32 (31:31) = char(jran)
               word32 (32:32) = '&'
               nmlst(2,ilist) = word32
               word32         = nmlst(1,ilist)
               nlist          = nlist + 1
               ilist          = nlist
               nmlst(1,ilist) = word32
               call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte
     1                       , word32 )
               go to 140
             endif
             temp32           = blank32
             temp32(1:jran-1) = word32(1:jran-1)
             word32           = temp32
           endif
           nmlst(2,ilist) = word32
           ilist          = ilisto
c
c      Make sure next character is a comma or a dollar sign.
c
           jbeg = jend + 1
           call nlsdac17 ( nlunit, line, jbeg, jend, char1 )
           if ( (char1 .ne. ',') .and. (char1 .ne. '$') ) go to 80
           jend = jbeg
c
c      If the variable is a vector, check to see if there are more data
c  (a propos for the syntax: iplot=1,1,1,2,0).
c
           if (nmlst(1,ilist)(32:32) .eq. 'v') then
             if (i1 .eq. 0) then
               nmlst(1,ilist)(9:9) = '1'
               i1 = 1
             endif
             if (i2 .eq. 0) i2 = i1
             if (char1 .eq. '$') return
             jbegold = jbeg
             jendold = jend
             oldline = line
             jbeg    = jend + 1
             call nlsdac16 ( nlunit, line, jbeg, jend, jran, iqte
     1                     , word32 )
             char2 = word32(1:1)
             jbeg  = jend + 1
             call nlsdac17 ( nlunit, line, jbeg, jend, char3 )
             if ( (char3 .eq. ',') .or. (char3 .eq. '$') .or.
     1            (char3 .eq. '*') .or. (char2 .eq. '"') .or.
     2            (char2 .eq. '''') ) then
               word32 = nmlst(1,ilist)
               nlist  = nlist + 1
               ilist  = nlist
               nmlst(1,ilist)( 1: 8) = word32( 1: 8)
               nmlst(1,ilist)(25:32) = word32(25:32)
               i1 = i2 + 1
               i2 = i1
               call nlsdac20 ( i1, word32 )
               nmlst(1,ilist)( 9:12) = word32(1:4)
             endif
             jbeg = jbegold
             jend = jendold
             if (line .ne. oldline) then
               backspace (nlunit)
               line = oldline
             endif
             if ( (char3 .eq. ',') .or. (char3 .eq. '$') .or.
     1            (char3 .eq. '*') .or. (char2 .eq. '"') .or.
     2            (char2 .eq. '''') ) go to 130
           endif
c
c     Go on to next variable in "line".
c
           if (char1 .eq. '$') return
           go to 40
         endif
       endif
c
c      No match for "cname" in current "line".  Read next "line".
c
       go to 30
c
c      Error:  "cname" does not exist, or was out of order and thus
c  skipped over.
c
150    continue
       call nlsdac15 ( cname, 0, 6 )
c
c      Error:  Problem reading input deck.
c
160    continue
       call nlsdac15 ( blank32, 0, 7 )
c
c-----------------------------------------------------------------------
c----------------------- Read format statements ------------------------
c-----------------------------------------------------------------------
c
1010   format(a72)
c
c-----------------------------------------------------------------------
c---------------------- Write format statements ------------------------
c-----------------------------------------------------------------------
c
2010   format(a72)
c
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 0 2             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 0 3             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac03 ( cscalar, cstring, nmlst, nmax )
c
c    dac:namelist.nlsdac03 <--------------------- fills character scalar
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1: February 1991, by David Clarke; rewritten to allow
c                  character strings to be specified in pieces.
c
c  PURPOSE:  This subroutine assigns the input value to a character
c  scalar.
c
c  EXTERNALS:  NONE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*8   cstring , word8   , blank8
       character*32  blank32 , word32
       character*(*) cscalar
       integer       nmax    , i       , ibeg    , mult    , lbeg
     1             , lend    , lran
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank8   / '        ' /
       data          blank32  / '                                ' /
c
c-----------------------------------------------------------------------
c
       mult = 0
       ibeg = 1
       lbeg = 1
10     continue
c
c      Find next entry in "nmlst(1,*)" which matches the name of the
c  variable to be set ("cstring").
c
       do 20 i=ibeg,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. cstring) go to 30
         if (word8 .eq. blank8 ) return
20     continue
       return
30     continue
       word32 = nmlst(2,i)
c
c      If character string is longer than 30 characters (final character
c  is '&'), it has been stored in "nmlst" in pieces.  Thus, set the next
c  piece of "cscalar" to the contents of "nmlst(2,i)", and then go back
c  to read the next piece.
c
       if (word32(32:32) .eq. '&') then
         mult = 1
         lran = ichar ( word32(31:31) )
         lend = lbeg + lran - 1
         cscalar(lbeg:lend) = word32(1:lran)
         lbeg = lend + 1
         ibeg = i + 1
         go to 10
       endif
c
c      If the character string was stored in pieces, "mult" will have
c  been set to 1.  If final character is not a '&', then this is the
c  last piece to read in.  Do not include trailing blanks in "cscalar".
c
c      If "mult" is 0, then the character string is stored in its
c  entirety in "nmlst(2,i)".  Thus, set cscalar trivially.
c
       if (mult .eq. 1) then
         do 60 lran=31,1,-1
           if (word32(lran:lran) .ne. ' ') go to 70
60       continue
70       continue
         lend = lbeg + lran - 1
         cscalar(lbeg:lend) = word32(1:lran)
       else
c
c      See comments from DAC, Jan/08 and Apr/08 in NLSDAC07.
c
         if (word32 .ne. blank32) cscalar = word32
c        cscalar = word32
       endif
c
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 0 3             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 0 4             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac04 ( iscalar, cstring, nmlst, nmax )
c
c    dac:namelist.nlsdac04 <----------------------- fills integer scalar
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine assigns the input value to an integer
c  scalar.
c
c  EXTERNALS:  NLSDAC19
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*8   cstring , word8   , blank8
       character*32  blank32 , word32
       integer       iscalar , nmax    , i
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank8   / '        ' /
       data          blank32  / '                                ' /
c
c      External statements
c
       external      nlsdac19
c
c-----------------------------------------------------------------------
c
       do 10 i=1,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. cstring) go to 20
         if (word8 .eq. blank8 ) return
10     continue
       return
20     continue
       word32 = nmlst(2,i)
       if (word32 .ne. blank32) call nlsdac19 ( word32, iscalar )
c
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 0 4             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 0 5             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac05 ( rscalar, cstring, nmlst, nmax )
c
c    dac:namelist.nlsdac05 <-------------------------- fills real scalar
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine assigns the input value to a real scalar.
c
c  EXTERNALS:  NLSDAC21
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*8   cstring , word8   , blank8
       character*32  blank32 , word32
       integer       nmax    , i
       real*8        rscalar
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank8   / '        ' /
       data          blank32  / '                                ' /
c
c      External statements
c
       external      nlsdac21
c
c-----------------------------------------------------------------------
c
       do 10 i=1,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. cstring) go to 20
         if (word8 .eq. blank8 ) return
10     continue
       return
20     continue
       word32 = nmlst(2,i)
       if (word32 .ne. blank32) call nlsdac21 ( word32, rscalar )
c
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 0 5             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 0 6             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac06 ( lscalar, cstring, nmlst, nmax )
c
c    dac:namelist.nlsdac06 <----------------------- fills logical scalar
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine assigns the input value to a logical
c  scalar.
c
c  EXTERNALS:  NONE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       logical       lscalar
       character*8   cstring , word8   , blank8
       integer       nmax    , i
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank8   / '        ' /
c
c-----------------------------------------------------------------------
c
       do 10 i=1,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. cstring) go to 20
         if (word8 .eq. blank8 ) return
10     continue
       return
20     continue
       word8 = nmlst(2,i)(1:8)
       if (word8 .eq. '.true.  ') lscalar = .true.
       if (word8 .eq. '.t.     ') lscalar = .true.
       if (word8 .eq. '.false. ') lscalar = .false.
       if (word8 .eq. '.f.     ') lscalar = .false.
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 0 6             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 0 7             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac07 ( cvector, cstring, nmlst, nmax )
c
c    dac:namelist.nlsdac07 <--------------------- fills character vector
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1: February 1991, by David Clarke; rewritten to allow
c                  character strings to be specified in pieces.
c
c  PURPOSE:  This subroutine assigns the input values to a character
c  vector.
c
c  EXTERNALS:  NLSDAC19
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*8   cstring , word8   , blank8
       character*12  cstr12  , word12
       character*32  word32  , blank32
       integer       nmax    , mult    , i       , ibeg    , j
     1             , jbeg    , jend    , lbeg    , lend    , lran
c
       character*(*) cvector(*)
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank8   / '        ' /
       data          blank32  / '                                ' /
c
c      External statements
c
       external      nlsdac19
c
c-----------------------------------------------------------------------
c
       ibeg = 1
10     continue
       lbeg = 1
       mult = 0
c
c      Find next entry in "nmlst(1,*)" which matches the name of the
c  variable to be set ("cstring").
c
       do 20 i=ibeg,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. cstring) go to 30
         if (word8 .eq. blank8 ) go to 999
20     continue
30     continue
       cstr12 = nmlst(1,i)(1:12)
c
c      Set range of variable index (jbeg through jend) to be set to
c  what will be stored in "cvector(jbeg)".
c
       word32      = blank32
       word32(1:4) = nmlst(1,i)(9:12)
c
c      Note from DAC, Apr/08: Added conditional go to 120 when jbeg = 0,
c  which will trip the `array bound checking' (-C) in loop 110 below.
c  If jbeg=0, the array isn't being set anyway, so skipping on to line
c  120 should be innocuous.
c
       if (word32 .eq. blank32) go to 120
       call nlsdac19 ( word32, jbeg )
       word32(1:4) = nmlst(1,i)(13:16)
       if (word32 .eq. blank32) then
         jend = jbeg
       else
         call nlsdac19 ( word32, jend )
       endif
c
c      If character string is longer than 30 characters (final character
c  is '&'), it has been stored in "nmlst" in pieces.  Thus, set the next
c  piece of "cvector(jbeg)" to the contents of "nmlst(2,i)", and then
c  read the next piece.
c
40     continue
       word32 = nmlst(2,i)
       if (word32(32:32) .eq. '&') then
         mult = 1
         lran = ichar ( word32(31:31) )
         lend = lbeg + lran - 1
         cvector(jbeg)(lbeg:lend) = word32(1:lran)
         lbeg = lend + 1
         ibeg = i + 1
         do 70 i=ibeg,nmax
           word12 = nmlst(1,i)(1:12)
           if (word12 .eq. cstr12 ) go to 80
70       continue
80       continue
         go to 40
       endif
c
c      If the character string was stored in pieces, "mult" will have
c  been set to 1.  If final character is not a '&', then this is the
c  last piece to read in.  Do not include trailing blanks in
c  "cvector(jbeg)".
c
c      If "mult" is 0, then the character string is stored in its
c  entirety in "nmlst(2,i)".  Thus, set cvector(jbeg) trivially.
c
       if (mult .eq. 1) then
         do 90 lran=31,1,-1
           if (word32(lran:lran) .ne. ' ') go to 100
90       continue
100      continue
         lend = lbeg + lran - 1
         cvector(jbeg)(lbeg:lend) = word32(1:lran)
       else
c
c      Comment from DAC, Jan/08: the line commented out below prevented
c  character strings from being set to blanks in namelists, such as
c  p1ylab=' ' in namelist plt1con so that no y-label is put on the plot
c  (but annotations remain).  I cannot recall at all why "cvector" was
c  protected from being set to blank.
c
c      Answer Apr/08: It's to prevent defaults from being clobbered by
c  blanks if the namelist variable is not set.  The original intent of a
c  blank in "nmlst(2,i)" was to indicate that the variable given in
c  "nmlst(1,i)" is not being set, and therefore the default shouldn't be
c  overwritten with a blank.  This means character variables cannot be
c  set to a blank, which is a problem.
c
c      Comment from DAC, Apr/08: If the variable is not being set,
c  jbeg=0 and this portion of the routine is skipped.  Thus, if
c  execution gets here, the variable is to be set, and there is no
c  default value to protect.  Thus, the original first line is commented
c  out, and the second "unprotected" line is revealed.  This tact should
c  work in NLSDAC11 (character arrays) but won't for NLSDAC03 (character
c  scalars) since the latter have no "jbeg" to set.
c
c        if (word32 .ne. blank32) cvector(jbeg) = word32
         cvector(jbeg) = word32
       endif
c
c      Set desired range of "cvector" to the desired constant.
c
       do 110 j=jbeg,jend
         cvector(j) = cvector(jbeg)
110    continue
c
c      Search for next range of elements in "cvector" to be set.
c
120    continue
       ibeg = i + 1
       if (ibeg .gt. nmax) go to 999
       go to 10
c
999    continue
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 0 7             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 0 8             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac08 ( ivector, cstring, nmlst, nmax )
c
c    dac:namelist.nlsdac08 <----------------------- fills integer vector
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine assigns the input values to an integer
c  vector.
c
c  EXTERNALS:  NLSDAC19
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*8   cstring , word8   , blank8
       character*32  word32  , blank32
       integer       nmax    , i       , ibeg    , j       , jbeg
     1             , jend    , ival
c
       integer       ivector (   *)
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank8  / '        ' /
       data          blank32 / '                                ' /
c
c      External statements
c
       external      nlsdac19
c
c-----------------------------------------------------------------------
c
       ibeg = 1
10     continue
       do 20 i=ibeg,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. blank8) go to 999
         if (word8 .eq. cstring) go to 30
20     continue
30     continue
       word32 = nmlst(2,i)
       if (word32 .eq. blank32) go to 999
       ibeg = i + 1
       if (ibeg .gt. nmax+1) go to 999
       word32 = blank32
       word32(1:4) = nmlst(1,i)(9:12)
       call nlsdac19 ( word32, jbeg )
       word32(1:4) = nmlst(1,i)(13:16)
       if (word32 .eq. blank32) then
         jend = jbeg
       else
         call nlsdac19 ( word32, jend )
       endif
       call nlsdac19 ( nmlst(2,i), ival )
       do 40 j=jbeg,jend
         ivector(j) = ival
40     continue
       go to 10
c
999    continue
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 0 8             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 0 9             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac09 ( rvector, cstring, nmlst, nmax )
c
c    dac:namelist.nlsdac09 <-------------------------- fills real vector
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine assigns the input values to a real vector.
c
c  EXTERNALS:  NLSDAC19, NLSDAC21
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*8   cstring , word8   , blank8
       character*32  word32  , blank32
       integer       nmax    , i       , ibeg    , j       , jbeg
     1             , jend
       real*8        rval
c
       real*8        rvector (   *)
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank8  / '        ' /
       data          blank32 / '                                ' /
c
c      External statements
c
       external      nlsdac19, nlsdac21
c
c-----------------------------------------------------------------------
c
       ibeg = 1
10     continue
       do 20 i=ibeg,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. blank8) go to 999
         if (word8 .eq. cstring) go to 30
20     continue
30     continue
       word32 = nmlst(2,i)
       if (word32 .eq. blank32) go to 999
       ibeg = i + 1
       if (ibeg .gt. nmax+1) go to 999
       word32 = blank32
       word32(1:4) = nmlst(1,i)(9:12)
       call nlsdac19 ( word32, jbeg )
       word32(1:4) = nmlst(1,i)(13:16)
       if (word32 .eq. blank32) then
         jend = jbeg
       else
         call nlsdac19 ( word32, jend )
       endif
       call nlsdac21 ( nmlst(2,i), rval )
       do 40 j=jbeg,jend
         rvector(j) = rval
40     continue
       go to 10
c
999    continue
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 0 9             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 1 0             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac10 ( lvector, cstring, nmlst, nmax )
c
c    dac:namelist.nlsdac10 <----------------------- fills logical vector
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine assigns the input values to a logical
c  vector.
c
c  EXTERNALS:  NLSDAC19
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       logical       lval
       character*8   cstring , word8   , blank8
       character*32  word32  , blank32
       integer       nmax    , i       , ibeg    , j       , jbeg
     1             , jend
c
       logical       lvector (   *)
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank8  / '        ' /
       data          blank32 / '                                ' /
c
c      External statements
c
       external      nlsdac19
c
c-----------------------------------------------------------------------
c
       ibeg = 1
10     continue
       do 20 i=ibeg,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. blank8) go to 999
         if (word8 .eq. cstring) go to 30
20     continue
30     continue
       word32 = nmlst(2,i)
       if (word32 .eq. blank32) go to 999
       ibeg = i + 1
       if (ibeg .gt. nmax+1) go to 999
       word32 = blank32
       word32(1:4) = nmlst(1,i)(9:12)
       call nlsdac19 ( word32, jbeg )
       word32(1:4) = nmlst(1,i)(13:16)
       if (word32 .eq. blank32) then
         jend = jbeg
       else
         call nlsdac19 ( word32, jend )
       endif
       word8 = nmlst(2,i)(1:8)
       if (word8 .eq. '.true.  ') lval = .true.
       if (word8 .eq. '.t.     ') lval = .true.
       if (word8 .eq. '.false. ') lval = .false.
       if (word8 .eq. '.f.     ') lval = .false.
       do 40 j=jbeg,jend
         lvector(j) = lval
40     continue
       go to 10
c
999    continue
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 1 0             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 1 1             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac11 ( carray, cstring, nmlst, nmax, nx )
c
c    dac:namelist.nlsdac11 <------ fills two-dimensional character array
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1: February 1991, by David Clarke; rewritten to allow
c                  character strings to be specified in pieces.
c
c  PURPOSE:  This subroutine assigns the input values to a two-
c  dimensional character array.
c
c  EXTERNALS:  NLSDAC19
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*8   cstring , word8   , blank8
       character*24  cstr24  , word24
       character*32  word32  , blank32
       integer       nmax    , nx      , mult    , i       , ibeg
     1             , j       , jbeg    , jend    , k       , kbeg
     2             , kend    , lbeg    , lend    , lran
c
       character*32  nmlst   (   2,   *)
       character*(*) carray  (  nx,   *)
c
c      Data statements
c
       data          blank8   / '        ' /
       data          blank32  / '                                ' /
c
c      External statements
c
       external      nlsdac19
c
c-----------------------------------------------------------------------
c
       ibeg  = 1
10     continue
       lbeg  = 1
       mult  = 0
c
c      Find next entry in "nmlst(1,*)" which matches the name of the
c  variable to be set ("cstring").
c
       do 20 i=ibeg,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. cstring) go to 30
         if (word8 .eq. blank8 ) go to 999
20     continue
30     continue
       cstr24 = nmlst(1,i)(1:24)
c
c      Set range of variable index (jbeg through jend; kbeg through
c  kend) to be set to what will be stored in "carray(jbeg,kbeg)".
c
c      See comments from DAC, Apr/08 in NLSDAC07.
c
       word32 = blank32
       word32(1:4) = nmlst(1,i)(9:12)
       if (word32 .eq. blank32) go to 130
       call nlsdac19 ( word32, jbeg )
       word32(1:4) = nmlst(1,i)(13:16)
       call nlsdac19 ( word32, jend )
       word32(1:4) = nmlst(1,i)(17:20)
       if (word32 .eq. blank32) go to 130
       call nlsdac19 ( word32, kbeg )
       word32(1:4) = nmlst(1,i)(21:24)
       call nlsdac19 ( word32, kend )
c
c      If character string is longer than 30 characters (final character
c  is '&'), it has been stored in "nmlst" in pieces.  Thus, set the next
c  piece of "carray(jbeg,kbeg)" to the contents of "nmlst(2,i)", and
c  then read the next piece.
c
40     continue
       word32 = nmlst(2,i)
       if (word32(32:32) .eq. '&') then
         mult = 1
         lran = ichar ( word32(31:31) )
         lend = lbeg + lran - 1
         carray(jbeg,kbeg)(lbeg:lend) = word32(1:lran)
         lbeg = lend + 1
         ibeg = i + 1
         do 70 i=ibeg,nmax
           word24 = nmlst(1,i)(1:24)
           if (word24 .eq. cstr24 ) go to 80
70       continue
80       continue
         go to 40
       endif
c
c      If the character string was stored in pieces, "mult" will have
c  been set to 1.  If final character is not a '&', then this is the
c  last piece to read in.  Do not include trailing blanks in
c  "carray(jbeg,kbeg)".
c
c      If "mult" is 0, then the character string is stored in its
c  entirety in "nmlst(2,i)".  Thus, set carray(jbeg,kbeg) trivially.
c
       if (mult .eq. 1) then
         do 90 lran=31,1,-1
           if (word32(lran:lran) .ne. ' ') go to 100
90       continue
100      continue
         lend = lbeg + lran - 1
         carray(jbeg,kbeg)(lbeg:lend) = word32(1:lran)
       else
c
c      See comments from DAC, Jan/08 and Apr/08 in NLSDAC07.
c
c        if (word32 .ne. blank32) carray(jbeg,kbeg) = word32
         carray(jbeg,kbeg) = word32
       endif
c
c      Set desired range of "carray" to the desired constant.
c
       do 120 k=kbeg,kend
         do 110 j=jbeg,jend
           carray(j,k) = carray(jbeg,kbeg)
110      continue
120    continue
c
c      Search for next range of elements in "carray" to be set.
c
130    continue
       ibeg = i + 1
       if (ibeg .gt. nmax) go to 999
       go to 10
c
999    continue
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 1 1             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 1 2             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac12 ( iarray, cstring, nmlst, nmax, nx )
c
c    dac:namelist.nlsdac12 <-------- fills two-dimensional integer array
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine assigns the input values to a two-
c  dimensional integer array.
c
c  EXTERNALS:  NLSDAC19
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*8   cstring , word8   , blank8
       character*32  word32  , blank32
       integer       nmax    , nx      , i       , ibeg    , j
     1             , jbeg    , jend    , k       , kbeg    , kend
     2             , ival
c
       character*32  nmlst   (   2,   *)
       integer       iarray  (  nx,   *)
c
c      Data statements
c
       data          blank8  / '        ' /
       data          blank32 / '                                ' /
c
c      External statements
c
       external      nlsdac19
c
c-----------------------------------------------------------------------
c
       ibeg = 1
10     continue
       do 20 i=ibeg,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. blank8) go to 999
         if (word8 .eq. cstring) go to 30
20     continue
30     continue
       word32 = nmlst(2,i)
       if (word32 .eq. blank32) go to 999
       ibeg = i + 1
       if (ibeg .gt. nmax+1) go to 999
       word32 = blank32
       word32(1:4) = nmlst(1,i)(9:12)
       call nlsdac19 ( word32, jbeg )
       word32(1:4) = nmlst(1,i)(13:16)
       call nlsdac19 (word32, jend )
       word32(1:4) = nmlst(1,i)(17:20)
       call nlsdac19 ( word32, kbeg )
       word32(1:4) = nmlst(1,i)(21:24)
       call nlsdac19 ( word32, kend )
       call nlsdac19 ( nmlst(2,i), ival )
       do 50 k=kbeg,kend
         do 40 j=jbeg,jend
           iarray(j,k) = ival
40       continue
50     continue
       go to 10
c
999    continue
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 1 2             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 1 3             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac13 ( rarray, cstring, nmlst, nmax, nx )
c
c    dac:namelist.nlsdac13 <----------- fills two-dimensional real array
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine assigns the input values to a two-
c  dimensional real array.
c
c  EXTERNALS:  NLSDAC19, NLSDAC21
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*8   cstring , word8   , blank8
       character*32  word32  , blank32
       integer       nmax    , nx      , i       , ibeg    , j
     1             , jbeg    , jend    , k       , kbeg    , kend
       real*8        rval
c
       character*32  nmlst   (   2,   *)
       real*8        rarray  (  nx,   *)
c
c      Data statements
c
       data          blank8  / '        ' /
       data          blank32 / '                                ' /
c
c      External statements
c
       external      nlsdac19, nlsdac21
c
c-----------------------------------------------------------------------
c
       ibeg = 1
10     continue
       do 20 i=ibeg,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. blank8) go to 999
         if (word8 .eq. cstring) go to 30
20     continue
30     continue
       word32 = nmlst(2,i)
       if (word32 .eq. blank32) go to 999
       ibeg = i + 1
       if (ibeg .gt. nmax+1) go to 999
       word32 = blank32
       word32(1:4) = nmlst(1,i)(9:12)
       call nlsdac19 ( word32, jbeg )
       word32(1:4) = nmlst(1,i)(13:16)
       call nlsdac19 ( word32, jend )
       word32(1:4) = nmlst(1,i)(17:20)
       call nlsdac19 ( word32, kbeg )
       word32(1:4) = nmlst(1,i)(21:24)
       call nlsdac19 ( word32, kend )
       call nlsdac21 ( nmlst(2,i), rval )
       do 50 k=kbeg,kend
         do 40 j=jbeg,jend
           rarray(j,k) = rval
40       continue
50     continue
       go to 10
c
999    continue
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 1 3             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 1 4             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac14 ( larray, cstring, nmlst, nmax, nx )
c
c    dac:namelist.nlsdac14 <-------- fills two-dimensional logical array
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine assigns the input values to a two-
c  dimensional logical array.
c
c  EXTERNALS:  NONE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       logical       lval
       character*8   cstring , word8   , blank8
       character*32  word32  , blank32
       integer       nmax    , nx      , i       , ibeg    , j
     1             , jbeg    , jend    , k       , kbeg    , kend
c
       logical       larray  (  nx,   *)
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank8  / '        ' /
       data          blank32 / '                                ' /
c
c-----------------------------------------------------------------------
c
       ibeg = 1
10     continue
       do 20 i=ibeg,nmax
         word8 = nmlst(1,i)(1:8)
         if (word8 .eq. blank8) go to 999
         if (word8 .eq. cstring) go to 30
20     continue
30     continue
       word32 = nmlst(2,i)
       if (word32 .eq. blank32) go to 999
       ibeg = i + 1
       if (ibeg .gt. nmax+1) go to 999
       word32 = blank32
       word32(1:4) = nmlst(1,i)(9:12)
       call nlsdac19 ( word32, jbeg )
       word32(1:4) = nmlst(1,i)(13:16)
       call nlsdac19 ( word32, jend )
       word32(1:4) = nmlst(1,i)(17:20)
       call nlsdac19 ( word32, kbeg )
       word32(1:4) = nmlst(1,i)(21:24)
       call nlsdac19 ( word32, kend )
       word8 = nmlst(2,i)(1:8)
       if (word8 .eq. '.true.  ') lval = .true.
       if (word8 .eq. '.t.     ') lval = .true.
       if (word8 .eq. '.false. ') lval = .false.
       if (word8 .eq. '.f.     ') lval = .false.
       do 50 k=kbeg,kend
         do 40 j=jbeg,jend
           larray(j,k) = lval
40       continue
50     continue
       go to 10
c
999    continue
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 1 4             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 1 5             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac15 ( phrase, ipoint, ierror )
c
c    dac:namelist.nlsdac15 <---------------------- issues error messages
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1: May, 2013 by DAC; corrected a long-standing format
c                  error (0x is not a legal descriptor, at least for
c                  gfortran; who knew!)
c
c  PURPOSE:  This subroutine is called when an error is detected in the
c  namelist by one of the subroutines.  The appropriate error message
c  is issued to the CRT, and execution is terminated.
c
c  EXTERNALS:  NONE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*(*) phrase
       character*9   frmt
       integer       ipoint  , ierror
c
c-----------------------------------------------------------------------
c
       write (   6, 2010)
       write (   6, 2020) phrase
       if (ipoint .eq. 1) write ( 6, 2021 )
       if (ipoint .gt. 1) then
         write (frmt, 2030) ipoint - 1
         write (   6, frmt)
       endif
c
       if (ierror .eq.  1) write (6,2040)
       if (ierror .eq.  2) write (6,2050)
       if (ierror .eq.  3) write (6,2060)
       if (ierror .eq.  4) write (6,2070)
       if (ierror .eq.  5) write (6,2080)
       if (ierror .eq.  6) write (6,2090)
       if (ierror .eq.  7) write (6,2100)
       if (ierror .eq.  8) write (6,2110)
       if (ierror .eq.  9) write (6,2120)
       if (ierror .eq. 10) write (6,2130)
       if (ierror .eq. 11) write (6,2140)
       if (ierror .eq. 12) write (6,2150)
       if (ierror .eq. 13) write (6,2160)
       if (ierror .eq. 14) write (6,2170)
       if (ierror .eq. 15) write (6,2180)
       if (ierror .eq. 16) write (6,2190)
       if (ierror .eq. 17) write (6,2200)
       if (ierror .eq. 18) write (6,2210)
       if (ierror .eq. 19) write (6,2220)
       if (ierror .eq. 20) write (6,2230)
c
       write (6,2010)
       stop
c
c-----------------------------------------------------------------------
c---------------------- Write format statements ------------------------
c-----------------------------------------------------------------------
c
2010   format('************************************'
     1       ,'************************************')
2020   format( (a) )
2021   format('^')
2030   format("(",i2.2,"x,'^')")
2040   format('NAMELIST ERROR  1 ---> column 1 reserved for comment '
     1       ,'sentinal: c')
2050   format('NAMELIST ERROR  2 ---> column 2 reserved for $ '
     1       ,'delimeter')
2060   format('NAMELIST ERROR  3 ---> variable not found in namelist')
2070   format('NAMELIST ERROR  4 ---> unexpected character - check '
     1       ,'syntax')
2080   format('NAMELIST ERROR  5 ---> invalid logical expression')
2090   format('NAMELIST ERROR  6 ---> namelist does not exist, or is '
     1       ,'out of sequence')
2100   format('NAMELIST ERROR  7 ---> error reading input deck')
2110   format('NAMELIST ERROR  8 ---> variable not declared as a '
     1       ,'vector')
2120   format('NAMELIST ERROR  9 ---> next namelist begun before '
     1       ,'closing quote found')
2130   format('NAMELIST ERROR 10 ---> missing opening quote')
2140   format('NAMELIST ERROR 11 ---> blank data')
2150   format('NAMELIST ERROR 12 ---> premature end of file')
2160   format('NAMELIST ERROR 13 ---> missing $ sentinal to close '
     1       ,'namelist')
2170   format('NAMELIST ERROR 14 ---> exponent must not exceed 999')
2180   format('NAMELIST ERROR 15 ---> no more than 15 digits in single '
     1       ,'precision')
2190   format('NAMELIST ERROR 16 ---> ')
2200   format('NAMELIST ERROR 17 ---> ')
2210   format('NAMELIST ERROR 18 ---> ')
2220   format('NAMELIST ERROR 19 ---> ')
2230   format('NAMELIST ERROR 20 ---> ')
c
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 1 5             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              N L S D A C 1 6              \\\\\\\\\\
c
c=======================================================================
c
c
       subroutine nlsdac16 ( nlunit, line, jbeg, jend, jran, iquote
     1                     , word32 )
c
c    dac:namelist.nlsdac16 <-------------- returns next "word" in "line"
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine returns the next 32 character "word" found
c  after the input position of "jbeg" in "line".  The beginning and
c  ending positions of "word" are returned as well ("jbeg" and "jend").
c
c  There are two types of delimiters which define "word".  The first
c  type include the double and single quotation marks [" and '].  The
c  quotes are returned with "word".  Both the beginning and ending of
c  "word" must be delimited with the same type of quotation mark.
c
c  The second type include the equals sign [=], the asterisk [*], the
c  comma [,], the full colon [:], the opening parenthesis [(], the
c  closing parenthseis [)], the dollar sign [$], the blank, and the end
c  of "line".  These delimiters are not returned with "word", and so
c  "word" may be as long as 32 characters.  "word" may begin and end
c  with different delimiters of the second type.  (Note that $ and the
c  end of "line" can only delimit the end of "word".)  Delimiters of the
c  first type may be used to end "word" beginning with a delimiter of
c  the first type, and will not be returned with "word".  The reverse is
c  not true.  Delimiters of the second type found between like quotation
c  marks are returned as characters in "word".  Note that the search for
c  the next word will start directly at abs(jbeg) if "jbeg" < 0.
c
c  If the end of "line" is encountered before the next "word" is found,
c  the next "line" is read.  Both "line" and "word" are returned to the
c  calling subroutine.
c
c  EXTERNALS:  NLSDAC15
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*1   char1   , char2   , char3
       character*32  word32  , blank32
       character*72  line    , oldline
       integer       nlunit  , jbeg    , jend    , jran    , iquote
     1             , j       , j1      , j2      , iflag
c
c      Data statements
c
       data          blank32 /'                                '/
c
c      External statements
c
       external      nlsdac15
c
c-----------------------------------------------------------------------
c
c      Scan "line" for beginning of next "word".
c
       iflag = 1
       if (jbeg .lt. 0) then
         iflag = - 1
         jbeg  = - jbeg
         if (jbeg .gt. 72) go to 30
         go to 70
       endif
       j1 = jbeg
       if (jbeg .eq. 0) j1 =  3
10     continue
       do 20 j=j1,72
         char1 = line(j:j)
         if ( (char1 .eq. ' ') .or. (char1 .eq. ',') .or.
     1        (char1 .eq. '=') .or. (char1 .eq. ')') .or.
     2        (char1 .eq. '(') .or. (char1 .eq. '*') .or.
     3        (char1 .eq. ':') ) go to 20
         jbeg = j
         go to 70
20     continue
30     continue
c
c      The only way for execution to have reached this point is for the
c  end of the line to have been reached without finding another "word".
c  Read next "line", and resume search.
c
       oldline = line
40     continue
       read  (nlunit, 1010, end=50, err=60) line
c
c      Error message if column 1 contains anything but a 'c', 'C', or a
c  blank.
c
       char2 = line(1:1)
       if ( (char2 .ne. 'c') .and. (char2 .ne. 'C') .and.
     1      (char2 .ne. ' ') ) call nlsdac15 ( line, 1, 1 )
c
c      Error message if column 2 contains anything but a '$' or a blank
c  (unless column 1 contains a 'c' or a 'C').
c
       char3 = line(2:2)
       if ( (char2 .ne. 'c') .and. (char2 .ne. 'C') .and.
     1      (char3 .ne. '$') .and. (char3 .ne. ' ') )
     2   call nlsdac15 ( line, 2, 2 )
c
c      If column 1 contains a 'c' or a 'C', the line is "commented out",
c  and is echoed on the terminal.
c
       if ( (char2 .eq. 'c') .or. (char2 .eq. 'C') ) then
         write (6, 2010) line
         go to 40
       endif
c
c      If column 2 contains a '$' sentinal, then a new namelist has
c  begun without the previous one being closed.  Issue error message.
c
       if (char3 .eq. '$') then
         if (iquote .eq. 0) then
           call nlsdac15 ( oldline, 72, 13 )
         else
           call nlsdac15 ( line, 2, 9 )
         endif
       endif
c
c      New "line" is OK, resume scanning.
c
       if (iflag .eq. -1) then
         jbeg = 3
         go to 70
       else
         j1 = 3
         go to 10
       endif
c
c      Error: premature end of file.
c
50     continue
       call nlsdac15 ( line, 72, 12 )
c
c      Error: problem reading input deck
c
60     continue
       call nlsdac15 ( blank32, 0, 7 )
c
c      "jbeg" has been determined.  Now find "jend", which cannot be
c  greater than "jbeg" + 31.
c
70     continue
       j1   = jbeg + 1
       j2   = jbeg + 31
       if (j2 .gt. 72) j2 = 72
       jend = jbeg
c
c      Case:  first delimiter is a single quote.
c
       if ( (iquote .eq. 1) .or. (char1 .eq. '''') ) then
         do 80 j=j1,j2
           jend  = j
           char1 = line(j:j)
           if (char1 .eq. '''') go to 110
80       continue
c
c      Closing single quote not found.  Set "iquote" to 1.
c
         iquote = 1
         go to 120
       endif
c
c      Case:  first delimiter is a double quote.
c
       if ( (iquote .eq. 2) .or. (char1 .eq. '"') ) then
         do 90 j=j1,j2
           jend  = j
           char1 = line(j:j)
           if (char1 .eq. '"') go to 110
90       continue
c
c      Closing double quote not found.  Set "iquote" to 2.
c
         iquote = 2
         go to 120
       endif
c
c      Case:  first delimiter is not a quote.
c
       do 100 j=j1,j2
         char1 = line(j:j)
         if ( (char1 .eq. ' ') .or. (char1 .eq. ',') .or.
     1        (char1 .eq. '=') .or. (char1 .eq. ')') .or.
     2        (char1 .eq. '(') .or. (char1 .eq. '*') .or.
     3        (char1 .eq. ':') .or. (char1 .eq. '"') .or.
     4        (char1 .eq. '''') .or. (char1 .eq. '$') ) go to 110
         jend = j
100    continue
110    continue
       iquote = 0
120    continue
       jran   = jend - jbeg + 1
       word32 = blank32
       word32(1:jran) = line(jbeg:jend)
c
       return
c
c-----------------------------------------------------------------------
c----------------------- Read format statements ------------------------
c-----------------------------------------------------------------------
c
1010   format(a72)
c
c-----------------------------------------------------------------------
c---------------------- Write format statements ------------------------
c-----------------------------------------------------------------------
c
2010   format(a72)
c
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////              N L S D A C 1 6              \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              N L S D A C 1 7              \\\\\\\\\\
c
c=======================================================================
c
c
       subroutine nlsdac17 ( nlunit, line, jbeg, jend, char1 )
c
c    dac:namelist.nlsdac17 <----------- returns next non-blank character
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine returns the first non-blank character found
c  at or after the input position "jbeg" in "line".  If no non-blank
c  character is found, the next line is read and scanned for the first
c  non-blank character.
c
c  EXTERNALS:  NLSDAC15
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*1   char1   , char2   , char3
       character*8   blank8
       character*72  line    , oldline
       integer       nlunit  , jbeg    , jend    , j       , j1
c
c      Data statements
c
       data          blank8 / '        ' /
c
c      External statements
c
       external      nlsdac15
c
c-----------------------------------------------------------------------
c
c      Scan "line" for next non-blank character.
c
       j1 = jbeg
       if (j1 .eq. 0) j1 = 3
10     continue
       do 20 j=j1,72
         jbeg  = j
         jend  = j
         char1 = line(j:j)
         if (char1 .ne. ' ') return
20     continue
c
c      End of line reached before non-blank character found.  Read
c  next line.
c
       oldline = line
30     continue
       read  (nlunit, 1010, end=40, err=50) line
c
c      Error message if column 1 contains anything but a 'c', 'C', or a
c  blank.
c
       char2 = line(1:1)
       if ( (char2 .ne. 'c') .and. (char2 .ne. 'C') .and.
     1      (char2 .ne. ' ') ) call nlsdac15 ( line, 1, 1 )
c
c      Error message if column 2 contains anything but a '$' or a blank
c  (unless column 1 contains a 'c' or a 'C').
c
       char3 = line(2:2)
       if ( (char2 .ne. 'c') .and. (char2 .ne. 'C') .and.
     1      (char3 .ne. '$') .and. (char3 .ne. ' ') )
     2   call nlsdac15 ( line, 2, 2 )
c
c      If column 1 contains a 'c' or a 'C', the line is "commented out",
c  and is echoed on the terminal.
c
       if ( (char2 .eq. 'c') .or. (char2 .eq. 'C') ) then
         write (6, 2010) line
         go to 30
       endif
c
c      If column 2 contains a '$' sentinal, then a new namelist has
c  begun without the previous one being closed.  Issue error message.
c
       if (char3 .eq. '$') call nlsdac15 ( oldline, 72, 13 )
c
c      New "line" is OK, resume scanning.
c
       j1 = 3
       go to 10
c
c      Error: premature end of file.
c
40     continue
       call nlsdac15 ( line, 72, 12 )
c
c      Error: problem reading input deck
c
50     continue
       call nlsdac15 ( blank8, 0, 7 )
c
c-----------------------------------------------------------------------
c----------------------- Read format statements ------------------------
c-----------------------------------------------------------------------
c
1010   format(a72)
c
c-----------------------------------------------------------------------
c---------------------- Write format statements ------------------------
c-----------------------------------------------------------------------
c
2010   format(a72)
c
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////              N L S D A C 1 7              \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              N L S D A C 1 8              \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac18 ( line, jbeg, word32 )
c
c    dac:namelist.nlsdac18 <---- checks for valid integer representation
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine calls nlsdac15 (which issues fatal error
c  messages) if the input character string ("word32") does not represent
c  an integer.
c
c  EXTERNALS:  NLSDAC15
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*1   char1
       character*32  word32
       character*72  line
       integer       jbeg    , ineg    , ipos    , idig    , i
     1             , ibeg    , iend    , j       , jpos
c
       character*1   ch      (  10)
c
c      Data statements
c
       data          ch       /'0','1','2','3','4','5','6','7','8','9'/
c
c      External statements
c
       external      nlsdac15
c
c-----------------------------------------------------------------------
c
c      Initialise parameters.
c
       ineg  = 0
       ipos  = 0
       idig  = 0
c
c      Find first non-blank character.
c
       do 10 i=1,32
         ibeg  = i
         char1 = word32(i:i)
         if (char1 .ne. ' ') go to 20
10     continue
c
c      "word32" is blank.  Issue an error message.
c
       call nlsdac15 ( line, jbeg, 11 )
c
c      Scan word32 up to the next blank.  An error message is issued if:
c
c   1. more than one sign (+ or -) is encountered, or
c   2. a sign is embedded in the integer, or
c   3. if the integer is found to have more than 15 digits, or
c   4. if a character other than +, -, or a digit is found.
c
20     continue
       do 40 i=ibeg,32
         jpos  = jbeg + i
         iend  = i
         char1 = word32(i:i)
         if (char1 .eq. ' ') go to 50
         if (char1 .eq. '-') then
           ineg = ineg + 1
           if ( (ineg .gt. 1) .or. (ipos .gt. 0) )
     1       call nlsdac15 ( line, jpos, 4 )
           if (idig .gt. 0) call nlsdac15 ( line, jpos, 4 )
           go to 40
         endif
         if (char1 .eq. '+') then
           ipos = ipos + 1
           if ( (ipos .gt. 1) .or. (ineg .gt. 0) )
     1       call nlsdac15 ( line, jpos, 4 )
           if (idig .gt. 0) call nlsdac15 ( line, jpos, 4 )
           go to 40
         endif
         do 30 j=1,10
           if (char1 .eq. ch(j)) then
             idig = idig + 1
             if (idig .gt. 15) call nlsdac15 ( line, jpos, 15 )
             go to 40
           endif
30       continue
         call nlsdac15 ( line, jpos, 4 )
40     continue
c
c      First blank after data entry is found.
c
50     continue
c
c      If there are no digits after sign, issue error message.
c
       if (idig .eq. 0) call nlsdac15 ( line, jpos, 4 )
c
c      Scan rest of "word32" and issue error message if a non-blank
c  character is found.
c
       if (iend .lt. 32) then
         do 60 i=iend,32
           char1 = word32(i:i)
           jpos  = jbeg + i
           if (char1 .ne. ' ') call nlsdac15 ( line, jpos, 4 )
60       continue
       endif
       return
c
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////              N L S D A C 1 8              \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              N L S D A C 1 9              \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac19 ( word32, ival )
c
c    dac:namelist.nlsdac19 <----- translates character string to integer
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine translates a character string to an
c  integer.  It is assumed that the character string is a valid
c  representation of an integer, and is error-free (i.e., NLSDAC18 has
c  been called).
c
c  EXTERNALS:  NONE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*1   char1
       character*32  word32
       integer       ival    , ichar1  , ifact   , i
c
       character*1   ch      (  10)
c
c      Data statements
c
       data          ch       /'0','1','2','3','4','5','6','7','8','9'/
c
c-----------------------------------------------------------------------
c
       ichar1 = 32
       ifact  = 1
       ival   = 0
10     continue
       char1 = word32(ichar1:ichar1)
       if (char1 .ne. ' ') then
         do 20 i=1,10
           if (char1 .eq. ch(i)) go to 30
20       continue
         if (char1 .eq. '-') then
           ival = - ival
           return
         endif
30       continue
         ival  = ival + ifact * (i - 1)
         ifact = ifact * 10
       endif
       ichar1 = ichar1 - 1
       if (ichar1 .gt. 0) go to 10
       return
c
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////              N L S D A C 1 9              \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              N L S D A C 2 0              \\\\\\\\\\
c
c=======================================================================
c
c
       subroutine nlsdac20 ( ival, word32 )
c
c    dac:namelist.nlsdac20 <----- translates integer to character string
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine translates an integer into a left-justified
c  character string.
c
c  EXTERNALS:  NONE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*32  word32  , blank32
       integer       ival    , j       , jj      , jbeg    , jend
     1             , ilog
       real*8        x       , x1      , xlog
c
       character*1   ch      (  10)
c
c      Data statements
c
       data          ch       /'0','1','2','3','4','5','6','7','8','9'/
       data          blank32  /'                                '/
c
c-----------------------------------------------------------------------
c
       word32 = blank32
       if (ival .lt. 0) then
         jbeg = 2
         ival = - ival
         word32(1:1) = '-'
       else
         jbeg = 1
       endif
       x = dfloat(ival)
       if (ival .eq. 0) then
         ilog = 0
       else
         xlog = dlog10(x)
         ilog = idint (xlog + .000001d0)
       endif
       jend = jbeg + ilog
       if (jend .ge. 32) jend = 32
       do 10 j=jbeg,jend
         x1 = x / 10.0d0**ilog
         jj = idint ( x1 + .000001d0 )
         word32(j:j) = ch(jj+1)
         x  = x - dfloat(jj) * 10**ilog
         ilog = ilog - 1
10     continue
       return
c
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////              N L S D A C 2 0              \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 2 1             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac21 ( word32, rval )
c
c    dac:namelist.nlsdac21 <-------- translates character string to real
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine translates a character string to a real
c  constant.
c
c  EXTERNALS:  NONE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*1   char1
       character*32  word32
       integer       i       , esign   , ipnt    , npnt    , more
     1             , iblnk   , iman    , iexp    , j
       real*8        rval    , msign   , term    , fact
c
       character*1   ch      (  10)
       integer       exponent(   4), mantissa(  16)
c
c      Data statements
c
       data          ch       /'0','1','2','3','4','5','6','7','8','9'/
c
c-----------------------------------------------------------------------
c
c      Initialise parameters.
c
       do 10 i=1,4
         exponent(i) = 0
10     continue
       do 20 i=1,16
         mantissa(i) = 0
20     continue
       msign = 1.0d0
       esign = 1
       ipnt  = 0
       npnt  = 0
       more  = 1
       iblnk = 0
       iman  = 0
       iexp  = 0
c
c      Scan "word32".  It is assumed that "word32" has already been
c  determined to be error-free, and that it is a valid character
c  representation of a real value (i.e., NLSDAC23 has been called).
c
       do 40 i=1,32
         char1 = word32(i:i)
         if (char1 .eq. ' ') then
           iblnk = iblnk + 1
           go to 40
         endif
         if (char1 .eq. '-') then
           if (more .eq. 1) msign = -1.0d0
           if (more .eq. 2) esign = -1
           go to 40
         endif
         if (char1 .eq. '+') go to 40
         if (char1 .eq. '.') then
           npnt = 1
           ipnt = i - 2 - iblnk
           if (msign .eq. -1.0d0) ipnt = ipnt - 1
           go to 40
         endif
         if ( (char1 .eq. 'e') .or. (char1 .eq. 'd') ) then
           more = 2
           if (npnt .eq. 0) then
             npnt = 1
             ipnt = i - 2 - iblnk
             if (msign .eq. -1.0d0) ipnt = ipnt - 1
             if (ipnt .lt. 0) ipnt = 0
           endif
           go to 40
         endif
         do 30 j=1,10
           if (char1 .eq. ch(j)) then
             if (more .eq. 1) then
               iman = iman + 1
               mantissa(iman) = j - 1
             endif
             if (more .eq. 2) then
               iexp = iexp + 1
               exponent(iexp) = j - 1
             endif
             go to 40
           endif
30       continue
40     continue
       if (npnt .eq. 0) ipnt = iman - 1
c
c      Evaluate the mantissa.
c
       rval = 0.0d0
       if (iman .gt. 0) then
         do 50 i=1,iman
           term = dfloat(mantissa(i)) * 10.0d0**ipnt
           rval = rval + term
           ipnt = ipnt - 1
50       continue
       else
         rval = 1.0d0
       endif
       rval = rval * msign
c
c      Evaluate the exponential factor.
c
       fact = 1.0d0
       if (iexp .gt. 0) then
         do 60 i=1,iexp
           ipnt = iexp - i
           fact = fact * 10.0d0 ** (exponent(i) * 10.0d0**ipnt)
60       continue
       endif
       if (esign .eq. -1) fact = 1.0d0 / fact
c
c      Return real representation of character string.
c
       rval = rval * fact
       return
c
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 2 1             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 2 2             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac22 ( line, jbeg, word32 )
c
c    dac:namelist.nlsdac22 <---- checks for valid logical representation
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine calls nlsdac15 (which issues fatal error
c  messages) if the input character string ("word32") does not represent
c  a valid logical value.
c
c  EXTERNALS:  NLSDAC15
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*32  word32
       character*72  line
       integer       jbeg
c
c      External statements
c
       external      nlsdac15
c
c-----------------------------------------------------------------------
c
       if ( (word32(1:8) .ne. '.true.  ') .and.
     1      (word32(1:8) .ne. '.t.     ') .and.
     2      (word32(1:8) .ne. '.false. ') .and.
     3      (word32(1:8) .ne. '.f.     ') )
     4   call nlsdac15 ( line, jbeg, 5 )
c
       return
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 2 2             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 2 3              \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac23 ( line, jbeg, word32 )
c
c    dac:namelist.nlsdac18 <-- determines if character string is a valid
c                                         representation of a real value
c                                                              may, 1989
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine calls nlsdac15 (which issues various fatal
c  error messages) if the input character string ("word32") does not
c  represent a real value.
c
c  EXTERNALS:  NLSDAC15
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*1   char1
       character*32  word32
       character*72  line
       integer       jbeg    , mneg    , eneg    , mpos    , epos
     1             , iexp    , iman    , ipnt    , more    , i
     2             , ibeg    , iend    , j       , jpos
c
       character*1   ch      (  10)
c
c      Data statements
c
       data          ch       /'0','1','2','3','4','5','6','7','8','9'/
c
c      External statements
c
       external      nlsdac15
c
c-----------------------------------------------------------------------
c
c      Initialise parameters.
c
       mneg  = 0
       eneg  = 0
       mpos  = 0
       epos  = 0
       iexp  = 0
       iman  = 0
       ipnt  = 0
       more  = 1
c
c      Find first non-blank character.
c
       do 10 i=1,32
         ibeg  = i
         char1 = word32(i:i)
         if (char1 .ne. ' ') go to 20
10     continue
c
c      "word32" is blank.  Issue an error message.
c
       call nlsdac15 ( line, jbeg, 11 )
c
c      Scan word32 up to the next blank.  An error message is issued if:
c
c   1. more than one sign (+ or -) is encountered in the mantissa, or
c   2. a sign is embedded in the mantissa, or
c   3. more than one sign (+ or -) is encountered in the exponent, or
c   4. a sign is embedded in the exponent, or
c   5. if more than one decimal point is encountered in the mantissa, or
c   6. if a decimal point is encountered in the exponent, or
c   7. if more than one exponential sentinal ("e" or "d") is found, or
c   8. if the mantissa is found to have more than 16 digits, or
c   9. if the mantissa is found to have more than  4 digits, or
c  10. if a character other than +, -, ., e, d, or a digit is found.
c
20     continue
       do 40 i=ibeg,32
         jpos  = jbeg + i
         iend  = i
         char1 = word32(i:i)
         if (char1 .eq. ' ') go to 50
         if ( (char1 .eq. '-') .and. (more .eq. 1) ) then
           mneg = mneg + 1
           if ( (mneg .gt. 1) .or. (mpos .gt. 0) )
     1       call nlsdac15 ( line, jpos, 4 )
           if (iman .gt. 0) call nlsdac15 ( line, jpos, 4 )
           go to 40
         endif
         if ( (char1 .eq. '-') .and. (more .eq. 2) ) then
           eneg = eneg + 1
           if ( (eneg .gt. 1) .or. (epos .gt. 0) )
     1       call nlsdac15 ( line, jpos, 4 )
           if (iexp .gt. 0) call nlsdac15 ( line, jpos, 4 )
           go to 40
         endif
         if ( (char1 .eq. '+') .and. (more .eq. 1) ) then
           mpos = mpos + 1
           if ( (mpos .gt. 1) .or. (mneg .gt. 0) )
     1       call nlsdac15 ( line, jpos, 4 )
           if (iman .gt. 0) call nlsdac15 ( line, jpos, 4 )
           go to 40
         endif
         if ( (char1 .eq. '+') .and. (more .eq. 2) ) then
           epos = epos + 1
           if ( (epos .gt. 1) .or. (eneg .gt. 0) )
     1       call nlsdac15 ( line, jpos, 4 )
           if (iexp .gt. 0) call nlsdac15 ( line, jpos, 4 )
           go to 40
         endif
         if ( (char1 .eq. '.') .and. (more .eq. 1) ) then
           ipnt = ipnt + 1
           iman = iman + 1
           if (ipnt .gt. 1) call nlsdac15 ( line, jpos, 4 )
           go to 40
         endif
         if ( (char1 .eq. '.') .and. (more .eq. 2) ) then
           call nlsdac15 ( line, jpos, 4 )
         endif
         if ( (char1 .eq. 'e') .or. (char1 .eq. 'd') ) then
           if (more .eq. 2) call nlsdac15 ( line, jpos, 4 )
           more = 2
           go to 40
         endif
         do 30 j=1,10
           if (char1 .eq. ch(j)) then
             if (more .eq.  1) iman = iman + 1
             if (iman .gt. 16) call nlsdac15 ( line, jpos, 15 )
             if (more .eq.  2) iexp = iexp + 1
             if (iexp .gt.  3) call nlsdac15 ( line, jpos, 14 )
             go to 40
           endif
30       continue
         call nlsdac15 ( line, jpos, 4 )
40     continue
c
c      First blank after data entry is found.
c
50     continue
c
c      If there are no digits after sign and/or decimal point, issue
c  error message.
c
       iman = iman - ipnt
       if (iman .eq. 0) call nlsdac15 ( line, jpos, 4 )
c
c      Scan rest of "word32" and issue error message if a non-blank
c  character is found.
c
       if (iend .lt. 32) then
         do 60 i=iend,32
           char1 = word32(i:i)
           jpos  = jbeg + i
           if (char1 .ne. ' ') call nlsdac15 ( line, jpos, 4 )
60       continue
       endif
       return
c
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 2 3             \\\\\\\\\\
c
c=======================================================================
c
c
c=======================================================================
c
c     \\\\\\\\\\     B E G I N   S U B R O U T I N E     //////////
c     //////////             N L S D A C 2 4             \\\\\\\\\\
c
c=======================================================================
c
       subroutine nlsdac24 ( lunit, cname, nmlst, nmax )
c
c    dac:namelist.nlsdac24 <---- dumps namelist values to specified unit
c                                                         february, 1991
c
c      Written by: David Clarke
c      Modified 1:
c
c  PURPOSE:  This subroutine dumps a formatted summary of all user-
c  specified namelist parameters in the namelist "cname" to the logical
c  unit "lunit".
c
c  EXTERNALS:  NONE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       logical       lval
       character*1   char1   , attr    , type
       character*4   jbeg    , jend    , kbeg    , kend
       character*8   cname   , word8   , temp8   , nword8
       character*25  var
       character*32  blank32
       character*512 cval
       integer       lunit   , nmax    , i       , ilist   , ival
     1             , lbeg    , lend    , lran    , j       , j1
     2             , j2      , mbeg    , mend
       real*8        rval
c
       character*32  nmlst   (   2,   *)
c
c      Data statements
c
       data          blank32  / '                                ' /
c
c      External statements
c
       external      nlsdac19, nlsdac21
c
c-----------------------------------------------------------------------
c
       lbeg = 1
       do 10 j=1,16
         j2 = 32 * j
         j1 = j2 - 31
         cval(j1:j2) = blank32
10     continue
c
c      First line to output unit.
c
       write (lunit, 2010) cname
c
c      Read namelist entries, and write contents to output unit.
c
       do 100 i=1,nmax
         ilist = i
         if (nmlst(2,ilist) .eq. blank32) go to 100
         char1 = nmlst(1,ilist)(31:31)
         if (char1 .eq. '*') go to 100
         word8 = nmlst(1,ilist)( 1: 8)
         if (word8 .eq. '        ') go to 110
         nmlst(1,ilist)(31:31) = '*'
20       continue
         var   = blank32(1:25)
         attr  = nmlst(1,ilist)(30:30)
         type  = nmlst(1,ilist)(32:32)
c
c      Representation of variable name (contents of "nmlst(1,*)")
c  depends on type of variable (ie, scalar, vector, or 2-D array).
c
         if (type .ne. 's') then
           jbeg = nmlst(1,ilist)( 9:12)
           jend = nmlst(1,ilist)(13:16)
           if (jend .eq. '    ') jend = jbeg
           kbeg = nmlst(1,ilist)(17:20)
           kend = nmlst(1,ilist)(21:24)
           if (kend .eq. '    ') kend = kbeg
           if (type .eq. 'v') write (var, 2020) word8, jbeg, jend
           if (type .eq. 'a') write (var, 2030) word8, jbeg, jend
     1                                               , kbeg, kend
         else
           write (var, 2040) word8
         endif
c
c      Representation of variable value (contents of "nmlst(2,*)")
c  depends on attribute (ie, integer, real, logical, or character).
c
c  Case:  integer
c
         if (attr .eq. 'i') then
           call nlsdac19 ( nmlst(2,ilist), ival )
           write (lunit, 2050) var, ival
         endif
c
c  Case:  real
c
         if (attr .eq. 'r') then
           call nlsdac21 ( nmlst(2,ilist), rval )
           write (lunit, 2060) var, rval
         endif
c
c  Case:  logical
c
         if (attr .eq. 'l') then
           temp8 = nmlst(2,ilist)(1:8)
           if (temp8 .eq. '.true.  ') lval = .true.
           if (temp8 .eq. '.t.     ') lval = .true.
           if (temp8 .eq. '.false. ') lval = .false.
           if (temp8 .eq. '.f.     ') lval = .false.
           if (      lval) write (lunit, 2070) var
           if (.not. lval) write (lunit, 2080) var
         endif
c
c  Case:  character
c
c  For character strings, there is the possibility that the string is
c  stored in pieces throughout "nmlst(2,*)".  Thus, delay writing the
c  entry to the output unit until all the character string can be
c  gathered.
c
         if (attr .eq. 'c') then
           if (nmlst(2,ilist)(32:32) .eq. '&') then
             lran = ichar ( nmlst(2,ilist)(31:31) )
             lend = lbeg + lran - 1
             cval(lbeg:lend) = nmlst(2,ilist)(1:lran)
             lbeg = lend + 1
           else
             do 50 lran=31,1,-1
               if (nmlst(2,ilist)(lran:lran) .ne. ' ') go to 60
50           continue
60           continue
             lend = lbeg + lran - 1
             cval(lbeg:lend) = nmlst(2,ilist)(1:lran)
             if (lend .le. 50) then
               write (lunit, 2090) var, cval(1:lend)
             else
               mbeg = 1
               mend = 50
               write (lunit, 2100) var, cval(mbeg:mend)
70             continue
               mbeg = mend + 1
               mend = mend + 50
               if (mend .ge. lend) then
                 write (lunit, 2110) cval(mbeg:lend)
               else
                 write (lunit, 2120) cval(mbeg:mend)
                 go to 70
               endif
             endif
             lbeg = 1
             do 80 j=1,16
               j2 = 32 * j
               j1 = j2 - 31
               cval(j1:j2) = blank32
80           continue
           endif
         endif
c
c      Check the rest of "nmlst(1,*)" for any other variable entries
c  with the same variable name.  Such entries may exist for vectors and
c  arrays of all attributes, and scalars for character variables.  An
c  asterisk is put (temporarily) in column 31 of "nmlst(1,ilist)" so
c  that this entry will not be read again.
c
         if (ilist .ne. nmax) then
           do 90 j=ilist+1,nmax
             nword8  = nmlst(1,j)( 1: 8)
             if (nword8 .eq. '        ') go to 100
             if (nword8 .eq. word8) then
               nmlst(1,j)(31:31) = '*'
               ilist = j
               go to 20
             endif
90         continue
         endif
100    continue
110    continue
c
c      Last line to output unit.
c
       write (lunit, 2130) cname
c
c      Replace asterisks put in column 31 of "nmlst(1,*)" with
c  blanks.
c
       do 120 i=1,nmax
         if (nmlst(1,i)(1:8) .eq. '        ') return
         nmlst(1,i)(31:31) = ' '
120    continue
c
       return
c
c-----------------------------------------------------------------------
c---------------------- Write format statements ------------------------
c-----------------------------------------------------------------------
c
2010   format(/,'======================================================'
     1       ,'> BEGIN NAMELIST: ',a8)
2020   format(a8,'(',a3,':',a3,')')
2030   format(a8,'(',a3,':',a3,',',a3,':',a3,')')
2040   format(a8)
2050   format(a25,' =',i10)
2060   format(a25,' =',1pg20.12)
2070   format(a25,' = .true.')
2080   format(a25,' = .false.')
2090   format(a25,' = "',a,'"')
2100   format(a25,' = "',a)
2110   format(29x,a,'"')
2120   format(29x,a)
2130   format('========================================================'
     1       ,'> END NAMELIST: ',a8)
c
       end
c
c=======================================================================
c
c     \\\\\\\\\\       E N D   S U B R O U T I N E       //////////
c     //////////             N L S D A C 2 4             \\\\\\\\\\
c
c=======================================================================
c
c
c---+----1----+----2----+----3----+----4----+----5----+----6----+----7--
c
c>>>>>>>>>>>>>>>>>>>>> EXTERNALLY CALLED ROUTINES <<<<<<<<<<<<<<<<<<<<<<
c
c  To use the capabilities of this library, the user would only call the 
c  following three routines in their program.  All subsequent routines 
c  are included to satisfy all the calls that the following three 
c  routines make.
c
c      GRFXOPN    opens the plot file
c      GRFX1D     makes a single page, multiple-frame, single variable 
c                 per frame plot
c      GRFXCLS    closes the plot file.
c
c  Consult the headers of each of these routines to determine the 
c  required syntax and purpose of each variable in the argument list.
c
c  As an example of a call to GRFX1D:
c
c      call grfx1d ( psi, den, 0, 1, npnt, 1, 1, 0.0, 0.0, 0.0, 0.0
c    1             , 2, 1.0, 1, 2, 2, 1, 2, 1, xlab, ylab )
c
c  creates a square linear-linear graph of all "npnt" points of the 
c  dependent variable "den" (density) plotted against the independent 
c  variable "psi", allowing the plotter to determine the optimal 
c  plotting window, joining all points with lines.  The last six 
c  integers the plotter to arrange the graphs in a 2 x 2 fashion on a 
c  single page (rather than one plot per page).
c
c  The last two parameters, "xlab" and "ylab" use something called "NCAR 
c  mark-up language" to label the plots.  You're perfectly safe to just 
c  set xlab='x' and ylab='rho' here, though if you wanted to be a little 
c  fancy, try: xlab = '|F10|x/t', ylab = '|F5|r'.
c
c=======================================================================
c
       subroutine grfx1d ( xvar, yvar, yexp, ibeg, iend, linlog, imm
     1                   , xmn, xmx, ymn, ymx, corl, aspect, allxl
     2                   , allyl, np1h, ip1h, np1v, ip1v, xlab, ylab )
c
c    dac:grfx03.grfx1d <-------------------------------- plots 1-d slice
c                                                             june, 1990
c
c    written by: David Clarke
c    modified 1: June, 1996 by David Clarke; rewritten to generate
c                publication quality plots.
c    modified 2: December, 1996 by David Clarke; added multiple frames
c                per page capability.
c    modified 3: October, 2003 by Alexander Men'shchikov; generalized to
c                include the PSPLOT graphics library.
c    modified 4: February, 2005 by David Clarke; streamlined Sasha's
c                modifications
c    modified 5: April, 2006 by DAC; introduced "yexp", in analogy to 
c                "qexp" and "vexp" in GRFX2D.
c    modified 6: December, 2016 by DAC; simplified from July, 2015 
c                version for this "cheap and cheerful" standalone 
c                library to do single-variable, multi-framed plots on a 
c                single page.
c
c  PURPOSE: This is a pared down version of the subroutine by the same 
c  name in GRFX03 (as of last changes made in July, 2015) that isolates 
c  all the calls to the PSPLOT library necessary for making a single 
c  page of possibly multiple frames, each with a single variable 
c  plotted.  There is no legend capability, and the NCAR option has been 
c  deleted.  While this version does not directly support writing a 
c  header (title) or footer to the graph, there is a standalone routine 
c  in this package called HDRFTR that can put a header and/or footer to 
c  the graph is desired.  For the full capabilities of GRFX03, that 
c  library should be used.
c
c  For clarity, the terms "page", "frame", and "plot" are used 
c  consistently.  Thus, on a single page, there may be numerous frames 
c  (np1h x np1v), and on each frame, there may be numerous plots (though 
c  in this version, only one plot per frame).  The word "graph" is used 
c  generically and refers to any graphical unit, be it a single plot, a 
c  frame, or the entire page.  Its meaning is gleaned from context.
c
c  If a plot is to start from within the frame rather than from the 
c  border (e.g., there may be no solution to what is being plotted for 
c  some values of the abscissa), the function value should be set to 
c  -1.1e-32 or more negative at the points to be ignored.  Such points 
c  will not have line segments drawn to or from them, even to intersect 
c  the border.
c
c  INPUT VARIABLES:
c    xvar        vector    of independent variable (x-axis)
c    yvar        vector(s) of   dependent variable (y-axis)
c    yexp        log10 of scaling factor generated when "yvar" was
c                converted to single precision (integer)
c    ibeg        index of first value in "xvar" and "yvar" to plot
c    iend        index of last  value in "xvar" and "yvar" to plot
c    linlog      = 1 => linear x and y;    = 2 => linear x, log y
c                = 3 => log x, linear y;   = 4 => log x and y
c    imm         = 0 => use input "ymn" and "ymx" to limit range of data
c                       plotted
c                =-1 => same as 0, and extend y-axis by one minor tick
c                = 1 => plot entire range of data; extend y-axis
c    xmn         minimum value on the x-axis
c    xmx         maximum value on the x-axis
c    ymn         minimum value on the y-axis
c    ymx         maximum value on the y-axis
c    corl        =-6 => open diamonds; one symbol per datum
c                =-5 => open pentagrams
c                =-4 => asterisk (*)
c                =-3 => open squares
c                =-2 => open triangles
c                =-1 => plus signs (+)
c                = 0 => crosses (x)
c                = 1 => open circles (o)
c                = 2 => smooth curve drawn through data
c    aspect      desired aspect ratio of axes.
c                < 1.0 => short, wide frame
c                = 1.0 => square frame
c                > 1.0 => tall, narrow frame
c    allxl       =2 => all x labels/annotations are drawn (default)
c                =1 => only the x-label/annotations on the bottom row
c                      of frames on the page are drawn.
c                =0 => labels aren't even drawn on the bottom row.
c                if allxl < 0, same as abs(allxl) except last annotation
c                is not drawn if it falls at xxmax.
c    allyl       =2 => all y labels/annotations are drawn (default)
c                =1 => only the y-label/annotations on the left column
c                      of frames on the page are drawn.
c                =0 => labels aren't even drawn on the left column.
c                if allyl < 0, same as abs(allyl) except last annotation
c                is not drawn if it falls at yymax.
c    np1h        number of frames to be arranged horizontally per page
c    ip1h        in the current row   , this is the ip1h'th frame
c    np1v        number of frames to be arranged vertically   per page
c    ip1v        in the current column, this is the ip1v'th frame
c    xlab        label for x-axis
c    ylab        label for y-axis
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    GRFXAXES
c    PSCHARS
c    SETLW               )
c    SETCOLR             )
c    OCIRCLE , TIMES     )
c    PLUS    , TRIANGLE  )---- PSPLOT routines
c    SQUARE  , ASTERISK  )
c    STAR    , DIAMOND   )
c    DSHCRV  , PLOT      )
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       integer       ndim
       real          tiny    , huge    
       parameter     ( ndim = 9999
     1               , tiny = 1.0e-32, huge = 1.0e32 )
c
       character*(*) xlab    , ylab    
       logical       lmax0   , lastpt  
       integer       yexp    , ibeg    , iend    , linlog  , imm      
     1             , corl    , allxl   , allyl   , np1h    , ip1h    
     2             , np1v    , ip1v    , lnlg    , nzi     , i       
     4             , i1      , ip1     , ip2     , ip3     , eye     
     6             , ndash   , n       , ista    , ifin    
       real          xmn     , xmx     , ymn     , ymx     , penwt   
     1             , aspect  , xmin    , xmax    , ymin    , ymax    
     2             , yming   , ymaxg   , yminl   , ymaxl   , aymaxl  
     2             , q1      , q2      , ymingt0 , p1x0    , p1y0       
     4             , size    , deks    , eksip1  , eksmin  , eksmax  
     7             , delwhy  , dwhyll  , dwhyl   , dwhyr   , dwhy    
     8             , dwhyrr  , whymin  , whymax  , whyip1  , ffx     
     9             , ffy     
c
       real          xvar    (   *), yvar    (   *)
     1             , eks     (ndim), ex      (ndim), why     (ndim)
c
       real          wxmin   , wxmax   , wymin   , wymax   , xxmin
     1             , xxmax   , yymin   , yymax   , scale
       common      / win1d /
     1               wxmin   , wxmax   , wymin   , wymax   , xxmin
     2             , xxmax   , yymin   , yymax   , scale
c
c      External statements
c
       external      grfxaxes, pschars , setlw   , setcolr , ocircle 
     1             , times   , plus    , triangle, square  , asterisk
     3             , star    , diamond , dshcrv  , plot     
c
c-----------------------------------------------------------------------
c
       lnlg  = max ( 1, min ( 4, linlog ) )
       penwt = 1.0
       call setlw ( 0.03 * penwt )
c
c      Set origin and symbol sizes.
c
       p1x0 = 4.5
       p1y0 = 10.8
       if (np1h*np1v .eq. 1) p1y0 = 13.4
       size = 0.085
c
c-----------------------------------------------------------------------
c---------------------------- E X T R E M A ----------------------------
c-----------------------------------------------------------------------
c
c      Set the domain, "xmin" and "xmax".  
c
       xmin = xmn
       xmax = xmx
       if (xmn .ge. xmx) then
         xmin = xvar(ibeg)
         xmax = xvar(iend)
       endif
c
c      If a log x-axis is desired (lnlg=3,4), make sure xvar > 0.  If
c  not, reset lnlg to give a linear x-axis and issue a warning.
c
       if (lnlg .ge. 3) then
         if (xmin .le. 0.0) then
           lnlg = lnlg - 2
           write (6, 2010)
         endif
       endif
c
c      Determine the actual x-limits of the plot ("ista" and "ifin").
c
       ista = ibeg
       ifin = iend
       do 20 i=ibeg,iend
         if ( xvar(i) .lt. xmin ) ista = i + 1
         if ( xvar(i) .le. xmax ) ifin = i
20     continue
c
c      Set the range of the frame.  Four measures of the extrema are 
c  considered:
c
c  1.  ymn, ymx    user-input extrema to be used for "yvar", and take
c                  precedent over all other evaluations.
c  2.  ymaxg,      global (ibeg:iend) extrema for "yvar", used below for
c        yming     detecting odd poles.
c  3.  ymaxl,      local (ista:iend) extrema for "yvar", used to compute
c        yminl     ymin and ymax.
c  4.  ymin, ymax  set to ymn, ymx if imm .le. 0, or the extrema of 
c                  ymaxl and yminl if imm = 1; used by GRFXAXES to 
c                  determine the scale for the ordinate.
c
       ymaxg =-huge
       yming = huge
       do 30 i=ibeg,iend
         ymaxg = max ( ymaxg, yvar(i) )
         yming = min ( yming, yvar(i) )
30     continue
       ymaxl =-huge
       yminl = huge
       do 40 i=ista,ifin
         if (yvar(i) .lt. huge) ymaxl = max ( ymaxl, yvar(i) )
         if (yvar(i) .gt.-huge) yminl = min ( yminl, yvar(i) )
40     continue
       aymaxl = max ( abs(yminl), abs(ymaxl) )
c
c      Set the range, "ymin" and "ymax".  
c
       if ( (imm .le. 0) .and. (ymx .gt. ymn) ) then
         ymin = ymn
         ymax = ymx
       else
         ymax = max (-huge, ymaxl )
         ymin = min ( huge, yminl )
       endif
c
c      Adjust extrema of range if both are zero, or if they differ by 
c  less than one part in 10**5.  NB: SIGN function is used (instead of 
c  testing ymin*ymax directly) in case ymin*ymax is larger than 1.0e38, 
c  thus triggering an overflow NaN.
c
       if ( (ymin .eq. 0.0) .and. (ymax .eq. 0.0) ) then
         ymin =-1.0
         ymax = 1.0
       else
         q1 = sign ( 1.0, ymin ) * sign ( 1.0, ymax )
         if (q1 .ge. 0.0) then
           q2 = 2.0 * ( ymax - ymin ) / ( ymax + ymin )
           if (abs(q2) .le. 1.0e-5) then
             if (ymax .gt. 0.0) then
               ymax = 1.005 * ymax
               ymin = 0.995 * ymin
             else
               ymax = 0.995 * ymax
               ymin = 1.005 * ymin
             endif
           endif
         endif
       endif
c
c      If a log y-axis is desired (lnlg=2,4), make sure yvar > 0.  If
c  ymin < 0 or if ymin = ymax = 0, reset "lnlg" to give a linear y-axis 
c  and issue a warning.  If ymin = 0 and ymax > 0, reset "ymin" and all 
c  zero-values of "yvar" (if any) to the smallest non-zero value in 
c  "yvar" ("ymingt0"), leave "lnlg" alone, and issue a warning.
c
       if ( (lnlg .eq. 2) .or. (lnlg .eq. 4) ) then
c
         if (ymin .lt. 0.0) then
           lnlg = lnlg - 1
           write (6, 2020)
         endif
c
         if (ymin .eq. 0.0) then
           lmax0 = .false.
           if (aymaxl .eq. 0.0) lmax0 = .true.
           if (lmax0) then
             lnlg = lnlg - 1
             write (6, 2021)
           else
             ymingt0 = huge
             do 90 i=ista,ifin
               if (yvar(i) .gt. 0.0) ymingt0 = min ( ymingt0, yvar(i) )
90           continue
             do 110 i=ista,ifin
               yvar(i) = max ( ymingt0, yvar(i) )
110          continue
             ymin = ymingt0
             write (6, 2030)
           endif
         endif
c
       endif
c
c-----------------------------------------------------------------------
c------------------------------- A X E S -------------------------------
c-----------------------------------------------------------------------
c
       call grfxaxes ( yexp, xlab, ylab, imm, xmin, xmax, ymin, ymax
     1               , aspect, p1x0, p1y0, allxl, allyl, np1h, ip1h
     2               , np1v, ip1v, lnlg )
c
c      Apply "scale" computed in GRFXAXES to font and symbol sizes.
c
       size = size * scale
c
c-----------------------------------------------------------------------
c-------------------------- P L O T   D A T A --------------------------
c-----------------------------------------------------------------------
c
       nzi = ifin - ista + 1
c
       call plot ( wxmin, wymin, 3 )
c
c      Set scaling factors "ffx" and "ffy".
c
       ffx = ( wxmax - wxmin ) / ( xxmax - xxmin )
       ffy = ( wymax - wymin ) / ( yymax - yymin )
       if (lnlg .ge. 3)
     1   ffx = ( wxmax - wxmin ) / log10 ( xxmax / xxmin )
       if ( (lnlg .eq. 2) .or. (lnlg .eq. 4) )
     1   ffy = ( wymax - wymin ) / log10 ( yymax / yymin )
c
c      Bounds to keep plots inside frame windows.
c
       eksmin = wxmin
       eksmax = wxmin + ffx * ( xxmax - xxmin )
       whymin = wymin
       whymax = wymin + ffy * ( yymax - yymin )
       if (lnlg .ge. 3)
     1   eksmax = wxmin + ffx * log10 ( xxmax / xxmin )
       if ( (lnlg .eq. 2) .or. (lnlg .eq. 4) )
     1   whymax = wymin + ffy * log10 ( yymax / yymin )
c
c       Fill array "eks" with scaled independent variable.  "ex" is a 
c  permanent copy of the abscissa should "eks" have to be adjusted to 
c  avoid a pole, start the graph on a border, etc.
c 
       do 220 i=1,nzi
         eye    = i + ista - 1
         eks(i) = wxmin + ffx * ( xvar(eye) - xxmin )
         if ( (lnlg .eq. 3) .or. (lnlg .eq. 4) )
     1     eks(i) = wxmin + ffx * log10 ( xvar(eye) / xxmin )
         ex (i) = eks(i)
220    continue
c
c       Fill array "why" with scaled dependent variable(s), and reset 
c  the abscissa.
c 
       do 250 i=1,nzi
         eye    = i + ista - 1
         why(i) = yvar(eye)
         if (why(i) .gt.-huge) then
           why(i) = wymin + ffy * ( yvar(eye) - yymin )
           if ( (lnlg .eq. 2) .or. (lnlg .eq. 4) )
     1       why(i) = wymin + ffy * log10 ( yvar(eye) / yymin )
         endif
         eks(i) = ex(i)
250    continue
c
c      Set colour.
c
       call setcolr ( 0.2, 0.2, 0.2 )
c
c      Draw plots with one symbol per datum, wikth the symbol type set 
c  by "corl".  Only points between "xmin" and "xmax" (by virtue of 
c  running "i" from "ista" to "ifin") and between "ymin" and "ymax" are 
c  plotted.
c
       if ( (corl .ge.-6) .and. (corl .le. 1) ) then
         do 260 i=1,nzi
           if ( ( eks(i) .le. eksmax ) .and. 
     1          ( eks(i) .ge. eksmin ) .and. 
     2          ( why(i) .le. whymax ) .and. 
     3          ( why(i) .ge. whymin ) ) then
             if (corl .eq. 1) call ocircle  ( eks(i), why(i), size )
             if (corl .eq. 0) call times    ( eks(i), why(i), size )
             if (corl .eq.-1) call plus     ( eks(i), why(i), size )
             if (corl .eq.-2) call triangle ( eks(i), why(i), size )
             if (corl .eq.-3) call square   ( eks(i), why(i), size )
             if (corl .eq.-4) call asterisk ( eks(i), why(i), size )
             if (corl .eq.-5) call star     ( eks(i), why(i), size )
             if (corl .eq.-6) call diamond  ( eks(i), why(i), size )
           endif
260      continue
       endif
c
c      Variables are plotted with lines.  Set line weight and dash 
c  pattern.
c
       if (corl .eq. 2) then
         call setlw ( 0.03 * penwt )
         ndash = 0
c
c      Scan the independent variable to make sure it stays within the
c  limits "whymax" and "whymin".  An effort is made to avoid drawing 
c  curves through "odd poles" (where the functional value goes from +/- 
c  infinity to -/+ infinity).
c
c  "i1"     is the beginning index of the next line segment to be drawn.
c  "delwhy" is a measure of the greatest jump in why(i) that might be 
c           expected without it being an odd pole.  There is no need to 
c           prepare "delwhy" for a possible log plot as an odd pole 
c           would already have reset the log-y request to a linear-y.
c  "lastpt" keeps track of whether why(ip1) is within the frame and is
c           used to determine if a line segment should be drawn.
c  
         i1     = 1
         delwhy = ( min ( ymax, ymaxg )
     1            - max ( ymin, yming ) ) * ffy * 0.8
         dwhyll = 0.0
         dwhyl  = 0.0
         dwhy   = why(i1+1) - why(i1  )
         dwhyr  = why(i1+2) - why(i1+1)
         lastpt = .false.
         if ( (why(i1) .ge. whymin) .and. 
     1        (why(i1) .le. whymax) ) lastpt = .true.
         do 270 i=1,nzi-1
           ip1    = i + 1
           ip2    = min ( i+2, nzi )
           ip3    = min ( i+3, nzi )
           dwhyrr = why(ip3) - why(ip2)
           deks   = eks(ip1) - eks(i  )
c
c      Note from DAC, 7/15: last two criteria added to help prevent 
c  points flagged not to be plotted (by setting why=-huge in calling 
c  routine) from being mistaken for an odd pole.  Presumably, "why" 
c  never gets to -huge, even for an odd pole.
c
           if ( ( dwhyll*dwhyl    .gt. 0.0    ) .and.
     1          ( dwhyl *dwhy     .lt. 0.0    ) .and. 
     2          ( dwhy  *dwhyr    .lt. 0.0    ) .and.
     3          ( dwhyr *dwhyrr   .gt. 0.0    ) .and.
     3          ( why(i)*why(ip1) .lt. 0.0    ) .and.
     4          ( abs(dwhy)       .gt. delwhy ) .and.
     3          ( why(i)          .gt.-huge   ) .and.
     1          ( why(ip1)        .gt.-huge   ) ) then
c
c      An odd pole has been detected.  If a line segment is pending
c  (e.g., lastpt .eq. .true.), draw the current segment up to the left
c  of the pole, and set the starting point for the next line segment.
c
             if ( lastpt ) then
               q1 = deks / ( dwhy + tiny )
               if ( why(i) .gt. why(ip1) ) then
                 eks(i  ) = eks(i) + ( whymax - why(i) ) * q1 
                 why(i  ) = whymax
                 eks(ip1) = eks(i) + ( whymin - why(i) ) * q1
                 why(ip1) = whymin
               else
                 eks(i  ) = eks(i) + ( whymin - why(i) ) * q1 
                 why(i  ) = whymin
                 eks(ip1) = eks(i) + ( whymax - why(i) ) * q1 
                 why(ip1) = whymax
               endif
               call dshcrv ( eks(i1), why(i1), ip1-i1, ndash, 0.0 )
               i1 = ip1
             endif
c
           else
c
c      If why(i) and why(ip1) do not straddle an odd pole, perform the 
c  tests necessary to keep the plot inside the frame window.
c
c  1.  If "why" enters the frame from below, interpolate eks(i) to the 
c      entry point, and reset the initial point of the current segment.
c
c  Note from DAC, 7/15: the if-test on -huge here and in case 3 below is
c  the mechanism introduced to allow some points not to be plotted at 
c  all.  Such points are set to -1.0e32 in the calling routine, and are 
c  then skipped altogether (as opposed to being interpolated to the plot 
c  border for points outside the frame).
c
             if ( (why(i  ) .lt. whymin) .and. 
     1            (why(ip1) .gt. whymin) ) then
               eks(i) = eks(i) + ( whymin - why(i) ) * deks / dwhy 
               if (why(i) .le.-huge) then
                 why(i) = why(ip1)
               else
                 why(i) = whymin
               endif
               i1 = i
             endif
c
c  2.  If "why" enters the frame from above, interpolate eks(i) to the 
c      entry point, and reset the initial point of the current segment.
c
             if ( (why(i  ) .gt. whymax) .and. 
     1            (why(ip1) .lt. whymax) ) then
               eks(i) = eks(i) + ( whymax - why(i) ) * deks / dwhy 
               why(i) = whymax
               i1     = i
             endif
c
c  3.  If "why" falls below the frame, interpolate eks(ip1) to the exit 
c      point, draw the current segment to the exit point, and set the 
c      starting point for the next segment.
c
             if ( (why(i  ) .gt. whymin) .and. 
     1            (why(ip1) .lt. whymin) ) then
               eksip1   = eks(ip1)
               whyip1   = why(ip1)
               eks(ip1) = eks(i) + ( whymin - why(i) ) * deks / dwhy 
               if (why(ip1) .le.-huge) then
                 why(ip1) = why(i)
               else
                 why(ip1) = whymin
               endif
               call dshcrv ( eks(i1), why(i1), i-i1+2, ndash, 0.0 )
               eks(ip1) = eksip1
               why(ip1) = whyip1
               i1       = ip1
             endif
c
c  4.  If "why" rises above the frame, interpolate eks(ip1) to the exit 
c      point, draw the current segment to the exit point, and set the 
c      starting point for the next segment.
c
             if ( (why(i  ) .lt. whymax) .and. 
     1            (why(ip1) .gt. whymax) ) then
               eksip1   = eks(ip1)
               whyip1   = why(ip1)
               eks(ip1) = eks(i) + ( whymax - why(i) ) * deks / dwhy 
               why(ip1) = whymax
               call dshcrv ( eks(i1), why(i1), i-i1+2, ndash, 0.0 )
               eks(ip1) = eksip1
               why(ip1) = whyip1
               i1       = ip1
             endif
c
           endif
c
c      Update "lastpt".
c
           lastpt = .false.
           if ( (why(ip1) .ge. whymin) .and. 
     1          (why(ip1) .le. whymax) ) lastpt = .true.
c
c      Advance "dwhyll" to "dwhyl", etc., before next iteration.
c
           dwhyll = dwhyl
           dwhyl  = dwhy
           dwhy   = dwhyr
           dwhyr  = dwhyrr
270      continue
c
c      Draw last segment, if necessary.
c
         if ( lastpt ) then
           call dshcrv ( eks(i1), why(i1), nzi-i1+1, ndash, 0.0 )
         endif
c
       endif
c
c      Restore pen weight and colour before completing graph.
c
       call setlw   ( 0.03 * penwt )
       call setcolr ( 0.0, 0.0, 0.0 )
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-----------------------------------------------------------------------
c
2010   format('GRFX1D  : |---WARNING---> The x-axis includes the '
     1       ,'origin.  Log option switched',/
     2       ,'GRFX1D  : to linear.')
2020   format('GRFX1D  : |---WARNING---> At least one dependent '
     1       ,'variable includes negative',/
     2       ,'GRFX1D  : values.  Log option switched to linear.')
2021   format('GRFX1D  : |---WARNING---> At least one dependent '
     1       ,'variable includes no positive',/
     2       ,'GRFX1D  : values.  Log option switched to linear.')
2030   format('GRFX1D  : |---WARNING---> The y-axis includes the '
     1       ,'origin.  Zeroes reset to ',/
     2       ,'GRFX1D  : minimum non-zero value; log option retained.')
c
       return
       end
c
c=======================================================================
c=======================================================================
c
       subroutine grfxcls
c
c    dac:grfx03.grfxcls <---------------------------- closes a plot file
c                                                             june, 1990
c
c    written by: David Clarke
c    modified 1: 
c
c  PURPOSE: This subroutine closes a PSPLOT plot file and should be 
c  called after each call to GRFX1D and before the next plot is begun.
c
c  INPUT VARIABLES:
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    PLOT     )---- PSPLOT routine
c
c-----------------------------------------------------------------------
c
       implicit      none
c
c      External statements
c
       external      plot
c
c-----------------------------------------------------------------------
c
       call plot ( 0.0, 0.0, 999 )
c
       return
       end
c
c=======================================================================
c=======================================================================
c
       subroutine grfxopn ( fname )
c
c    dac:grfx03.grfxopn <----------------------------- opens a plot file
c                                                             june, 1990
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE: This subroutine opens a PSPLOT plot file with the specified 
c  file name.  It should be called before GRFX1D is called.
c
c  INPUT VARIABLES:
c    fname       file name of plot file as set by calling routine
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    NEWDEV  )---- PSPLOT routines
c    PSINIT  )
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*(*) fname   
c
c      External statements
c
       external      newdev  , psinit
c
c-----------------------------------------------------------------------
c
       call newdev ( fname )
       call psinit ( .true. )
c
       return
       end
c
c=======================================================================
c=======================================================================
c
       subroutine hdrftr ( phdr, iftr, filename )
c
c    dac:1dplot.hdrftr <------------------ writes header/footer to graph
c                                                             june, 2018
c
c    written by: David Clarke
c    modified 1: 
c
c  PURPOSE: This subroutine writes the header (title) "phdr" to the top-
c  centre of the graph.  If desired, a footer with the filename and the 
c  date and time of creation is created and written at the bottom left 
c  of the graph.
c
c  INPUT VARIABLES:
c    phdr        header (title) to be written to graph.
c                blank => no header
c    iftr        =1 => footer created and written to graph
c                =0 => not
c    filename    used to create footer
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    SETLW            )  PSPLOT routine
c    STRLEN
c    PSCHARS
c    TIMEDATE
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*(*) phdr    , filename
       character*8   chtime
       character*10  chdate
       character*80  pftr
       integer       iftr    , nch   
       real          pt12    , pt20    , x1      , y1      
c
c      External statements
c
       integer       strlen
       external      setlw   , strlen  , pschars , timedate
c
c-----------------------------------------------------------------------
c
c      Set line weight and font sizes.
c
       call setlw ( 0.03 )
       pt12 = 0.264
       pt20 = 0.440
c
c      Write user-specified title, if desired.
c
       nch  = strlen ( phdr )
       if (nch .gt. 0) then
         x1 = 9.5
         y1 = 24.3
         call pschars ( x1, y1, pt20, phdr(1:nch), 0.0, 1 )
       endif
c
c      Create and write footer, if desired.
c
       if (iftr .eq. 1) then
         call timedate ( chtime, chdate )
         nch = strlen ( filename )
         write ( pftr, 2010 ) filename(1:nch), chdate, chtime
         nch = strlen ( pftr )
         x1  = 2.5 
         y1  = 7.0
         call pschars ( x1, y1, pt12, pftr(1:nch), 0.0, 0 )
       endif
c
c-----------------------------------------------------------------------
c
2010   format('Plot file `',a,''' created on ',a10,' at ',a8,'.')
c
       return
       end
c
c=======================================================================
c
c>>>>>>>>>>>>>>>>>>>>> INTERNALLY CALLED ROUTINES <<<<<<<<<<<<<<<<<<<<<<
c
c++++++++++++++++++++++++ GRFX03 (DAC) ROUTINES ++++++++++++++++++++++++
c
c=======================================================================
c
       subroutine axislab ( xory, wxmin, wxmax, wymin, wymax, scale
     1                    , wconv, nchmax, clab, fact, iexp )
c
c    dac:grfx03.axislab <----------------------------------- labels axis
c                                                             june, 1996
c
c    written by: David Clarke
c    modified 1: December 1996, by David Clarke; added multiple plot per
c                frame capability.
c    modified 2: June, 2003, by David Clarke; generalised for use by 2-D
c                plotter as well.
c    modified 3: October 2003 by Alexander Men'shchikov; generalized to
c                include the PSPLOT graphics library.
c    modified 5: June, 2014 by DAC; introduced "fact" to scale down 
c                annotation sizes for long, short plots.
c
c  PURPOSE: This subroutine labels the axis.
c
c  INPUT VARIABLES:
c    xory        =1 => label the x-axis
c                =2 => label the y-axis
c    wxmin       minimum x-window coordinate
c    wxmax       maximum x-window coordinate
c    wymin       minimum y-window coordinate
c    wymax       maximum y-window coordinate
c    scale       scaling factor for multiple plots per frame
c    wconv       window conversion factor.  1-D plots are done in window
c                coordinates, and thus wconv = 1.0.  2-D plots are done
c                in grid coordinates, and thus wconv = width of window
c                in grid units.
c    nchmax      maximum number of characters in y-annotation (needed
c                so the y-axis label won't overwrite the annotations).
c    clab        character string containing the label.  Should be
c                written in `NCAR markup language'.
c    fact        scale factor of annotation size for long, short plots
c    iexp        .eq. 0 => ignored
c                .ne. 0 => "iexp" is the exponent of the annotations
c                          expressed in scientific notation and is
c                          incorporated into the axis label.
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    STRLEN
c    PSCHARS
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*(*) clab
       character*34  clab2
       integer       xory    , iexp    , nchmax  , nch     , nch2
     1             , aexp
       real          xpos    , ypos    , angle   , wxmin   , wxmax
     1             , wymin   , wymax   , scale   , wlab    , wlab2
     2             , fnt1    , fnt2    , wconv   , q1      , q2
     3             , fact    , scfa    , scfa2
c
c      External statements
c
       integer       strlen
       external      strlen  , pschars
c
c-----------------------------------------------------------------------
c
       scfa  = scale * fact
       scfa2 = scale * fact**0.8
       fnt1  = 0.5 * scfa
       fnt2  = 0.4 * scfa
       q1    = wconv * scfa
       q2    = 0.02  * q1
c
c      Determine the number of characters in "clab".
c
       nch  = strlen ( clab )
       if (nch .eq. 0) return
c
c      If necessary, create a label for the axis scaling factor.
c
       wlab2 = 0.0
       if (iexp .ne. 0) then
         iexp = min ( 999, max ( -999, iexp ) )
         nch2 = 17
         clab2(1:nch2) = ' (|F18|U|F0|10|S|'
         if (iexp .lt. 0) then
           clab2(nch2+1:nch2+10) = '|F18|-|F0|'
           nch2 = nch2 + 10
         endif
         aexp = abs ( iexp )
         if (aexp .lt. 10) then
           write ( clab2(nch2+1:nch2+1), 2010 ) aexp
           nch2 = nch2 + 1
         endif
         if ( ( (aexp .ge. 10) .and. (aexp .lt. 99) ) ) then
           write ( clab2(nch2+1:nch2+2), 2020 ) aexp
           nch2 = nch2 + 2
         endif
         if (aexp .ge. 100) then
           write ( clab2(nch2+1:nch2+3), 2030 ) aexp
           nch2 = nch2 + 3
         endif
         clab2(nch2+1:nch2+4) = '|N|)'
         nch2 = nch2 + 4
       endif
c
c      Set coordinates and orientation of label, depending on whether
c  this is the x- or y-axis, and depending on whether the exponent of
c  the annotations has to be incorporated into the axis label.
c
       if (iexp .ne. 0.0) wlab2 = 4.5 * fnt2
       if (xory .eq. 1) then
         angle = 0.0
         xpos  = 0.5 * ( wxmax + wxmin - wlab2 ) + 0.1 * scfa
         ypos  = wymin - 1.6 * scfa2
       else
         angle = 90.0
         xpos  = wxmin - 0.75 * fnt1 * real ( nchmax ) - 0.15 * scale
         ypos  = 0.5 * ( wymax + wymin - wlab2 ) + 0.1 * scfa
       endif
c
c      Write the axis label.
c
       call pschars ( xpos, ypos, fnt1, clab(1:nch), angle, 1 )
       if (iexp .ne. 0) then
         call pschars ( 999.0, 999.0, fnt2, clab2(1:nch2), angle, 1)
       endif
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-----------------------------------------------------------------------
c
2010   format ( i1 )
2020   format ( i2 )
2030   format ( i3 )
c
       return
       end
c
c=======================================================================
c=======================================================================
c
       subroutine axisset1 ( xory, qmin, qmax, imm, aspect, ndec, iexp
     1                     , nchq, qqmin, qqmax, nqdiv, nqann, qann
     2                     , cqann)
c
c    dac:grfx03.axisset1 <------------------- parameters for linear axis
c                                                             june, 1996
c
c    written by: David Clarke
c    modified 2: July, 2010 by DAC; modified to give proper minus signs 
c                for PSPLOT, and upgraded over-crowding algorithm to 
c                take into account number of characters per annotation.
c
c  PURPOSE: This subroutine computes the parameters needed to draw a
c  linear axis.  Parameters for logarithmic axes are computed by
c  AXISSET2.  Constraints include:
c
c      1. annotation and tick spacings chosen so they are not crowded.
c      2. annotations will always be multiples of 1, 2, or 5.
c      3. annotations will be expressed in scientific notation if the
c         absolute value of the annotation is .le. 0.001 or .ge. 1000.
c      4. new extrema are chosen to be multiples of the tick spacing.
c      5. Axes are not extended to fall exactly on an annotated tick.
c      6. y-axes are extended by one minor tick for imm=1,-1.
c
c  INPUT VARIABLES:
c    xory        = 1 => x-axis
c                = 2 => y-axis
c                = 3 => polar axis
c    qmin        minimum value on axis.
c    qmax        maximum value on axis.
c    imm         = 1,-1 =>     extend grid by one minor tick mark
c                = 0 => do not extend grid by one minor tick mark
c    aspect      desired aspect ratio of axes.
c                < 1 => short, wide plot
c                = 1 => square plot
c                > 1 => tall, narrow plot
c    ndec        the number of significant figures to appear to the
c                right of the decimal in the annotations.
c                =-1 => minimum number of decimal places required to
c                       make each annotation unique.
c
c  OUTPUT VARIABLES:
c    iexp        If the absolute annotation values are .le. 0.001 or
c                .ge. 1000., "iexp" is the exponent of the annotations
c                expressed in scientific notation.  Otherwise, iexp = 0.
c    nchq        maximum number of characters in any annotation.
c    qqmin       greatest multiple of the major tick spacing < "qmin".
c    qqmax       smallest multiple of the major tick spacing > "qmax".
c    nqdiv       total number of divisions along axis (always > 0).
c    nqann       total number of annotations along axis.
c    qann        1-D real array containing the annotation locations
c                along the axis.
c    cqann       1-D character array (at least character*21) containing 
c                the axis annotations in NCAR markup language.
c
c  LOCAL VARIABLES:
c    ccqann      temporary 1-D character array.
c    mant        1-D integer array containing int(qann).
c    idec        1-D integer array containing the decimal portion of
c                "qann".
c
c  EXTERNALS: [NONE]
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       integer       mann
       parameter     ( mann=20 )
c
       integer       xory    , ndec    , iexp    , nchq    , nqdiv
     1             , nqann   , i1      , i       , j       , jj
     2             , nmax    , imm     , ndeco   , nchars
       real          qmin    , qmax    , aspect  , qqmin   , qqmax
     1             , fact    , dq1     , dq2     , dq3     , dq4
     2             , dq5     , dq6     , qamin   , qamax   , sqmin
     3             , sqmax   , sqqmin  , sqqmax  , aqamax  , q1
     4             , q2      , qdec    , tick
c
       character*(*) cqann   (   *)
       character*16  ccqann  (mann)
       integer       mant    (mann), idec    (mann)
       real          qann    (   *)
c
c-----------------------------------------------------------------------
c
       ndeco = ndec
       if (xory .eq. 1) fact = 0.1 * max ( 1.0, aspect )
       if (xory .eq. 2) fact = 0.1 / min ( 1.0, aspect )
       if (xory .eq. 3) fact = 0.1 * max ( 1.0, aspect )
c
c      dq1    is the tentative spacing between annotations.
c      i1     is chosen so that 0.1 .lt. dq2 = dq1 * 10**(-i1) .le. 1.0
c      dq3    is the spacing between annotations (modulo a factor
c             10**i1)
c
       dq1   = ( qmax - qmin ) * fact
       i1    = int ( alog10 ( dq1 ) )
       if (dq1 .gt. 1.0) i1 = i1 + 1
       dq2   = dq1 * 10.0**(-i1)
       if (dq2 .le. 0.1) then
         i1  = i1 - 1
         dq2 = dq2 * 10.0
       endif
       dq3   = 1.0
       if ( (dq2 .gt. 0.2) .and. (dq2 .le. 0.5) ) dq3 = 0.5
       if ( (dq2 .gt. 0.1) .and. (dq2 .le. 0.2) ) dq3 = 0.2
c
c      dq4    is the final spacing between annotations (dq3 * 10**i1)
c      dq5    is the final spacing between ticks
c      qqmin  is the greatest multiple of dq5 less    than  qmin
c      qamin  is the smallest multiple of dq4 greater than qqmin
c      qqmax  is the smallest multiple of dq5 greater than  qmax
c      qamax  is the greatest multiple of dq4 less    than qqmax
c
10     continue
       dq4    = dq3 * 10.0**i1
       tick   = 5.0
       if ( dq3 .eq. 0.2 ) tick = 4.0
       dq5    = dq4 / tick
c
       sqmin  = sign ( 0.5 , qmin )
       sqmax  = sign ( 0.5 , qmax )
       qqmin  = real ( int ( qmin / dq5 - 0.499 + sqmin ) ) * dq5
       qqmax  = real ( int ( qmax / dq5 + 0.499 + sqmax ) ) * dq5
       if ( (xory .eq. 2) .and. (imm .ne. 0) ) then
         if ( qmin-qqmin .lt. 0.5*dq5) qqmin = qqmin - dq5
         if (qqmax- qmax .lt. 0.5*dq5) qqmax = qqmax + dq5
       endif
       if (xory .eq. 3) then
         if ( (qqmin .gt.  0.0) .and. (qqmin .lt.  5.0) ) qqmin =  0.0
         if ( (qqmax .gt. 85.0) .and. (qqmax .lt. 90.0) ) qqmax = 90.0
       endif
       sqqmin = sign ( 0.5, qqmin )
       sqqmax = sign ( 0.5, qqmax )
       qamin  = real ( int ( qqmin / dq4 + 0.499 + sqqmin ) ) * dq4
       qamax  = real ( int ( qqmax / dq4 - 0.499 + sqqmax ) ) * dq4
c
c      nqann  is the number of annotations along the axis
c      nqdiv  is the number of divisions along the axis (always > 0)
c
c      Note: if nqann > nmax, the value of dq3 is increased and
c  execution is returned to statement 10 above so that nqann is .le.
c  nmax.
c
       nqann = int ( ( qamax - qamin ) / dq4 + 0.01 ) + 1
       nmax  = int ( 8.0 / max ( 1.0, aspect ) + 0.5 )
       if (xory .eq. 2)
     1   nmax = int ( 8.0 * min ( 1.0, aspect ) + 0.5 )
       if (nqann .gt. nmax) then
         if (dq3 .eq. 0.2) then
           dq3 = 0.5
           go to 10
         endif
         if (dq3 .eq. 0.5) then
           dq3 = 1.0
           go to 10
         endif
         if (dq3 .eq. 1.0) then
           dq3 = 0.2
           i1 = i1 + 1
           go to 10
         endif
       endif
       nqdiv = int ( ( qqmax - qqmin ) / dq5 + 0.01 )
c
c      If the largest absolute value at a major tick mark (aqmax) is
c  greater than 1000 or less than 0.01, the axis values are printed in
c  scientific notation where the mantissa is written next to the major
c  tick mark and the exponent (q1 = 10**iexp) is given as part of the
c  axis label.
c      q2     is the reciprocal of q1.
c      dq6    is the spacing between annotations, modulo q2
c
       aqamax = max ( abs(qamax), abs(qamin) )
       iexp  = 0
       if (aqamax .ge. 1000.) iexp = int ( alog10 ( aqamax ) )
       if (aqamax .lt. 0.01 ) iexp = int ( alog10 ( 0.1001 * aqamax ) )
       q1    = 10.0**iexp
       q2    = 10.0**(-iexp)
       dq6   = dq4 * q2
c
c      ndec   is the number of decimal places to appear in the mantissa.
c
c      If ndec=-1, then the routine automatically selects ndec to the
c  minimum number required for all annotations along the axis to be
c  different.  e.g., for the values 1.210, 1.220, 1.230, 1.240,
c  1.250, ndec should be 2.  If it were 1, then all values would appear
c  as 1.2.
c      If the automatic routine doesn't work for some pathelogical case,
c  or if the user requires additional zeroes after the last decimal
c  place chosen by the routine, then specifying ndec will override the
c  automatic routine.  Note that ndec can be no larger than 6.
c
       if (ndeco .eq. -1) ndec =-int ( alog10 ( dq6 ) - 0.7 )
       ndec = max ( 0, min ( 6, ndec ) )
       if ( (aqamax .lt. 20.0) .and. (ndec .eq. 0) ) ndec = 1
c
c      qann   is a real 1-D array with the annotations for the axis.
c
       qann(1) = qamin * q2
       do 20 i=2,nqann
         qann(i) = qann(i-1) + dq6
20     continue
c
c      The annotations are converted to character strings with all the
c  machine round-off noise cleaned out.  This is done by separating the
c  portions to the left (mant) and right (idec) of the decimal point,
c  then formatting the character string according to the value of
c  "ndec".  Loop 40 squeezes out extraneous blanks.
c
       do 30 i=1,nqann
         mant(i) = int ( abs ( qann(i) ) + 0.000005 )
         qdec    = abs ( qann(i) ) - real ( mant(i) )
         idec(i) = int ( 0.1 + 10.0**ndec * qdec )
         qann(i) = qann(i) * q1
30     continue
       nchq = 1
       do 50 i=1,nqann
         if (ndec .eq. 0) write (ccqann(i),2010) mant(i)
         if (ndec .eq. 1) write (ccqann(i),2020) mant(i), idec(i)
         if (ndec .eq. 2) write (ccqann(i),2030) mant(i), idec(i)
         if (ndec .eq. 3) write (ccqann(i),2040) mant(i), idec(i)
         if (ndec .eq. 4) write (ccqann(i),2050) mant(i), idec(i)
         if (ndec .eq. 5) write (ccqann(i),2060) mant(i), idec(i)
         if (ndec .eq. 6) write (ccqann(i),2070) mant(i), idec(i)
         cqann(i) = ' '
         jj = 0
         if (qann(i) .lt. 0.0) then
           cqann(i)(1:10) = '|F18|-|F0|'
           jj = 10
         endif
         do 40 j=1,16
           if (ccqann(i)(j:j) .ne. ' ') then
             jj = jj + 1
             cqann(i)(jj:jj) = ccqann(i)(j:j)
           endif
40       continue
         if (qann(i) .lt. 0.0) jj = jj - 9
         nchq = max ( nchq, jj )
50     continue
c
c      To avoid x-axis crowding, count the total number of characters 
c  (excluding decimal points) to be printed along the x-axis.  If this
c  is greater than 30, reset "dq3" and return to statement 10 to try
c  again.
c
       nchars = nqann * ( nchq - 1 )
       if ( (xory .eq. 1) .and. (nchars .gt. 30) ) then
         if (dq3 .eq. 0.2) then
           dq3 = 0.5
           go to 10
         endif
         if (dq3 .eq. 0.5) then
           dq3 = 1.0
           go to 10
         endif
         if (dq3 .eq. 1.0) then
           dq3 = 0.2
           i1  = i1 + 1
           go to 10
         endif
       endif
c
c      "nchq" is returned so that the y-axis label may be positioned 
c  just left of the left-most character in the annotations.
c
       nchq = nchq + 1
       if (ndec .eq. 0) nchq = nchq + 1
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-----------------------------------------------------------------------
c
2010   format(i4,12x)
2020   format(i4,'.',i1.1,10x)
2030   format(i4,'.',i2.2, 9x)
2040   format(i4,'.',i3.3, 8x)
2050   format(i4,'.',i4.4, 7x)
2060   format(i4,'.',i5.5, 6x)
2070   format(i4,'.',i6.6, 5x)
c
       return
       end
c
c=======================================================================
c=======================================================================
c
       subroutine axisset2 ( xory, qmin, qmax, yexp, aspect, nchq, qqmin
     1                     , qqmax, nqdiv, nqann, qann, cqann )
c
c    dac:grfx03.axisset2 <-------------- parameters for logarithmic axis
c                                                             july, 1996
c
c    written by: David Clarke
c    modified 2: July, 2010 by DAC; modified to give proper minus signs 
c                for PSPLOT.
c
c  PURPOSE: This subroutine computes the parameters needed to draw a
c  logarithmic axis.  Parameters for linear axes are computed by
c  AXISSET1.  Constraints include:
c
c      1. annotation and tick spacings chosen so they are not crowded.
c      2. annotations will always be powers of 10.
c      3. annotations will be expressed in scientific notation if the
c         absolute value of the annotation is .le. 0.001 or .ge. 1000.
c      4. new extrema are chosen to be multiples of the tick spacing.
c      5. Axes are not extended to fall exactly on an annotated tick.
c
c  INPUT VARIABLES:
c    xory        = 1 => x-axis
c                = 2 => y-axis
c    qmin        minimum value on axis.
c    qmax        maximum value on axis.
c    yexp        log10 of scaling factor generated when y was converted
c                to single precision
c    aspect      desired aspect ratio of axes.
c                < 1 => short, wide plot
c                = 1 => square plot
c                > 1 => tall, narrow plot
c
c  OUTPUT VARIABLES:
c    nchq        maximum number of characters in any annotation.
c    qqmin       location of first tick mark.
c    qqmax       location of last  tick mark.
c    nqdiv       >0 => one less than the total number of decades along
c                      axis.  These will mark location of minor ticks.
c                =0 => minor ticks will mark units within each decade.
c    nqann       total number of annotations along axis.
c    qann        1-D real array containing the annotation locations
c                along the axis.
c    cqann       1-D character array (at least character*21) containing 
c                the axis annotations in NCAR markup language.
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [NONE]
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       integer       mann
       parameter     ( mann=20 )
c
       integer       xory    , yexp    , nchq    , nqdiv   , nqann   
     1             , i1      , i2      , i3      , i4      , i5      
     2             , i6      , i7      , i8      , dnq     , i
     3             , n
       real          qmin    , qmax    , aspect  , qqmin   , qqmax
     1             , fact    , ri5     , ri6     , rdnqi
c
       character*(*) cqann   (   *)
       real          qann    (   *)
c
c-----------------------------------------------------------------------
c
c      i1     is the highest power of ten less than "qmin".
c      i2     is a multiple of 10**i1 less than "qmin".
c      qqmin  is the log of the location of the first tick mark
c
       i1    = int ( alog10 ( qmin * 1.01 ) )
       if (qmin .lt. 1.0) i1 = i1 - 1
       i1    = max ( -32, min ( 32, i1 ) )
       i2    = int ( qmin * 10.0**(-i1) + 0.01 )
       if (i2 .ge. 8) i2 = i2 - 1
       if (i2 .ge. 6) i2 = i2 - 1
       if (i2 .ge. 4) i2 = i2 - 1
       qqmin = alog10 ( real ( i2 ) ) + real ( i1 )
c
c      i3     is the highest power of ten less than "qmax".
c      i4     is a multiple of 10**i3 greater than "qmax".
c      qqmax  is the log of the location of the last tick mark
c
       i3    = int ( alog10 ( qmax * 0.99 ) )
       if (qmax .lt. 1.0) i3 = i3 - 1
       i3    = max ( -32, min ( 32, i3 ) )
       i4    = int ( qmax * 10.0**(-i3) + 1.01 )
       if (i4 .ge. 8) i4 =            i4 + 1
       if (i4 .ge. 6) i4 = min ( 10, i4 + 1 )
       if (i4 .ge. 4) i4 = min ( 10, i4 + 1 )
       qqmax = alog10 ( real ( i4 ) ) + real ( i3 )
c
c      If "qqmin" and "qqmax" fall within the same decade, reset them to
c  the extremes of that decade.
c
       if (i3 .eq. i1) then
         qqmin = real ( i1     )
         qqmax = real ( i1 + 1 )
       endif
c
c      i5     is the smallest integer greater than "qqmin"
c      i6     is the largest  integer less    than "qqmax"
c
       i5    = int ( qqmin - 0.001 )
       if (qqmin .gt. 0.0) i5 = i5 + 1
       i6    = int ( qqmax + 0.001 )
       if (qqmax .lt. 0.0) i6 = i6 - 1
c
c      To avoid over-crowded annotations, only every "dnq"'th decade is
c  annotated.  If there are fewer than 8 or 9 decades between "qqmin"
c  and "qqmax", "dnq" is 1 and minor ticks mark off units within each
c  decade.  If there are more than 8 or 9 decades, "dnq" is greater than
c  1 and minor ticks mark off non-annotated decades.
c
c      i5     is the log of the location of the first annotation.
c      i6     is the log of the location of the last  annotation.
c
       if (xory .eq. 1) fact = max ( 1.0, aspect ) / 8.0
       if (xory .eq. 2) fact = 0.1 / min ( 1.0, aspect )
       dnq = int ( real ( i6 - i5 ) * fact ) + 1
       if (dnq .gt. 1) then
         ri5   = real ( i5 )
         ri6   = real ( i6 )
         rdnqi = 1.0 / real ( dnq )
         i5    = int ( ri5 * rdnqi + sign ( 0.001, ri5 ) ) * dnq
         i6    = int ( ri6 * rdnqi + sign ( 0.999, ri6 ) ) * dnq
         qqmin = real ( i5 )
         qqmax = real ( i6 )
         nqann = ( i6 - i5 ) / dnq + 1
c        nqdiv = i2 - i1
         nqdiv = i6 - i5
       else
         nqann = i6 - i5 + 1
         nqdiv = 0
       endif
c
c      Create real and character annotations for the axis.
c
       i5    = i5 + yexp
       i6    = i6 + yexp
       i7    = i5
       nchq  = 4
       do 10 i=1,nqann
         qann (i) = real ( i7 - yexp )
         cqann(i) = ' '
         i8 = abs ( i7 )
         if (i8 .le. 2) then
           if (i7 .eq. -2) then
             cqann(i)(1:4) = '0.01'
             nchq = max ( nchq, 5 )
           endif
           if (i7 .eq. -1) cqann(i)(1:3) = '0.1'
           if (i7 .eq.  0) cqann(i)(1:3) = '1.0'
           if (i7 .eq.  1) then
             cqann(i)(1:4) = '10.0'
             nchq = max ( nchq, 5 )
           endif
           if (i7 .eq.  2) then
             cqann(i)(1:5) = '100.0'
             nchq = max ( nchq, 6 )
           endif
         else
           cqann(i)(1:5) = '10|S|'
           n = 5
           if (i7 .lt. 0) then
             cqann(i)(6:15) = '|F18|-|F0|'
             n = 15
           endif
           if (i8 .le. 9) write ( cqann(i)(n+1:n+1), '(i1)' ) i8
           if ( (i8 .ge. 10) .and. (i8 .le. 99) )
     1       write ( cqann(i)(n+1:n+2), '(i2)' ) i8
           if (i8 .ge. 100) write ( cqann(i)(n+1:n+3), '(i3)' ) i8
         endif
         i7 = i7 + dnq
10     continue
       if ( (i5 .lt.    0) .or. (i6 .ge.  10) ) nchq = max ( nchq, 5 )
       if ( (i5 .le.  -10) .or. (i6 .ge. 100) ) nchq = 6
       if   (i5 .le. -100)                      nchq = 7
c
       return
       end
c
c=======================================================================
c=======================================================================
c
       subroutine grfxaxes ( yexp, xlab, ylab, imm, xmin, xmax, ymin
     1                     , ymax, aspect, p1x0, p1y0, allxl, allyl
     2                     , np1h, ip1h, np1v, ip1v, lnlg )
c
c    dac:grfx03.grfxaxes <---------------------- draws axes for graphics
c                                                             june, 1996
c
c    written by: David Clarke
c    modified 1: December 1996, by David Clarke; added multiple frame 
c                per page capability.
c    modified 2: October 2003 by Alexander Men'shchikov; generalized to
c                include the PSPLOT graphics library.
c    modified 3: June, 2010 by DAC; moved determination of extrema and 
c                log-plot checks to GRFX1D.
c
c  PURPOSE: This subroutine draws and labels the axes (perimeter style)
c  for 1-D PSPLOT graphics.
c
c  INPUT VARIABLES:
c    yexp        log10 of scaling factor generated when y was converted
c                to single precision
c    xlab        character string for x-axis label (NCAR markup 
c                language).
c    ylab        character string for y-axis label (ditto).
c    imm         = 1,-1 =>     extend grid by one minor tick mark
c                = 0 => do not extend grid by one minor tick mark
c    xmin        minimum value on the x-axis
c    xmax        maximum value on the x-axis
c    ymin        minimum value on the y-axis
c    ymax        maximum value on the y-axis
c    aspect      desired aspect ratio of axes.
c                < 1 => short, wide frame
c                = 1 => square frame
c                > 1 => tall, narrow frame
c    p1x0        x-coordinate of the lower left corner of the frame
c    p1y0        y-coordinate of the lower left corner of the frame
c    allxl       =2 => all x labels/annotations are drawn (default)
c                =1 => only the x-label/annotations on the bottom row
c                      of frames on the page are drawn.
c                =0 => labels aren't even drawn on the bottom row.
c                if allxl < 0, same as abs(allxl) except last annotation
c                is not drawn if it falls at xxmax.
c    allyl       =2 => all y labels/annotations are drawn (default)
c                =1 => only the y-label/annotations on the left column
c                      of frames on the page are drawn.
c                =0 => labels aren't even drawn on the left column.
c                if allyl < 0, same as abs(allyl) except last annotation
c                is not drawn if it falls at yymax.
c    np1h        number of frames to be arranged horizontally per page
c    ip1h        in the current row   , this is the ip1h'th frame
c    np1v        number of frames to be arranged vertically   per page
c    ip1v        in the current column, this is the ip1v'th frame
c    lnlg        = 1 => linear x and y;    = 2 => linear x, log y
c                = 3 => log x, linear y;   = 4 => log x and y
c
c  OUTPUT VARIABLES (via common block win1d):
c    scale       scaling factor for multiple frames per page
c    wxmin, etc. coordinates of plotting window in graphical units
c    xxmin, etc. coordinates of plotting window in "real" units after 
c                adjustments to "xmin", etc. have been made to allow x- 
c                and y-axes to start and end on major/minor tick marks
c
c  LOCAL VARIABLES:
c    cxann       1-D character array containing all x-axis annotations
c                in NCAR markup language.
c    cyann       1-D character array containing all y-axis annotations
c                in NCAR markup language.
c
c  EXTERNALS:
c    AXISSET1, AXISSET2
c    STRLEN
c    PSCHARS
c    AXISLAB
c    SLDLIN     )---- PSPLOT routine
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       integer       mann
       real          small
c       parameter     ( mann=20, small=0.001 )
       parameter     ( mann=20, small=0.05 )
c
       character*(*) xlab    , ylab
       integer       yexp    , imm     , allxl   , allyl   , np1h    
     1             , ip1h    , np1v    , ip1v    , lnlg    
     2             , aallxl  , aallyl  , ixsen   , iysen   
     3             , ndec    , iend    , i       , nch     , iexp    
     3             , nchmax  , ndiv    , nann    , i1      , ibeg
       real          xmin    , xmax    , ymin    , ymax    , aspect 
     1             , p1x0    , p1y0    , phw     , wx      , dwx     
     2             , dwx1    , dwx2    , wy      , dwy     , dwy1    
     3             , dwy2    , fact    , q1      , q2      , q3      
     4             , xx      , dxx     , yy      , dyy
       real          wxmin   , wxmax   , wymin   , wymax   , xxmin
     1             , xxmax   , yymin   , yymax   , scale
c
       character*32  cxann   (mann), cyann   (mann)
       real          xann    (mann), yann    (mann)
c
       common      / win1d /
     1               wxmin   , wxmax   , wymin   , wymax   , xxmin
     2             , xxmax   , yymin   , yymax   , scale
c
c      External statements
c
       integer       strlen
       external      axisset1, axisset2, strlen  , pschars , sldlin  
     1             , axislab
c
c-----------------------------------------------------------------------
c
c      Set axes sentinels.  If i*sen=1, *=x,y, *-label and annotations
c  are drawn.
c
       aallxl = abs ( allxl )
       ixsen  = 0
       if   (aallxl .eq. 2)                           ixsen = 1
       if ( (aallxl .eq. 1) .and. (ip1v  .eq. np1v) ) ixsen = 1
       aallyl = abs ( allyl )
       iysen  = 0
       if   (aallyl .eq. 2)                           iysen = 1
       if ( (aallyl .eq. 1) .and. (ip1h  .eq. 1   ) ) iysen = 1
c
c-----------------------------------------------------------------------
c----------------  W I N D O W   C O O R D I N A T E S  ----------------
c-----------------------------------------------------------------------
c
c      Compute window coordinates ("wxmin", etc.) and the scaling factor 
c  ("scale"), all in common block "win1d" declared also by GRFX1D.  
c
       scale = 1.0
       wxmin = p1x0
       wxmax = p1x0 + 10.0
       wymin = p1y0
       wymax = p1y0 + 10.0
       phw   = 5.0
c
       if (aspect .gt. 1.0) then
         wxmin = wxmin + phw * ( 1.0 - 1.0 / aspect )
         wxmax = wxmax - phw * ( 1.0 - 1.0 / aspect )
       endif
c
       if (aspect .lt. 1.0) then
         wymin = wymin + phw * ( 1.0 - aspect )
         wymax = wymax - phw * ( 1.0 - aspect )
       endif
c
c      Adjust for multiple frames per page.  The factor of 17.0 in 
c  "scale" allows room for the footer to be put at the bottom of the 
c  page in GRFX1D.
c
       dwx = 0.0
       dwy = 0.0
       if (np1h*np1v .gt. 1) then
         dwx   = 3.5
         dwy   = 2.2
         dwx1  = ( wxmax - wxmin + dwx ) * real ( np1h )
         dwy1  = ( wymax - wymin + dwx ) * real ( np1v )
         scale = 17.0 / max ( dwx1, dwy1 )
         if (aallyl .le. 1) then
           dwx  = 0.3
           dwx1 = ( wxmax - wxmin + dwx ) * real ( np1h ) + 3.2
         endif
         if (aallxl .le. 1) then
           dwy  = 0.3
           dwy1 = ( wymax - wymin + dwx ) * real ( np1v ) + 1.9
         endif
         wx    = 12.8
         wy    = 12.5
         dwx1  = ( wxmax - wxmin + dwx ) * scale
         dwx2  = ( wxmax - wxmin       ) * scale
         wxmax = wx - dwx1 * real ( np1h - ip1h ) + p1x0
         dwy1  = ( wymax - wymin + dwy ) * scale
         dwy2  = ( wymax - wymin       ) * scale
         wymax = wy - dwy1 * real ( ip1v - 1    ) + p1y0
         wxmin = wxmax - dwx2
         wymin = wymax - dwy2
c
       endif
c
c-----------------------------------------------------------------------
c------------------------  T H E   X - A X I S  ------------------------
c-----------------------------------------------------------------------
c
c      Set parameters for the x-axis.
c
       ndec =-1
       if (lnlg .le. 2) then
         call axisset1 ( 1, xmin, xmax, imm, aspect, ndec, iexp, nchmax
     1                 , xxmin, xxmax, ndiv, nann, xann, cxann )
       else
         call axisset2 ( 1, xmin, xmax, 0, aspect, nchmax
     1                 , xxmin, xxmax, ndiv, nann, xann, cxann )
         iexp = 0
       endif
c
c      Write the x-axis annotations and draw the annotated ticks.
c
       fact = ( wxmax - wxmin ) / ( xxmax - xxmin )
       wy   = wymin - 0.75  * scale
       q1   = 0.06 * scale
       q2   = 0.45 * scale
       q3   = 0.38 * scale
c
       iend = nann
       if (allxl .lt. 0) then
         q1 =  abs ( xxmax - xann(nann) ) / ( xxmax - xxmin )
         if (q1 .lt. small) iend = nann - 1
       endif
       do 20 i=1,iend
         wx = wxmin + ( xann(i) - xxmin ) * fact - q1
         if (ixsen .eq. 1) then
           nch = strlen ( cxann(i) )
           call pschars ( wx, wy, q2, cxann(i)(1:nch), 0.0, 1 )
         endif
         wx  = wx + q1
         call sldlin ( wx, wymin, wx, wymin+q3, 0.0 )
         call sldlin ( wx, wymax, wx, wymax-q3, 0.0 )
20     continue
c
c      Draw the unannotated ticks.
c
       q1 = 0.25  * scale
       if (ndiv .gt. 0) then
         dxx = ( xxmax - xxmin ) / real ( ndiv )
         do 30 i=2,ndiv
           xx = real ( i - 1 ) * dxx
           wx = wxmin + xx * fact
           call sldlin ( wx, wymin, wx, wymin+q1, 0. )
           call sldlin ( wx, wymax, wx, wymax-q1, 0. )
30       continue
       else
         xx = xxmin
40       continue
         i1 = int ( xx - 0.499 + sign ( 0.5, xx ) )
         xx = alog10 ( 10.0**xx + 10.0**i1 )
         if (xx .lt. xxmax) then
           wx = wxmin + ( xx - xxmin ) * fact
           call sldlin ( wx, wymin, wx, wymin+q1, 0. )
           call sldlin ( wx, wymax, wx, wymax-q1, 0. )
           go to 40
         endif
       endif
c
c      Draw and label the axis.
c
       call sldlin ( wxmin, wymin, wxmax, wymin, 0. )
       call sldlin ( wxmin, wymax, wxmax, wymax, 0. )
       if (ixsen .eq. 1)
     1   call axislab ( 1, wxmin, wxmax, wymin, wymax, scale, 1.0
     2                , nchmax, xlab, 1.0, iexp )
c
c-----------------------------------------------------------------------
c------------------------  T H E   Y - A X I S  ------------------------
c-----------------------------------------------------------------------
c
c      Set the parameters for the y-axis.
c
       ndec =-1
       if ( (lnlg .eq. 1) .or. (lnlg .eq. 3) ) then
         call axisset1 ( 2, ymin, ymax, imm, aspect, ndec, iexp, nchmax
     1                 , yymin, yymax, ndiv, nann, yann, cyann )
         iexp = iexp + yexp
       else
         call axisset2 ( 2, ymin, ymax, yexp, aspect, nchmax
     1                 , yymin, yymax, ndiv, nann, yann, cyann )
         iexp = 0
       endif
c
c      Write the y-axis annotations and draw the annotated ticks.  The
c  first y-axis annotation is skipped if both it and the first x-axis
c  annotation are at the origin.
c
       wx   = wxmin - 0.20  * scale
       fact = ( wymax - wymin ) / ( yymax - yymin )
       q1   = abs ( xann(1) - xxmin ) / ( xxmax - xxmin )
       q2   = abs ( yann(1) - yymin ) / ( yymax - yymin )
       ibeg = 1
       if ( (q1 .le. small) .and. (q2 .le. small) ) ibeg = 2
       q2   = 0.45 * scale
       q3   = 0.38 * scale
c
       iend = nann
       if (allyl .lt. 0) then
         q1 =  abs ( yymax - yann(nann) ) / ( yymax - yymin )
         if (q1 .lt. small) iend = nann - 1
       endif
       do 90 i=ibeg,iend
         wy = wymin + ( yann(i) - yymin ) * fact
         if (iysen .eq. 1) then
           nch = strlen ( cyann(i) )
           call pschars ( wx, wy-q2/2.0, q2, cyann(i)(1:nch), 0.0, 2 )
         endif
         call sldlin ( wxmin, wy, wxmin+q3, wy, 0.0 )
         call sldlin ( wxmax, wy, wxmax-q3, wy, 0.0 )
90     continue
c
c      Draw the unannotated ticks.
c
       q1 = 0.25  * scale
       if (ndiv .gt. 0) then
         dyy = ( yymax - yymin ) / real ( ndiv )
         do 100 i=2,ndiv
           yy = real ( i - 1 ) * dyy
           wy = wymin + yy * fact
           call sldlin ( wxmin, wy, wxmin+q1, wy, 0. )
           call sldlin ( wxmax, wy, wxmax-q1, wy, 0. )
100      continue
       else
         yy = yymin
110      continue
         i1 = int ( yy - 0.499 + sign ( 0.5, yy ) )
         yy = alog10 ( 10.0**yy + 10.0**i1 )
         if (yy .lt. yymax) then
           wy = wymin + ( yy - yymin ) * fact
           call sldlin ( wxmin, wy, wxmin+q1, wy, 0. )
           call sldlin ( wxmax, wy, wxmax-q1, wy, 0. )
           go to 110
         endif
       endif
c
c      Draw and label the axis.
c
       call sldlin ( wxmin, wymin, wxmin, wymax, 0. )
       call sldlin ( wxmax, wymin, wxmax, wymax, 0. )
       if (iysen .eq. 1)
     1   call axislab ( 2, wxmin, wxmax, wymin, wymax, scale, 1.0
     2                , nchmax, ylab, 1.0, iexp )
c
c-----------------------------------------------------------------------
c
c      Final adjustments for log axes.
c
       if (lnlg .ge. 3) then
         xxmin = 10.0**xxmin
         xxmax = 10.0**xxmax
       endif
       if ( (lnlg .eq. 2) .or. (lnlg .eq. 4) ) then
         yymin = 10.0**yymin
         yymax = 10.0**yymax
       endif
c
       return
       end
c
c=======================================================================
c=======================================================================
c
       subroutine pschars ( x, y, fsize, chars, ang, just )
c
c    abm:grfx03.pschars <------------- emulate NCAR typesetting language
c                                                          october, 2003
c
c    written by: Alexander Men'shchikov
c    modified 1: February, 2005 by David Clarke, generalised for format 
c                statements 2050 and 2060 in GRFXCNTR.
c    modified 2: July, 2010 by DAC; completely rewritten to make it more 
c                readable and robust, and to enable sub-sub, sub-super, 
c                super-sub, and super-super scripting.
c
c  PURPOSE: This subroutine emulates the NCAR library's typesetting
c  language used to draw labels and legends in plots, and is for use 
c  with the PSPLOT graphics library.
c
c  NCAR markup language:
c
c    Fonts: |F0|  = times-roman (default)
c           |F10| = times-roman italic
c           |F14| = times-roman italic bold
c           |F5|  = greek italic
c           |F18| = math characters
c           |F34| = more math characters
c
c    Position: |B| => subscript; |S| => superscript; |N| => normal
c
c    One may combine one position designator with one font change in
c    each "field" (bounded by ||), and then string as many fields
c    together as one needs
c
c  Note: this routine is not general, and works based on several 
c  formatting assumptions.  
c
c  1.  Position control characters (B, S, N) may appear in the same 
c      control sequence as the font specifier (e.g., F0) but must appear 
c      before the font specifier.  Thus, |BF0| is OK, |F0B| is not.  
c
c  2.  There may be one S, one B, or one or two Ns before a font 
c      specifier.  Thus, |SF0| and |NNF0| are OK, |NSF0| is not.
c
c  3.  Some key characters are sensitive to extraneous blanks.  Thus, 
c      and for example, |F18|U gives a multiplication symbol, |F18| U 
c      gives a 'Y'.
c
c  INPUT VARIABLES:
c    x,y        position coordinates (in grid units) of the first 
c               character sequence printed.  (999.,999.) => current pen 
c               position.
c    fsize      font size (in cm?)
c    chars      character string (in NCAR markup language) to be parsed 
c               and interpreted
c    ang        angle of text.  Currently, allowed angles are 0 (x-axis) 
c               and 90 (y-axis).  With a little work, any angle could be 
c               interpreted as PSPLOT doesn't restrict angles to 0 & 90.
c    just       =0 => (x,y) at left side of "text"
c               =1 => (x,y) at centre of "text"
c               =2 => (x,y) at right side of "text"
c               where "text" is the entire character string for (x,y)
c              .ne. (999.,999.), only the first character "token" 
c               otherwise.
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    STRLEN
c    KEKSYM   )
c    SUPER    )---- PSPLOT routines
c    SUBBER   )
c    SETFNT   )
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       integer       nlim
       parameter     ( nlim=100 )
c
       character*(*) chars
       character*128 word
       integer       i       , j       , k       , n       , nch
     1             , nch2    , i1      , i2      , nfont   , nfnt    
     2             , just    , k0      , kmax    , nmax    , lev
       real          x       , y       , xi      , yi      , fsize
     1             , fs      , ang     , width   , fact
c
       character*128 control (nlim), token   (nlim)
       integer       tlen    (nlim), idc     (nlim), level   (nlim)
c
c      External statements
c
       integer       strlen
       external      strlen  , keksym  , super   , subber  , setfnt
c
c-----------------------------------------------------------------------
c
       do 10 i=1,nlim
         control(i) = ''
         token  (i) = ''
         tlen   (i) = 0
10     continue
c
c      Find the location of all delineation characters (|) in the 
c  string.
c
       nch = strlen ( chars )
       n   = 0
       do 20 i=1,nch
         if (chars(i:i) .eq. '|') then
           n      = n + 1
           idc(n) = i
         endif
20     continue
       nmax = n
       if (nmax .eq. 0) then
c
c      If no delineation characters were found, assume the entire string 
c  is to be printed as is at the default font (F0 => nfnt=24).
c
         control(1) = 'F0'
         tlen   (1) = nch
         token  (1) = chars(1:nch)
         kmax       = 1
c
       else
c
c      If the first character (including blanks!) is not a |, assume 
c  this string is the first token to be printed at the default font, F0.
c
         k0 = 0
         if (idc(1) .gt. 1) then
           control(1) = 'F0'
           tlen   (1) = idc(1)-1
           token  (1) = chars(1:tlen(1))
           k0         = 1
         endif
c
c      For syntactically correct NCAR markup language, control codes 
c  follow an odd number of delineation characters, while character 
c  substrings to be printed (tokens) follow an even number of 
c  delineation characters.
c
         do 30 n=1,nmax
           k  = n/2
           i1 = idc(n) + 1
           i2 = nch
           if (n .lt. nmax) i2 = idc(n+1) - 1
           if (i2 .ge. i1) then
             if (n .eq. 2*k) then
               token  (k+k0  ) = chars(i1:i2)
               tlen   (k+k0  ) = i2 - i1 + 1
             else
               control(k+k0+1) = chars(i1:i2)
             endif
           endif
30       continue
         kmax = nmax/2 + k0
c
c      Squeeze out any blanks that may be in the control sequences.
c
         do 50 k=1,kmax
           write ( word, '(64x)' )
           i   = 0
           nch2 = strlen ( control(k) )
           do 40 n=1,nch2
             if (control(k)(n:n) .ne. ' ') then
               i = i + 1
               word(i:i) = control(k)(n:n)
             endif
40         continue
           control(k) = word
           if ( (word(1:1) .ne. 'F') .and. (word(1:1) .ne. 'N') .and. 
     1          (word(1:1) .ne. 'S') .and. (word(1:1) .ne. 'B') ) then
             write (6, 2010) word(1:i), chars(1:nch)
             control(k) = 'NF0'
           endif
50       continue
c
       endif
c
c      Determine the text level (sub-superscripting) applicable for each 
c  token.  Text can be at any one of seven positions:
c
c  lev = -3 => sub-subscript                     1 => super-subscript
c        -2 => subscript         0 => normal     2 => superscript
c        -1 => sub-superscrpt                    3 => super-superscript
c
       lev = 0
       do 60 k=1,kmax
         if (control(k)(1:1) .eq. 'N') then
           if (abs(lev) .eq. 2) lev = 0
           if ( (lev .eq. 1) .or. (lev .eq. 3) ) lev = 2
           if ( (lev .eq.-1) .or. (lev .eq.-3) ) lev =-2
         endif
         if (control(k)(2:2) .eq. 'N') then
           if (abs(lev) .eq. 2) lev = 0
           if ( (lev .eq. 1) .or. (lev .eq. 3) ) lev = 2
           if ( (lev .eq.-1) .or. (lev .eq.-3) ) lev =-2
         endif
         if (control(k)(1:1) .eq. 'B') then
           if (abs(lev) .eq. 2) lev = lev - 1
           if (lev .eq. 0) lev =-2
         endif
         if (control(k)(1:1) .eq. 'S') then
           if (abs(lev) .eq. 2) lev = lev + 1
           if (lev .eq. 0) lev = 2
         endif
         level(k) = lev
60     continue
c
c      The first token is centred according to "just" after which the 
c  remaining tokens are appended.  This is fine for left-justification
c  but not for centring or right-justification.  For the latter two, it 
c  is necessary to adjust the initial coordinates (xi,yi), which is done 
c  by estimating the length of text to be printed, including an account 
c  of the smaller characters that may be sub-superscripted.
c
       xi = x
       yi = y
       if (kmax .gt. 1) then
         width = 0.0
         do 70 k=2,kmax
           fact = 0.4
           if (    level(k)  .eq. 0) fact = 0.9
           if (abs(level(k)) .eq. 2) fact = 0.6
           width = width + fact * fsize * real ( tlen(k) )
           do 61 i=1,tlen(k)
             if (token(k)(i:i) .eq. '.') 
     1         width = width - 0.5 * fact * fsize
61         continue
70       continue
         if ( (ang .eq. 0.0) .and. (x .ne. 999.0) ) then
           if (just .eq. 1) xi = x - width * 0.5
           if (just .eq. 2) xi = x - width
         endif
         if ( (ang .eq. 90.0) .and. (y .ne. 999.0) ) then
           if (just .eq. 1) yi = y - width * 0.5
           if (just .eq. 2) yi = y - width
         endif
       endif
c
c      Print each token with the specified font and at the specified 
c  sub-superscript level.
c
       do 80 k=1,kmax
c
c      Set font.
c
         if ( (control(k)(1:1) .eq. 'F') .or.
     1        (control(k)(2:2) .eq. 'F').or.
     1        (control(k)(3:3) .eq. 'F') ) then
c
           if (control(k)(1:1) .eq. 'F') then
             read (control(k)(2:),'(i2)') nfont
           else
             if (control(k)(2:2) .eq. 'F') then
               read (control(k)(3:),'(i2)') nfont
             else
               read (control(k)(4:),'(i2)') nfont
             endif
           endif
c
c      Substitution of characters that are different in NCAR and PSPLOT
c
           if ( (nfont .eq. 18) .and. (token(k)(1:1) .eq. 'R') ) then
             token(k) = '||'
             tlen (k) = 2
           endif
           if ( (nfont .eq. 18) .and. (token(k)(1:1) .eq. 'U') ) then
             token(k) = '264'
             tlen (k) = -999
           endif
           if ( (nfont .eq. 18) .and. (token(k)(1:1) .eq. 'A') ) then
             token(k) = '136'
             tlen (k) = -999
           endif
           if ( (nfont .eq. 34) .and. (token(k)(1:1) .eq. 'Q') ) then
             token(k) = '321'
             tlen (k) = -999
           endif
           if ( (nfont .eq. 34) .and. (token(k)(1:1) .eq. 'W') ) then
             token(k) = '327'
             tlen (k) = -999
           endif
           if ( (nfont .eq. 34) .and. (token(k)(1:1) .eq. '0') ) then
             token(k) = '260'
             tlen (k) = -999
           endif
c
           nfnt =-1
c
c      default is NewCenturySchlbk-Roman
c
           if (nfont .eq.  0) nfnt = 24
c
c      Italic --> NewCenturySchlbk-Italic
c
           if (nfont .eq. 10) nfnt = 23
c
c      Bold --> NewCenturySchlbk-Bold
c
           if (nfont .eq. 13) nfnt = 21
c
c      Bold italic --> NewCenturySchlbk-BoldItalic
c
           if (nfont .eq. 14) nfnt = 22
c
c      Greek --> Symbol
c
           if (nfont .eq.  5) nfnt = 29
c
c      Symbol Set 1 --> Symbol
c
           if (nfont .eq. 18) nfnt = 29
c
c      Math Symbols --> Symbol
c
           if (nfont .eq. 34) nfnt = 29
c
           if (nfnt .eq. -1) then
             write (6, 2020) nfont, chars(1:nch)
             nfnt = 24
           endif
           call setfnt ( nfnt )
c
         endif
c
c      Font size.
c
         fs = fsize
         if (abs(level(k)) .eq. 1) fs = 2.0 * fsize / 3.0
         if (abs(level(k)) .eq. 3) fs = 2.0 * fsize / 3.0
c
c      Additional positional shift for secondary sub-superscripting.
c
         fact = 1.0
         if (level(k) .eq.-3) fact = 2.5
         if (level(k) .eq.-1) fact = 0.08
         if (level(k) .eq. 1) fact = 0.8
         if (level(k) .eq. 3) fact = 2.5
c
c      Write text.
c
         if (level(k) .ge. 1) then
           call super  ( token(k), tlen(k), fs, ang, fact )
         endif
         if (level(k) .eq. 0) then
           call keksym ( xi, yi, fs, token(k), ang, tlen(k), just )
         endif
         if (level(k) .le.-1) then
           call subber ( token(k), tlen(k), fs, ang, fact )
         endif
c
         xi = 999.0
         yi = 999.0
80     continue
c
c      Restore default font.
c
       call setfnt ( 24 )
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-----------------------------------------------------------------------
c
2010   format('PSCHARS : |---WARNING---> Unrecognised control '
     1       ,'sequence "',a,'" in input string:',/
     2       ,'PSCHARS : "',a,'"',/
     3       ,'PSCHARS : Recognised control sequences begin with "F", '
     4       ,'"N", "S", or "B".',/
     5       ,'PSCHARS : Resetting control sequence to "NF0"; '
     6       ,'execution continuing...')
2020   format('PSCHARS : |---WARNING---> Unknown font F',i2,' in input '
     1       ,'character string:',/
     2       ,'PSCHARS : "',a,'"',/
     3       ,'PSCHARS : Recognised fonts include F0, F5, F10, F13, '
     4       ,'F14, F18, and F34.',/
     5       ,'PSCHARS : Resetting to default font, F0; execution '
     6       ,'continuing...')
c
       return
       end
c
c=======================================================================
c=======================================================================
c
       function strlen ( word )
c
c    dac:grfx03.strlen <--------------------- length of character string
c    from dac:editor.strlen                               february, 1992
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE: This function returns the number of characters in the
c  character string "word".  Note that PSPLOT has a function LENSTR with 
c  essentially the same purpose.
c
c  INPUT VARIABLES:
c    word        input character*(*) string
c
c  OUTPUT VARIABLES:
c    strlen      number of characters up to and including the last
c                non-blank character in "word"
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [NONE]
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*(*) word
       integer       strlen  , i       , imax
c
c-----------------------------------------------------------------------
c
       imax   = len ( word )
       strlen = 0
       do 10 i=1,imax
         if (word(i:i) .ne. ' ') strlen = i
10     continue
c
       return
       end
c
c=======================================================================
c=======================================================================
c
       subroutine timedate ( chtime, chdate )
c
c    dac:plotit.timedate <------------------------ returns time and date 
c    from dac:zeus3d.wclock                                   june, 2018
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE: This routine returns the wall-clock time (hh:mm:ss) and date 
c  (dd/mm/yyyy) in character variables; a pared-down version of WCLOCK 
c  in dzeus36 meant only for gfortran, ifort, and pgfortran (not pgf77 
c  or pgf95).
c
c  INPUT VARIABLES:
c
c  OUTPUT VARIABLES:
c    chtime      time returned as hh:mm:ss
c    chdate      date returned as dd/mm/yyyy
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [none]
c
c-----------------------------------------------------------------------
c
       implicit      none
c
       character*(*) chtime  , chdate
       character*10  chstr
       integer       year    , month   , day     , hour    , minute  
     1             , second  , ncht    , nchd    , nch     
c
       integer       values  (   8)
c
c-----------------------------------------------------------------------
c
c  DATE_AND_TIME returns eight integers in "values", where:
c
c  values(1)   the year (e.g., 2011)
c  values(2)   the month of the year (1-12)
c  values(3)   the day of the month (1-31)
c  values(4)   the number of minutes from GMT (e.g., -240 for Halifax)
c  values(5)   the hour of the day (0-23)
c  values(6)   the minute of the hour (0-59)
c  values(7)   the second of the minute (0-59)
c  values(8)   the number of milliseconds of the second (0-999)
c
       call date_and_time ( values=values )
       year   = values(1)
       month  = values(2)
       day    = values(3)
       hour   = values(5)
       minute = values(6)
       second = values(7)
c
c      Set "chtime" and "chdate"
c
       ncht   = len ( chtime )
       chtime = ' '
       write ( chstr, 2010 ) hour, minute, second
       nch           = min ( 8, ncht )
       chtime(1:nch) = chstr(1:nch)
c
       nchd   = len ( chdate )
       chdate = ' '
       write ( chstr, 2020 ) day, month, year
       nch           = min ( 10, nchd )
       chdate(1:nch) = chstr(1:nch)
c
c-----------------------------------------------------------------------
c
2010   format(i2.2,':',i2.2,':',i2.2)
2020   format(i2,'/',i2.2,'/',i4.4)
c
       return
       end
c
c=======================================================================
c
c+++++++++++++++++++++++++++ PSPLOT ROUTINES +++++++++++++++++++++++++++
c
c  The following modules represent a subset of the full PSPLOT library 
c  (/Users/dclarke/grfx/psplot) that support simple line graphs.  No 2-D 
c  colour or gray-scale contouring routines are included.
c
c  All PSPLOT modules were kindly provided by the late Kevin Kohler,
c  formerly of Nova Southeastern University, Fort Lauderdale, FL.
c
c  Modifications to individual routines made by Alexander Men'shchikov 
c  and David Clarke.
c
c=======================================================================
       subroutine asterisk ( xc, yc, radius )
c----------------------------------------------------------------------- 
c  written by DAC, June 2010 using dk times as a template:  This routine 
c  draws an asterisk symbol on a 1 - D plot.
c----------------------------------------------------------------------- 
       implicit      none
c
       real          xc      , yc      , radius
c
       real          xa      (   2), ya      (   2)
c
       external      drwcrv  
c----------------------------------------------------------------------- 
       xa(1) = xc
       ya(1) = yc - radius
       xa(2) = xc
       ya(2) = yc + radius
       call drwcrv ( xa, ya, 2, 0.0, .false. )
c
       xa(1) = xc - 0.87 * radius
       ya(1) = yc - 0.50 * radius
       xa(2) = xc + 0.87 * radius
       ya(2) = yc + 0.50 * radius
       call drwcrv ( xa, ya, 2, 0.0, .false. )
c
       ya(1) = yc + 0.50 * radius
       ya(2) = yc - 0.50 * radius
       call drwcrv ( xa, ya, 2, 0.0, .false. )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine blkstp ( ch, ndim, a, leng )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine strips out blanks
c
c  idim        leading dimension of arr in calling program
c  ch          input character*1 array
c  ndim        dimension of "ch" and "a"
c  leng        number of non-blank characters
c----------------------------------------------------------------------- 
       implicit      none
c
       integer       ndim    , leng    , i       , l
c
       character*1   ch      (ndim), a       (ndim)
c----------------------------------------------------------------------- 
c
c      Strip out blanks only (leave in esc, etc.)
c
       i    = 1
       leng = 0
10     continue
       if ( ichar(ch(i)) .ne. 32 )then
         leng    = leng + 1
         a(leng) = ch(i)
       endif
c
       if ( i .eq. ndim ) then
c
c      Blankfill remainder of output array
c
         do 20 l=leng+1,ndim
           a(l) = ' '
20       continue
         go to 30
       endif
c
       i = i + 1
       go to 10
30     continue
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine circle ( xc, yc, radius, fill )
c----------------------------------------------------------------------- 
c  written by KEK: writes a (filled) circle as a symbol on a 1-D plot
c
c  xc, yc      coordinates of centre of circle
c  radius      radius of circle (inches)
c  fill        logical toggle indicating if circle is to be filled
c----------------------------------------------------------------------- 
       implicit      none
c
       logical        fill
       character*132  cmdstr  , scrc
       real           xc      , yc      , radius  , conver  , xci     
     1              , yci     , rad
c
       common      / plt1     / cmdstr
       common      / cnvcom   / conver
c
       integer       lenstr
       external      lenstr  , filler
c----------------------------------------------------------------------- 
       xci  = xc     * conver
       yci  = yc     * conver
       rad  = radius * conver
       scrc = ' '
       write ( scrc, '(f8.2,'' '',f8.2,'' '',f8.2,'' C'')' )
     1   xci, yci, rad
       if ( fill ) then
         cmdstr = 'Np '//scrc(1:lenstr(scrc,132))//' fill'
       else
         cmdstr = 'Np '//scrc(1:lenstr(scrc,132))//' stroke'
       endif
       call filler
c
       return
       end
c======================================================================= 
c======================================================================= 
c----------------------------------------------------------------------- 
       subroutine dshcrv ( x, y, npts, idshpn, thk )
c----------------------------------------------------------------------- 
c  Written by KEK: this routine draws a dashed curve connecting points 
c  (x(n),y(n)), n=1,npts.
c
c  x,y         coordinate arrays in inches
c  npts        dimension of x and y; number of points to connect
c              starting at (x(1),y(1))
c  idshpn      number of solid/blank patterns per inch.  0 => solid 
c              curve is drawn.
c  thk         thickness of the dashes (inches).  0.0 => a single line
c              is drawn for the dashes
c----------------------------------------------------------------------- 
       implicit      none
c
       integer       npts    , idshpn  , iset    , ipen    , n       
     1             , nn      
       real          thk     , curwid  , curwids , pi      , absx    
     1             , ang     , dist    , xe      , xnew    , ye      
     2             , ynew
       real*8        dlen    , dleft   , xsd     , ysd     , xed
     1             , yed
c
       real          x       (npts), y       (npts)
 
       common      / lcom     / curwid
c
       external      drwcrv  , plot
c
       integer       icall
       data          icall    / 1 /
       save          icall
c----------------------------------------------------------------------- 
       if (icall .eq. 0) then
         write(6,3000)
3000     format('PSPLOT: EXTERNAL: DSHCRV')
         icall = 1
       endif
c
       if (idshpn .le. 0) then
         call drwcrv ( x, y, npts, thk, .false. )
         return
       endif
c
       pi = acos (-1.0)
c
       iset = 0
       if ( (thk .ne. 0.0) .and. (thk.ne.curwid) ) then 
         iset    = 1
         curwids = curwid
         call setlw ( thk )
       endif
c
       dlen = 0.5d0 / dble(idshpn)
       ipen = 3
       call plot ( x(1), y(1), ipen )
       xsd  = dble(x(1))
       ysd  = dble(y(1))
       n    = 2
c
10     continue
       if ( n .gt. npts ) go to 30         ! we're done
       ipen = 5 - ipen                     !(3-->2, 2-->3)
c
c      Find angle to next data point.
c
       absx = abs ( x(n) - real(xsd) )
       if ( absx .lt. 1.0e-5 ) then
         ang = pi/2.0 * sign ( 1.0, y(n) - real(ysd) )
       else
         ang = atan2 ( y(n) - real(ysd), x(n) - real(xsd) )
       endif
c
c      Calculate end point of dash or space.
c
       xed = xsd + dlen * dble ( cos(ang) )
       yed = ysd + dlen * dble ( sin(ang) )
       xe  = real ( xed )
       ye  = real ( yed )
c
c      See if we can fit this in before next data point.
c
       dist = sqrt ( (x(n) - real(xsd) )**2 + ( y(n) - real(ysd) )**2 )
       if ( dist .le. dlen ) then
c
c      Can't fit
c
         dleft = dlen
         do 20 nn=n,npts
c
c      Find angle to next data point.
c
           absx = abs ( x(nn) - real(xsd) )
           if ( absx .lt. 1.0e-5 ) then
             ang = pi/2.0 * sign ( 1.0, y(nn) - real(ysd) )
           else
             ang = atan2 ( ( y(nn) - real(ysd) )
     1                   , ( x(nn) - real(xsd) ) )
           endif
           dist = sqrt ( ( x(nn) - real(xsd) )**2 
     1                 + ( y(nn) - real(ysd) )**2 )
           if ( dist .le. dleft ) then
c
c      Draw line to next data point and calculate rest.
c
             call plot ( x(nn), y(nn), ipen )
             xsd   = dble ( x(nn) )
             ysd   = dble ( y(nn) )
             dleft = dleft - dble ( dist )     !Remaining portion
             if ( dleft .eq. 0.0d0 ) then 
c
c      Exact fit, so just do next point
c
               n = nn + 1
               go to 10
             endif
           else
c
c      Draw remaining portion and move toward.
c
             xnew = real ( xsd + dleft * dcos ( dble(ang) ) )
             ynew = real ( ysd + dleft * dsin ( dble(ang) ) )
             call plot ( xnew, ynew, ipen )
             xsd  = dble ( xnew )
             ysd  = dble ( ynew )
             n    = nn
             go to 10
           endif
20       continue
       else
c
c      Can fit, so just draw normally.  Check to see if we should just 
c  move to data point itself since roundoff could cause weird angles.
c
         dist = sqrt ( ( x(n) - real(xed) )**2
     1               + ( y(n) - real(yed) )**2 )
         if ( dist .le. 1.0e-5 ) then
           xed = dble(x(n))
           yed = dble(y(n))
           xe  = real(xed)
           ye  = real(yed)
           n   = n + 1
         endif
         call plot ( xe, ye, ipen )
         xsd = dble ( xed )
         ysd = dble ( yed )
         go to 10
       endif
c
30     continue
c
c      Reset linewidth to original
c
       if ( iset .eq. 1 ) call setlw ( curwids )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine diamond ( xc, yc, height )
c----------------------------------------------------------------------- 
c  written by DAC, June 2010 using dk times as a template:  This routine 
c  draws a diamond symbol on a 1-D plot.
c----------------------------------------------------------------------- 
       implicit      none
c
       real          xc      , yc      , height  , width
c
       real          xa      (   4), ya      (   4)
c
       external      drwcrv
c----------------------------------------------------------------------- 
       width = 0.8 * height
       xa(1) = xc - width
       ya(1) = yc
       xa(2) = xc
       ya(2) = yc - height
       xa(3) = xc + width
       ya(3) = yc
       xa(4) = xc
       ya(4) = yc + height
       call drwcrv ( xa, ya, 4, 0.0, .true. )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine drwcrv ( xarr, yarr, n, thkk, closer )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine draws a curved line on the current plot.
c
c  xarr, yarr  coordinates through which curve is drawn
c  n           dimension of xarr, yarr
c  thkk        desired thickness of line; 0.0 => use current thickness
c  closer      .true.  => dashed line
c              .false. => solid line
c----------------------------------------------------------------------- 
       implicit      none
c
       logical       closer
       character*132 cmdstr
       integer       n       , nn
       real          thkk    , curwids , curwid
c
       real          xarr    (   n), yarr    (   n)
c
c      "curwid" is set in routine setlw
c
       common      / lcom     / curwid
       common      / plt1     / cmdstr
c
       external      filler  , setlw   , movet   , linet   , stroke
c----------------------------------------------------------------------- 
       cmdstr = 'Np'
       call filler
c
       if ( thkk .ne. 0.0 ) then
         curwids = curwid
         call setlw ( thkk )
       endif
       call movet ( xarr(1), yarr(1) )
       do 10 nn=2,n
         call linet ( xarr(nn), yarr(nn) )
         if ( mod ( nn, 100 ) .eq. 0 ) then
c
c      stroke now to avoid possible limit
c
           call stroke
           call movet ( xarr(nn), yarr(nn) )
         endif
10     continue
c
       if ( closer ) then
         cmdstr = 'Cs'
       else
         cmdstr = 'S'
       endif
       call filler
       if ( thkk .ne. 0.0 ) call setlw ( curwids )   !Reset linewidth
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine factor ( facc )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine ... 
c----------------------------------------------------------------------- 
       implicit      none
c
       character*132 cmdstr
       real          facc    , fac     , recipx  , recipy
c
       common      / plt1     / cmdstr
       common      / plt2     / fac
c
       external      filler
c----------------------------------------------------------------------- 
c
c      Unscale previous scaling
c
       recipx = 1.0 / fac
       recipy = 1.0 / fac
       write ( cmdstr, '(2f7.3,'' scale'')' ) recipx, recipy
       call filler
       fac = facc
       write ( cmdstr, '(2F7.3,'' scale'')' ) fac, fac
       call filler
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine filler
c----------------------------------------------------------------------- 
c  Written by KEK: this routine...
c----------------------------------------------------------------------- 
       implicit      none
c
       logical       ispace
       character*132 cmdstr
       integer       iunit   , ibslash , lc      , lcc     , itot
     1             , icclst  , l       , icc     , ii
c
       character*1   cmdc    ( 132)
c
       common      / plt1     / cmdstr
       common      / pspunit  / iunit
c
       integer       lenstr
       external      lenstr
c----------------------------------------------------------------------- 
       ibslash = 92
       lc      = 0
       lcc     = lenstr ( cmdstr, 132 )
       ispace  = .false.
c
c      "itot" is running total of left/right parentheses in text string
c      If itot=0 then we are not in text mode, i.e. left = right
c
       itot   = 0
       icclst =-999
       do 10 l=1,lcc
         icc = ichar ( cmdstr(l:l) )
c
c      Don't place 2 or more spaces together not in text mode.
c
         if ( (icc .eq. 32) .and. (ispace) .and. (itot .eq. 0) )
     1     go to 10
         if ( (icc .ge. 32) .and. (icc .le. 127) ) then
           lc       = lc + 1
           cmdc(lc) = cmdstr(l:l)
         endif
         if (icc .eq. 32) then
           ispace = .true.
         else
           ispace = .false.
         endif
         if (icclst .ne. ibslash) then
           if (icc .eq. 40) itot = itot + 1
           if (icc .eq. 41) itot = itot - 1
         endif
         icclst = icc
10     continue
c
c      Write cmdstr to output file
c
       write ( iunit, '(132a1)' ) ( cmdc(ii), ii=1,lc )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine keksym ( xp, yp, size, titlein, ang, nchar1, mjus )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine ...
c
c  Modified by ABM, 10/03: converted Hollerith strings to character 
c  strings; generalized to accept special characters (fonts 29 and 35)
c----------------------------------------------------------------------- 
       implicit      none
c
       character*1   bslash
       character*132 cmdstr  , curfnt  , scrc    , titleb  , titlec
       character*(*) titlein
       integer       nchar1  , mjus    , nchar   , mchar   , iht
     1             , ifntsz  , nfont   , ititle  , m       , njus
     2             , numc    
       real          xp      , yp      , size    , ang     , conver  
     1             , pi      , high    , rsize   , szrat   , wide    
     2             , xarg    , yarg    , xpos    , ypos    , arg     
     3             , argdeg
c
       character*4  ctitle(20)
c
       common      / fntcom   / curfnt  , ifntsz  , nfont
       common      / plt1     / cmdstr
       common      / cnvcom   / conver
       common      / kkplot   / szrat
c
       integer       lenstr
       external      filler  , blkstp  , lenstr
c----------------------------------------------------------------------- 
c
c      Stroke previous paths before this write
c
       cmdstr = 'S'
       call filler
       bslash = char(92)
       pi     = acos (-1.0)
c
       if ( nchar1 .eq. -999 ) then        ! octal code
         ctitle(1) = titlein
         read ( ctitle(1), '(I4)' ) ititle
         nchar     = 1
       else
         nchar     = nchar1
         titlec    = titlein
         if ( iabs(nchar) .lt. 132 ) titlec(nchar+1:132) = ' '
         read ( titlec, '(20a4)' ) ctitle
       endif
c
c      Choose proper font height, using current font.
c
       mchar = iabs ( nchar )
c
c      Set character size
c
       iht = int ( size * conver / 0.6 )     !0.6 factor is empirical
       if ( iht .ne. ifntsz ) then
         cmdstr = ' '
         write ( cmdstr, '(I3,'' Setf'')' ) iht
         call filler
         ifntsz = iht
       endif
c
       if ( nchar1 .eq. -999 ) then        !Octal code
         write ( titlec, '(A1,I10)' ) bslash, ititle
         call blkstp ( titlec, 132, titlec, numc )
       else
         write ( titlec, '(20a4)' ) ctitle
c
c      Check if "titlec" contains (, ), or "backslash." These characters 
c  must be treated specially by preceding them with a "backslash". Do 
c  this to ( and ) even though they might be balanced, i.e. () within a 
c  string, which can be treated normally.
c
         titleb = titlec
         numc   = 0
         do 10 m=1,mchar
           if ( (titleb(m:m) .eq. '(') .or. (titleb(m:m) .eq. ')') .or.
     1          (titleb(m:m) .eq. bslash) ) then
             numc              = numc + 1
             titlec(numc:numc) = bslash
           endif
           numc              = numc + 1
           titlec(numc:numc) = titleb(m:m)
10       continue
       endif
 
       mchar = numc
       xpos  = xp
       ypos  = yp
       if ( nchar .lt. 0 ) then
         njus = 0
       else
         njus = mjus
       endif
       rsize = size
c
       if ( xpos .eq. 999.0 ) then
         cmdstr = '/Xpos Xpres def'
         njus   = 0
       else
         cmdstr = ' '
         write ( cmdstr, '(f8.2,'' Xpd'')' ) xp * conver
       endif
       call filler
c
       if ( ypos .eq. 999.0 ) then
         cmdstr = '/Ypos Ypres def'
         njus   = 0
       else
         cmdstr = ' '
         write ( cmdstr, '(f8.2,'' Ypd'')' ) yp * conver
       endif
       call filler
c
       if ( (njus .ne. 0) .and. (njus .ne. 1) .and. (njus .ne. 2) ) then
         write ( *,2010 ) njus
2010     format (1x,'Incorrect justification code ',i5,'found in '
     1          ,'KEKSYMC, zero used')
         njus = 0
       endif
c
c      Strlen has already been "factored" by the choice of font height.
c  Since it will eventually be factored again, we must divide by factor 
c  now.
c
       cmdstr = '('//titlec(1:mchar)//') Le'
       arg  = ang * pi / 180.0
       xarg = 0.5 * cos(arg) * float (njus)
       yarg = 0.5 * sin(arg) * float (njus)
c
       if ( xarg .ne. 0.0 ) then
         scrc = ' '
         write ( scrc, '(f7.3,'' Xpjd'')') xarg
         cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1            //scrc(1:lenstr(scrc,132))
       endif
c
       if ( yarg .ne. 0.0 ) then
         scrc = ' '
         write ( scrc, '(f7.3,'' Ypjd'')' ) yarg
         cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1            //scrc(1:lenstr(scrc,132))
       endif
c
       if ( nchar .eq. -1 ) then    !centered symbol
         high = rsize
         wide = high * szrat
         xarg = 0.5 * ( high * sin(arg) - wide * cos(arg) )
         yarg =-0.5 * ( high * cos(arg) + wide * sin(arg) )
         xpos = xpos + xarg
         ypos = ypos + yarg
c
         if ( xarg .ne. 0.0 ) then
           scrc   = ' '
           write ( scrc, '(''/Xpos Xpos'',f8.2,'' add def'')' )
     1       xarg * conver
           cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1              //scrc(1:lenstr(scrc,132))
         endif
c
         if ( yarg .ne. 0.0 ) then
           scrc = ' '
           write ( scrc, '(''/Ypos Ypos'',f8.2,'' add def'')' )
     1       yarg * conver
           cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1              //scrc(1:lenstr(scrc,132))
         endif
       endif
c
       call filler
c
c      Move pen to proper coordinates
c
       scrc   = 'xyd mv'
       cmdstr = scrc(1:lenstr(scrc,132))
c
c      Set angle
c
       if ( ang .ne. 0.0 ) then
         scrc = ' '
         write(scrc,'(F7.1,'' rotate'')') ang
         cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1            //scrc(1:lenstr(scrc,132))
       endif
c
       scrc   = '('//titlec(1:mchar)//') show'
       cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1          //scrc(1:lenstr(scrc,132))
c
c      Reset angle
c
       if ( ang .ne. 0.0 ) then
         scrc = ' '
         write ( scrc, '(F7.1,'' rotate'')' ) -ang
         cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1            //scrc(1:lenstr(scrc,132))
       endif
       call filler
c
c      Start next char at 0.5 char width away
c
       argdeg = arg * 180.0 / pi
       cmdstr = ' '
       write ( cmdstr, '(f6.1,'' Xyps'')' ) argdeg
       call filler
c
       return
       end
c======================================================================= 
c======================================================================= 
       function lenstr ( string, ls )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine finds actual length of string by 
c  eliminating trailing blanks
c
c  string      character string
c  ls          declared length of string in calling program
c----------------------------------------------------------------------- 
       implicit      none
c
       character*(*) string
       integer       lenstr  , ls      , i       , is
c----------------------------------------------------------------------- 
       do 10 i=ls,1,-1
         is = i
         if ( string(i:i) .ne. char(32) ) go to 20
10     continue
       is = 0
20     continue
       lenstr = is
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine linet ( xcall, ycall )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine is same as a call to plot with ip=2, but 
c  does not perform the "stroke".
c----------------------------------------------------------------------- 
       implicit      none
c
       character*132 cmdstr
       real          xcall   , ycall   , conver
c
       common      / plt1     / cmdstr
       common      / cnvcom   / conver
c
       external      filler
c----------------------------------------------------------------------- 
       write ( cmdstr, '(2f8.2,'' L'')' ) xcall * conver, ycall * conver
       call filler
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine movet ( xcall, ycall )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine is same as call to plot with ip=3, but 
c  does not perform the "stroke".
c----------------------------------------------------------------------- 
       implicit      none
c
       character*132 cmdstr
       real          xcall   , ycall   , conver
c
       common      / plt1     / cmdstr
       common      / cnvcom   / conver
c
       external      filler
c----------------------------------------------------------------------- 
       write ( cmdstr, '(2f8.2,'' M'')' ) xcall * conver, ycall * conver
       call filler
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine newdev ( iusnb )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine changes the output device/filename to 
c  the user specifications. It must be called before psinit.
c
c  iusnb       new file name ascii string
c----------------------------------------------------------------------- 
       implicit      none
c
       character*80  fileout
       character*(*) iusnb
       integer       inew    , lus
c
       common      / io       / fileout , inew
c
       integer       lenstr
       external      lenstr
c----------------------------------------------------------------------- 
       lus     = lenstr ( iusnb, len(iusnb) )
       fileout = ' '
       fileout = iusnb(1:lus)
c
c      Set inew to 999 so that init knows to use this name
c
       inew = 999
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine ocircle ( xc, yc, rad )
c----------------------------------------------------------------------- 
c  Written by DAC: This routine draws an open circle.
c
c  xc, yc      coordinates of circle centre
c  rad         radius of circle (inches)
c----------------------------------------------------------------------- 
       implicit      none
c
       real          xc      , yc      , rad
c
       external      circle
c----------------------------------------------------------------------- 
       call circle ( xc, yc, 0.95*rad, .false. )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine plot ( xcall, ycall, ip )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine ...
c
c  idim        leading dimension of arr in calling program
c  xcall,ycall coordinates of...
c  ip          ...
c----------------------------------------------------------------------- 
       implicit      none
c
       character*80  scr
       character*132 cmdstr
       integer       ip      , iunit   , npage   , npageo  , ipp
     1             , nch
       real          xcall   , ycall   , conver
 
       common      / plt1     / cmdstr
       common      / cnvcom   / conver
       common      / pspunit  / iunit
       common      / pagcom   / npage, npageo
c
       external      filler  , blkstp
c----------------------------------------------------------------------- 
       ipp = iabs(ip)
       if ( ip .eq. 999 ) then   !    Terminate plot session.
         cmdstr = 'stroke showpage'
         call filler
         cmdstr = '%%Trailer'
         call filler
         write ( scr, '(i6)' ) npage
         call blkstp ( scr, 80, scr, nch )
         cmdstr = '%%Pages: '//scr(1:nch)
         call filler
         cmdstr = '%%EOF'
         call filler
         npage  = npageo
         close(iunit)
         return
       endif
c
c      Moving pen
c
       if ( ipp .eq. 3 ) then 
c
c      Stroke to paint previous path, then move to...
c
         write ( cmdstr, '(2f8.2,'' SM'')' )
     1     xcall * conver, ycall * conver
       else
         write ( cmdstr, '(2f8.2,'' LM'')' )
     1     xcall * conver, ycall * conver
       endif
       call filler
c
c      Reset origin if ip < 0
c
       if (ip .lt. 0 ) then
         write ( cmdstr, '(2f8.2,'' translate'')' )
     1     xcall * conver, ycall * conver
         call filler
       endif
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine plus ( xc, yc, radius )
c----------------------------------------------------------------------- 
c  written by DAC, 8/08 using dk square as a template: writes a plus 
c  sign as a symbol on a 1-D plot
c
c  xc, yc      coordinates of centre of plus sign
c  radius      radius of plus sign (inches)
c----------------------------------------------------------------------- 
       implicit      none
c
       real          xc      , yc      , radius
c
       real          xa      (2), ya      (2)
c
       external      drwcrv
c----------------------------------------------------------------------- 
       xa(1) = xc - radius
       ya(1) = yc
       xa(2) = xc + radius
       ya(2) = yc
       call drwcrv ( xa, ya, 2, 0.0, .false. )
       xa(1) = xc
       ya(1) = yc - radius
       xa(2) = xc
       ya(2) = yc + radius
       call drwcrv ( xa, ya, 2, 0.0, .false. )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine psinit ( portrait )
c----------------------------------------------------------------------- 
c  Written by KEK: initializes plot for hp plotter
c
c  portrait    .true. => portrait orientation (.false. => landscape)
c----------------------------------------------------------------------- 
       implicit      none
c
       logical       portrait, prtrt   , iopen
       character*132 cmdstr  , curfnt
       character*80  fileout
       integer       ioffp   , inew    , ientry  , ifntsz  
     1             , nfont   , iunit   , npage   , npageo
       real          spval   , conver  , fac     , szrat   , xsh     
     1             , ysh
c
       common      / conre1   / ioffp   , spval
       common      / plt1     / cmdstr
       common      / cnvcom   / conver
       common      / plt2     / fac
       common      / io       / fileout , inew
       common      / kkplot   / szrat
       common      / chpcom   / ientry  , prtrt
       common      / fntcom   / curfnt  , ifntsz  , nfont
       common      / pspunit  / iunit
       common      / pagcom   / npage   , npageo
c
       data          npageo   / 0 /
c
       integer       lenstr
       external      lenstr  , filler  , setfnt  , factor  , setlw
     1             , setgry  , setcolr , plot
c----------------------------------------------------------------------- 
       ioffp = 0
       spval = 0.0
c
c      Set conversion factor: conver=72.0 for inches; 72.0/2.54 for cm,
c  etc.
c
       conver = 72.0 / 2.54
       if (npageo .eq. 0) then
         npageo = 1
       else
         npageo = npage
       endif
       npage = 1
       prtrt = portrait
c
c      Use default name unless "newdev" has already been called (inew = 
c  999).
c
       if ( inew .eq. 0 ) then
         fileout = 'psplot.ps'
         inew    = 1
       endif
       if ( inew .eq. 999 )then
         inew    = 1
       endif
c
c      Open output file.
c
       iunit = 60
10     continue
       inquire ( iunit, opened=iopen )    ! get a non-existent unit
       if ( iopen ) then
         iunit = iunit + 1
         go to 10
       endif
c
       open ( iunit, file=fileout, status='unknown', form='formatted')
c
       cmdstr = '%!PS-Adobe-3.0'
       call filler
c
c      Next PS command introduced by Alexander Men'shchikov, Dec 2003.
c  It helps to avoid large white space at the top of a page when
c  displaying it with GhostView. This gives instant view of the entire 
c  plot, without any need to scroll it in the window. There may be a 
c  problem printing such a page, but one can switch Media in GhostView 
c  back to Letter or just delete/comment the line in the PS file.
c
c      cmdstr = '%%DocumentMedia: JustForScreen 585 652 80 white ( )'
c      call filler
c
       cmdstr =  '%%Title: '//fileout(1:lenstr(fileout,80))
       call filler
c
c       cmdstr =  '%%CreationDate: '//cdate//' '//ctime
c       call filler
c
       cmdstr =  '%%Creator: PSPLOT PostScript Plotting Package'
       call filler
c
       cmdstr =  '%%Pages: (atend)'
       call filler
c
       cmdstr = '%%EndComments'
       call filler
c
       cmdstr = '%Library Creator: Kevin E. Kohler <kevin@nova.edu>'
       call filler
c
       cmdstr = '%%BeginProlog'
       call filler
c
       cmdstr = '/inch {72 mul} bind def'
       call filler
c
       cmdstr = '/Ah {moveto lineto lineto stroke} def'
       call filler
       cmdstr = '/Ar {moveto 2 copy lineto 4  - 2 roll'
       call filler
       cmdstr = '     moveto lineto lineto stroke } def'
       call filler
c
       cmdstr = '/arcit {S /A2 exch def /A1 exch def /Rad exch def'
       call filler
       cmdstr = '        /Yc exch def /Xc exch def'
       call filler
       cmdstr = '        Xc Rad A1 cos mul add Yc Rad A1 sin mul add'
       call filler
       cmdstr = '        moveto newpath'
       call filler
       cmdstr = '        Xc Yc Rad A1 A2 arc stroke} def'
       call filler
c
       cmdstr = '/C {/Rad exch def /Yc exch def /Xc exch def'
       call filler
c
c      Closepath needed to avoid notch with fat line width
c
       cmdstr = '         Xc Yc Rad 0 360 arc closepath'
       call filler
       cmdstr = '        } def'
       call filler
c
       cmdstr = '/c0sf {closepath 0 setgray fill} def'
       call filler
c
       cmdstr = ' '
       write (cmdstr,'(''/cc {'',f7.4,'' mul def} def'')') conver
       call filler
c
       cmdstr = '/cd {cc /dy} def'
       call filler
c
       cmdstr = '/cf {closepath fill} def'
       call filler
c
       cmdstr = '/Cs {closepath stroke} def'
       call filler
c
       cmdstr = '/Cln {newpath 3 1 roll'
       call filler
       cmdstr = '      moveto {lineto} repeat clip newpath'
       call filler
       cmdstr = '     } def'
       call filler
c
       cmdstr = '/Cs {closepath stroke} def'
       call filler
c
       cmdstr = '/dd {/Dx dx def /Dy dy def} def'
       call filler
c
       cmdstr = '/Fb {newpath moveto '
       call filler
       cmdstr = ' Dx 0 rlineto 0 Dy rlineto Dx neg 0 rlineto closepath'
       call filler
       cmdstr = ' fill } def'
       call filler
c
       cmdstr = '/Fbn { newpath 3 1 roll moveto {lineto} repeat'
       call filler
       cmdstr = '       closepath fill } def'
       call filler
c
       cmdstr = '/Fbnc { newpath 3 1 roll moveto'
       call filler
       cmdstr = '       {lineto} repeat closepath fill } def'
       call filler
c
       cmdstr = '/L /lineto load def'
       call filler
c
       cmdstr = '/Le {/Strlen exch stringwidth pop def} def'
       call filler
c
c      Define stringlength slightly increased for integrand placement
c
       cmdstr = '/Lei {/Strlen exch stringwidth pop 1.5 mul def} def'
       call filler
c
c      Define stringlength slightly increased for summation placement
c
       cmdstr = '/Les {/Strlen exch stringwidth pop 1.1 mul def} def'
       call filler
c
       cmdstr = '/Lenssd '//
     1        '{/Strlenss exch stringwidth pop 3 mul 4 div def} def'
       call filler
c
       cmdstr = '/LM {2 copy lineto stroke moveto} def'
       call filler
c
       cmdstr = '/lm {Xp Yp lineto stroke mv} def'
       call filler
c
       cmdstr = '/M /moveto load def'
       call filler
c
       cmdstr = '/mv {Xp Yp moveto} def'
       call filler
c
       cmdstr = '/Np {newpath} def'
       call filler
c
       cmdstr = '/S /stroke load def'
       call filler
c
       cmdstr = '/Sc {setrgbcolor} def'
       call filler
c
       cmdstr = '/Sg {setgray} def'
       call filler
c
       cmdstr = '/Setf {Curfnt exch scalefont setfont} def'
       call filler
c
       cmdstr = '/SM {stroke moveto} def'
       call filler
c
       cmdstr = '/sm {stroke mv} def'
       call filler
c
       write ( cmdstr, '(''/Slw {'',f7.4,'' mul setlinewidth} def'')' )
     1   conver
       call filler
c
c      Minimum line width 300 dpi
c
       cmdstr = '/Slw0 {.24 setlinewidth} bind def'
       call filler
c
       cmdstr = '/Xpd {/Xpos exch def} def'
       call filler
c
       cmdstr = '/Xpjd '//
     1        ' {/Xpos exch Xpos exch Strlen mul sub def} def'
       call filler
c
       cmdstr = '/xyd {/Xp Xpos def /Yp Ypos def} def'
       call filler
c
       cmdstr = '/Xyp0d {/Xpos0 Xpres def /Ypos0 Ypres def} def'
       call filler
c
       cmdstr = '/Xyps {dup '//
     1        '/Xpres exch cos Strlen mul Xpos add def'
       call filler
       cmdstr = '              '//
     1        '/Ypres exch sin Strlen mul Ypos add def} def'
       call filler
c
       cmdstr = '/Xyps0 {dup '//
     1        '/Xpres exch cos Strlen mul Xpos0 add def'
       call filler
       cmdstr = '               '//
     1        '/Ypres exch sin Strlen mul Ypos0 add def} def'
       call filler
c
       cmdstr = '/Ypd {/Ypos exch def} def'
       call filler
c
       cmdstr = '/Ypjd '//
     1        ' {/Ypos exch Ypos exch Strlen mul sub def} def'
       call filler
c
       cmdstr = '/space ( ) def'
       call filler
c
       cmdstr = '%%EndProlog'
       call filler
c
       cmdstr = '%%Page: 1 1'
       call filler
c
c      "szrat" is the ratio of width to height of characters, determined
c  empirically to be: szrat = 0.6.  Set initial font to helvetica, 12 
c  point.
c
       szrat  = 0.6
       ifntsz = 12
       call setfnt ( 20 )
c
c      Set factor to 1 for initialization, reset later if CHOPIT called
c
       fac = 1.0
       call factor ( fac )
c
c      Set initial lineweight to 0.
c
       call setlw ( 0.0 )
c
c      Set initial grayscale to 0.
c
       call setgry ( 0.0 )
c
c      Set initial rgb colors to black(0).
c
       call setcolr ( 0.0, 0.0, 0.0 )
c
       if (.not. portrait) then
         cmdstr = '90 rotate 0 -8.5 inch translate '
         call filler
       endif
c
       call plot ( 18.0 / conver, 18.0 / conver, -3 )
c
       if ( portrait ) then
         xsh = 18.0 / conver
         ysh = 0.0
       else
         xsh = 0.0
         ysh = 18.0 / conver
       endif
       call plot ( xsh, ysh, -3 )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine setcolr ( red, green, blue )
c----------------------------------------------------------------------- 
c  Written by KEK: this routines sets the current colour
c
c  red, green, blue  rgb saturation ratios between 0 and 1
c----------------------------------------------------------------------- 
       implicit      none
c
       character*132 cmdstr
       real          red     , green   , blue    , cred    , cgreen  
     1             , cblue   , cgry    , r       , g       , b
c
       common      / plt1     / cmdstr
       common      / colrcom  / cred    , cgreen  , cblue   , cgry
c
       external      filler
c----------------------------------------------------------------------- 
       r = amin1 ( 1.0, amax1 ( 0.0, red   ) )
       g = amin1 ( 1.0, amax1 ( 0.0, green ) )
       b = amin1 ( 1.0, amax1 ( 0.0, blue  ) )
c
       cmdstr = ' '
       write ( cmdstr, '(3F7.3,'' Sc'')' ) r, g, b
       call filler
       cred   = r
       cgreen = g
       cblue  = b
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine setfnt ( numfnt )
c----------------------------------------------------------------------- 
c  Written by KEK: This routines changes the typeface of the current 
c  font.
c
c  Modified by DAC (8/08): fonts described and rationalised
c
c  numfnt         Font            numfnt         Font
c     1    AvantGarde-Book           2    AvantGarde-BookOblique
c     3    AvantGarde-Demi           4    AvantGarde-DemiOblique
c     5    Bookman-Demi              6    Bookman-DemiItalic
c     7    Bookman-Light             8    Bookman-LightItalic
c     9    Courier-Bold             10    Courier-BoldOblique
c    11    Courier-Oblique          12    Courier
c    13    Helvetica-Bold           14    Helvetica-BoldOblique
c    15    Helvetica-Narrow-Bold    16    Helvetica-Narrow-BoldOblique
c    17    Helvetica-Narrow-Oblique 18    Helvetica-Narrow
c    19    Helvetica-Oblique        20    Helvetica
c    21    NewCenturySchlbk-Bold    22    NewCenturySchlbk-BoldItalic
c    23    NewCenturySchlbk-Italic  24    NewCenturySchlbk-Roman
c    25    Palatino-Bold            26    Palatino-BoldItalic
c    27    Palatino-Italic          28    Palatino-Roman
c    29    Symbol                   30    Times-Bold
c    31    Times-BoldItalic         32    Times-Italic
c    33    Times-Roman              34    ZapfChancery-MediumItalic
c    35    ZapfDingbats
c
c----------------------------------------------------------------------- 
       implicit      none
c
       character*132 cmdstr  , scrc    , curfnt
       integer       numfnt  , ifntsz  , nfont
c
       character*40  fntnam  (  35)
c
       common      / fntcom   / curfnt  , ifntsz  , nfont
       common      / plt1     / cmdstr
c
       data          fntnam
     1             / 'AvantGarde-Book'
     2             , 'AvantGarde-BookOblique'
     3             , 'AvantGarde-Demi'
     4             , 'AvantGarde-DemiOblique'
     5             , 'Bookman-Demi'
     6             , 'Bookman-DemiItalic'
     7             , 'Bookman-Light'
     8             , 'Bookman-LightItalic'
     9             , 'Courier-Bold'
     1             , 'Courier-BoldOblique'
     2             , 'Courier-Oblique'
     3             , 'Courier'
     4             , 'Helvetica-Bold'
     5             , 'Helvetica-BoldOblique'
     6             , 'Helvetica-Narrow-Bold'
     7             , 'Helvetica-Narrow-BoldOblique'
     8             , 'Helvetica-Narrow-Oblique'
     9             , 'Helvetica-Narrow'
     1             , 'Helvetica-Oblique'
     2             , 'Helvetica'
     3             , 'NewCenturySchlbk-Bold'
     4             , 'NewCenturySchlbk-BoldItalic'
     5             , 'NewCenturySchlbk-Italic'
     6             , 'NewCenturySchlbk-Roman'
     7             , 'Palatino-Bold'
     8             , 'Palatino-BoldItalic'
     9             , 'Palatino-Italic'
     1             , 'Palatino-Roman'
     2             , 'Symbol'
     3             , 'Times-Bold'
     4             , 'Times-BoldItalic'
     5             , 'Times-Italic'
     6             , 'Times-Roman'
     7             , 'ZapfChancery-MediumItalic'
     8             , 'ZapfDingbats' /
c
       integer        lenstr
       external       filler  , lenstr
c----------------------------------------------------------------------- 
       nfont = numfnt
       if ( (numfnt .lt. 1) .or. (numfnt .gt. 35) ) then
         write (*,'(a)') 'Invalid font number encountered in **setfnt**'
         write (*,'(a)') 'Using Times-Roman default'
         nfont = 30
       endif
       scrc   = fntnam ( nfont )
       cmdstr = '/Curfnt /'//scrc(1:lenstr(scrc,132))//' findfont def'
       call filler
       write ( cmdstr, '(i3,'' Setf'')' ) ifntsz
       call filler
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine setgry ( gry )
c----------------------------------------------------------------------- 
c  Written by KEK: This routines sets the current gray level
c
c  gry         real between 0.0 and 1.0
c----------------------------------------------------------------------- 
       implicit      none
c
       character*132 cmdstr
       real          gry     , cred    , cgreen  , cblue   , cgry
     1             , g
c
       common      / plt1     / cmdstr
       common      / colrcom  / cred    , cgreen  , cblue   , cgry
c
       external      filler
c----------------------------------------------------------------------- 
       g      = amax1 ( 0.0, amin1 ( 1.0, gry) )
       cmdstr = ' '
       write ( cmdstr, '(F7.3,'' Sg'')' ) g
       call filler
       cgry   = g
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine setlw ( rlwi )
c----------------------------------------------------------------------- 
c  Written by KEK: this routines sets the current linewidth.
c
c  rlwi        desired linewidth in inches
c----------------------------------------------------------------------- 
       implicit      none
c
       character*132 cmdstr
       real          rlwi    , curwid
c
       common      / plt1     / cmdstr
       common      / lcom     / curwid
c
       external      filler
c----------------------------------------------------------------------- 
       if ( abs(rlwi) .lt. 1.0e-5 ) then
         cmdstr = 'Slw0'
       else
         cmdstr = ' '
         write ( cmdstr, '(F7.3,'' Slw'')' ) rlwi
       endif
       call filler
       curwid = rlwi
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine sldlin ( x1, y1, x2, y2, thk )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine draws a solid line between specified 
c  points.
c
c  x1, y1      beginning coordinate of line
c  x2, y2      end coordinate of line
c  thk         thickness of the line (inches).  0.0 => line is drawn
c              with current line weight setting.  The width of the line 
c              is centred on the line joining the two points.
c----------------------------------------------------------------------- 
       implicit      none
c
       integer       iset
       real          x1      , y1      , x2      , y2      , thk
     1             , curwid  , curwids
c
       common      / lcom     / curwid
c
       external      setlw   , plot
c----------------------------------------------------------------------- 
       iset = 0
       if ( (thk .gt. 0.0) .and. (thk .ne. curwid) ) then
         iset    = 1
         curwids = curwid
         call setlw ( thk )
       endif
       call plot ( x1, y1, 3 )
       call plot ( x2, y2, 2 )
c
c      Reset linewidth to original value.
c
       if (iset .eq. 1) call setlw ( curwids )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine square ( xc, yc, radius )
c----------------------------------------------------------------------- 
c  written by KEK: writes a square as a symbol on a 1-D plot
c
c  xc, yc      coordinates of centre of square
c  radius      radius of square (inches)
c----------------------------------------------------------------------- 
       implicit      none
c
       real          xc      , yc      , radius  , rad
c
       real          xa      (   4), ya      (   4)
c
       external      drwcrv
c----------------------------------------------------------------------- 
       rad   = 0.85 * radius
       xa(1) = xc - rad
       ya(1) = yc - rad
       xa(2) = xc + rad
       ya(2) = ya(1)
       xa(3) = xa(2)
       ya(3) = yc + rad
       xa(4) = xa(1)
       ya(4) = ya(3)
       call drwcrv ( xa, ya, 4, 0.0, .true. )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine star ( xc, yc, radius )
c----------------------------------------------------------------------- 
c  written by DAC, 6/10 using dk square as a template: writes a star 
c  as a symbol on a 1-D plot
c
c  xc, yc      coordinates of centre of star
c  radius      radius of star (inches)
c----------------------------------------------------------------------- 
       implicit      none
c
       real          xc      , yc      , radius
c
       real          xa      (  10), ya      (  10)
c
       external      drwcrv
c----------------------------------------------------------------------- 
       xa( 1) = xc
       ya( 1) = yc + radius
       xa( 2) = xc - 0.225 * radius
       ya( 2) = yc + 0.309 * radius
       xa( 3) = xc - 0.951 * radius
       ya( 3) = ya(2)
       xa( 4) = xc - 0.363 * radius
       ya( 4) = yc - 0.118 * radius
       xa( 5) = xc - 0.588 * radius
       ya( 5) = yc - 0.809 * radius
       xa( 6) = xc
       ya( 6) = yc - 0.382 * radius
       xa( 7) = xc + 0.588 * radius
       ya( 7) = ya(5)
       xa( 8) = xc + 0.363 * radius
       ya( 8) = ya(4)
       xa( 9) = xc + 0.951 * radius
       ya( 9) = ya(3)
       xa(10) = xc + 0.225 * radius
       ya(10) = ya(2)
       call drwcrv ( xa, ya, 10, 0.0, .true. )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine stroke
c----------------------------------------------------------------------- 
c  Written by KEK: This routine ... 
c----------------------------------------------------------------------- 
       implicit      none
c
       character*132 cmdstr
c
       common      / plt1     / cmdstr
c
       external      filler
c----------------------------------------------------------------------- 
       cmdstr = 'stroke'
       call filler
c
       return
       end
c======================================================================= 
c=======================================================================
       subroutine subber ( titlein, nchar1, size, ang, fact )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine produces subscripted text.
c
c  titlein     character(s) to be subscripted
c  nchar1      number of characters to be subscripted
c  size        height of subscripted variable
c  ang         angle of text
c  fact        parameter for second level of subscripting
c----------------------------------------------------------------------- 
c  Modified by ABM, 10/2003: replaced Hollerith for character strings; 
c  generalized to accept special characters (of fonts 29 and 35).
c
c  Modified by DAC, 7/2010: introduced "fact" for second level of super- 
c  subscripting (fact = 1.0  = > no change from before)
c----------------------------------------------------------------------- 
       implicit      none
c
       character*(*) titlein
       character*132 cmdstr  , curfnt  , scrc
       character*80  titleb  , titlec
       character*1   bslash
       integer       nchar1  , ifntsz  , nfont   , nchar   , mchar   
     1             , numc    , m       , ireset  , iht     , ititle
       real          size    , ang     , fact    , conver  , pi      
     1             , angrad  , ss      , down    , xshift  , yshift  
c
       character*4   ctitle  (  20)
c
       common      / fntcom   / curfnt  , ifntsz  , nfont
       common      / plt1     / cmdstr
       common      / cnvcom   / conver
c
       integer       lenstr
       external      filler  , blkstp  , lenstr  , setfnt
c----------------------------------------------------------------------- 
       bslash = char(92)
       pi     = acos (-1.0)
       angrad = ang * pi / 180.0
       ss     = 2.0 * size / 3.0
       down   = fact * ss / 2.0
       xshift = sin(angrad) * down
       yshift = cos(angrad) * down
c
       cmdstr = 'Xyp0d'
       call filler
c
       cmdstr = ' '
       write ( cmdstr, '(''/Xpos Xpres '',f8.2,'' add def '')' )
     1   xshift * conver
       call filler
       cmdstr = ' '
       write ( cmdstr, '(''/Ypos Ypres '',F8.2,'' sub def '')' )
     1   yshift * conver
       call filler
c
       if ( nchar1 .eq. -999 ) then
         ctitle(1) = titlein
         read ( ctitle(1), '(I4)' ) ititle
         nchar     = 1
       else
         nchar     = nchar1
         titlec    = titlein
         if ( iabs(nchar) .lt. 80 ) titlec(nchar+1:80) = ' '
         read ( titlec, '(20a4)' ) ctitle
       endif
c
c      Choose font proper height using currently set font
c
       mchar = iabs(nchar)
c
       if ( nchar1. eq. -999 ) then
         write ( titlec, '(A1,I10)' ) bslash, ititle
         call blkstp ( titlec, 80, titlec, mchar )
       else
         write ( titlec, '(20a4)' ) ctitle
c
c      Check if titlec contains (, ), or "backslash". These characters 
c  must be treated specially by preceding them with a "backslash". Do 
c  this to ( and ) even though they might be balanced, i.e. () within a 
c  string, which can be treated normally.
c
         titleb = titlec
         numc   = 0
         do 10 m=1,mchar
           if ( (titleb(m:m) .eq. '(') .or. (titleb(m:m) .eq. ')') .or.
     1          (titleb(m:m) .eq. bslash) ) then
             numc = numc + 1
             titlec(numc:numc) = bslash
           endif
           numc = numc + 1
           titlec(numc:numc) = titleb(m:m)
10       continue
         mchar = numc
       endif
c
       cmdstr = '('//titlec(1:mchar)//') Le'
c
c      Set character size
c
       iht = int ( ss * conver / 0.6 )     !0.6 factor is empirical
c
       ireset = 0
       if ( titlec(1:mchar) .ne. char(39) ) then
         if ( iht .ne. ifntsz ) then
           cmdstr = ' '
           write ( cmdstr, '(I3,'' Setf'')' ) iht
           call filler
           ifntsz = iht
         endif
       else        !Use prime, not apostrophe (do not change ifntsz)
         ireset = 1
         cmdstr = ' '
         write ( cmdstr
     1         ,'(''/Symbol findfont '',I3,'' scalefont setfont'')')iht
         call filler
       endif
c
       if ( titlec(1:mchar) .ne. char(39) ) then
         cmdstr = '('//titlec(1:mchar)//') Le'
       else
         cmdstr = '('//bslash//'242) Le'
       endif
c
c      Move pen to proper coordinates
c
       scrc   = 'xyd mv'
       cmdstr = cmdstr (1:lenstr(cmdstr,132))//' '
     1                 //scrc(1:lenstr(scrc,132))
c
c      Set angle
c
       if ( ang .ne. 0.0 ) then
         scrc = ' '
         write ( scrc, '(F7.1,'' rotate'')' ) ang
         cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1                     //scrc(1:lenstr(scrc,132))
       endif
c
       if ( titlec(1:mchar) .ne. char(39) ) then
         scrc = '('//titlec(1:mchar)//') show'
       else
         scrc = '('//bslash//'242) show'
       endif
       cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1                   //scrc(1:lenstr(scrc,132))
c
c      Reset angle.
c
       if ( ang .ne. 0.0 ) then
         scrc = ' '
         write ( scrc, '(F7.1,'' rotate'')') -ang
         cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1                     //scrc(1:lenstr(scrc,132))
       endif
       call filler
c
c      Compute next character position.
c
       cmdstr = ' '
       write ( cmdstr, '(f6.1,'' Xyps0'')' ) ang
       call filler
c
c      Reset font if we drew a prime
c
       if ( ireset .eq. 1 ) call setfnt ( nfont )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine super ( titlein, nchar1, size, ang, fact )
c----------------------------------------------------------------------- 
c  Written by KEK: This routine produces superscripted text.
c
c  titlein     character(s) to be superscripted
c  nchar1      number of characters to be superscripted
c  size        height of superscripted variable
c  ang         angle of text
c  fact        parameter for second level of superscripting
c----------------------------------------------------------------------- 
c  Modified by ABM, 10/2003: replaced Hollerith for character strings; 
c  generalized to accept special characters (of fonts 29 and 35).
c
c  Modified by DAC, 7/2010: introduced "fact" for second level of super- 
c  subscripting (fact = 1.0  = > no change from before)
c----------------------------------------------------------------------- 
       implicit      none
c
       character*(*) titlein
       character*132 cmdstr  , curfnt  , scrc    
       character*80  titlec
       character*1   bslash
       integer       nchar1  , ifntsz  , nfont   , nchar   , mchar   
     1             , iht     , ireset  
       real          size    , ang     , fact    , conver  , pi      
     1             , angrad  , ss      , up      , xshift  , yshift 
c
       common      / fntcom   / curfnt  , ifntsz  , nfont
       common      / plt1     / cmdstr
       common      / cnvcom   / conver
c
       integer       lenstr
       external      filler  , lenstr  , setfnt
c----------------------------------------------------------------------- 
       bslash = char(92)
       titlec = titlein
c
c      If nchar1 is lt 0, we're calling from oversup and we should
c  lower superscript slightly
c
       nchar  = abs(nchar1)
       pi     = acos(-1.0)
       angrad = ang * pi/180.
       ss     = 2.0 * size / 3.0
       if ( nchar1 .ge. 0 ) then
         up = fact * 0.8 * ss
       else
         up = fact * 0.5 * ss
       endif
       xshift = sin(angrad) * up
       yshift = cos(angrad) * up
       cmdstr = 'Xyp0d'
       call filler
c
       cmdstr = ' '
       write ( cmdstr, '(''/Xpos Xpres '',f8.2,'' sub def '')' )
     1   xshift * conver
       call filler
       cmdstr = ' '
       write ( cmdstr, '(''/Ypos Ypres '',F8.2,'' add def '')' )
     1   yshift * conver
       call filler
c
c      Choose proper height.
c
       mchar = iabs(nchar)
c
c      Set character size.
c
       iht = int ( ss * conver / 0.6 )     !0.6 factor is empirical
c
       ireset = 0
       if ( titlec(1:mchar) .ne. char(39) ) then
         if ( iht .ne. ifntsz ) then
           cmdstr = ' '
           write ( cmdstr, '(I3,'' Setf'')' ) iht
           call filler
           ifntsz = iht
         endif
       else        !Use prime, not apostrophe (do not change ifntsz)
         ireset = 1
         cmdstr = ' '
         write ( cmdstr
     1         ,'(''/Symbol findfont '',I3,'' scalefont setfont'')') iht
         call filler
       endif
c
c      DRAW PRIME, NOT APOSTROPHE
c
       if ( titlec(1:mchar) .ne. char(39) ) then
         cmdstr = '('//titlec(1:mchar)//') Le'
       else
         cmdstr = '('//bslash//'242) Le'
       endif
c
c      Move pen to proper coordinates.
c
       scrc   = 'xyd mv'
       cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1                   //scrc(1:lenstr(scrc,132))
c
c      Set angle.
c
       if ( ang .ne. 0.0 ) then
         scrc = ' '
         write ( scrc, '(F7.1,'' rotate'')' ) ang
         cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1                     //scrc(1:lenstr(scrc,132))
       endif
c
       if ( titlec(1:mchar) .ne. char(39) ) then
         scrc = '('//titlec(1:mchar)//') show'
       else
         scrc = '('//bslash//'242) show'
       endif
       cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1                   //scrc(1:lenstr(scrc,132))
c
c      Set angle.
c
       if ( ang .ne. 0.0 ) then
         scrc = ' '
         write ( scrc, '(F7.1,'' rotate'')' ) -ang
         cmdstr = cmdstr(1:lenstr(cmdstr,132))//' '
     1                     //scrc(1:lenstr(scrc,132))
       endif
       call filler
c
c      Compute next character position.
c
       cmdstr = ' '
       write ( cmdstr, '(f6.1,'' Xyps0'')' ) ang
       call filler
c
c      Reset font if we drew a prime.
c
       if ( ireset .eq. 1 ) call setfnt ( nfont )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine times ( xc, yc, radius )
c----------------------------------------------------------------------- 
c  written by DAC, 8/08 using dk square as a template: writes a 
c  multiplication sign as a symbol on a 1-D plot
c
c  xc, yc      coordinates of centre of multiplication sign
c  radius      radius of multiplication sign (inches)
c----------------------------------------------------------------------- 
       implicit      none
c
       real          xc      , yc      , radius  , rad
c
       real          xa      (2), ya      (2)
c
       external      drwcrv
c----------------------------------------------------------------------- 
       rad   = 0.9 * radius
       xa(1) = xc - rad
       ya(1) = yc - rad
       xa(2) = xc + rad
       ya(2) = yc + rad
       call drwcrv ( xa, ya, 2, 0.0, .false. )
       ya(1) = yc + rad
       ya(2) = yc - rad
       call drwcrv ( xa, ya, 2, 0.0, .false. )
c
       return
       end
c======================================================================= 
c======================================================================= 
       subroutine triangle ( xc, yc, radius )
c----------------------------------------------------------------------- 
c  written by DAC, 8/08 using dk square as a template: writes a triangle
c  as a symbol on a 1-D plot
c
c  xc, yc      coordinates of centre of triangle
c  radius      radius of triangle (inches)
c----------------------------------------------------------------------- 
       implicit      none
c
       real          xc      , yc      , radius  
c
       real          xa      (3), ya      (3)
c
       external      drwcrv
c----------------------------------------------------------------------- 
       xa(1) = xc - radius
       ya(1) = yc - radius
       xa(2) = xc + radius
       ya(2) = ya(1)
       xa(3) = xc
       ya(3) = yc + radius
       call drwcrv ( xa, ya, 3, 0.0, .true. )
c
       return
       end
c======================================================================= 
